/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "CapiMsg.h"
#include "c2msgdef.h"
#include "DCMTrace.h"
#include "osmem.h"
#include "atomar.h"
#include "dbgCAPI.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#ifndef NDEBUG

#define REGISTER_STATE_RELEASED     0
#define REGISTER_STATE_RELEASING    1
#define REGISTER_STATE_REGISTERING  2
#define REGISTER_STATE_REGISTERED   3

#endif

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

cp_connectresp   cpV_ConnectResp_Ignore   = { creject_Ignore,        0,
                                              { b1prot_HDLC, b2prot_X75, b3prot_Transparent, 0, 0, 0 },
                                              { 0, 0, 0, 0 } };
cp_connectresp   cpV_ConnectResp_Reject   = { creject_Reject_Normal, 0,
                                              { b1prot_HDLC, b2prot_X75, b3prot_Transparent, 0, 0, 0 },
                                              { 0, 0, 0, 0 }  };
cp_connectresp   cpV_ConnectResp_Busy     = { creject_Reject_Busy,   0,
                                              { b1prot_HDLC, b2prot_X75, b3prot_Transparent, 0, 0, 0 },
                                              { 0, 0, 0, 0 }  };
cp_connectb3resp cpV_ConnectB3Resp_Reject = { creject_Reject_Normal, 0 };

/*===========================================================================*\
    Interne Makros
\*===========================================================================*/

// NOTE: 1024 could be to small in some situations
//#define CapiCalcRegMsgBufSize(chans)    (1024 * (1 + (chans)))
#define CapiCalcRegMsgBufSize(chans)        (4096 * (1 + (chans)))

#define CapiMsg_SetMsgLen(name, len)                                    \
    set16 ((name), (len))


#define CapiMsg_SetApplID(name, applid)                                 \
    set16 ((name) + 2, (applid))


#define CapiMsg_SetCmd(name, capiCmd)                                   \
    (name)[4] = (capiCmd)


#define CapiMsg_SetSubCmd(name, capiSubCmd)                             \
    (name)[5] = (capiSubCmd)


#define CapiMsg_SetHeader(name, cmd, subcmd, msgnr, addr)               \
    CapiMsg_SetCmd (name, cmd);                                         \
    CapiMsg_SetSubCmd (name, subcmd);                                   \
    set16 ((name) + 6, (msgnr));                                        \
    set32 ((name) + 8, (addr))

#define CapiMsg_SetFullHeader(name, applid, cmd, subcmd, msgnr, addr)   \
    CapiMsg_SetMsgLen (name, CAPI_MSG_HEADER_ADDR_SIZE);                \
    CapiMsg_SetApplID (name, applid);                                   \
    CapiMsg_SetHeader (name, cmd, subcmd, msgnr, addr);

#if 0
#define CapiMsg_AddByte(name, val)                                      \
    set8 (name + get16(name), (val));                                   \
    set16 (napiMsg_AddByte(name, val)                                      \
    set8 (name + get16(name), (val));                                   \
    set16 (name, get16(name) + sizeof (c_byte))


#define CapiMsg_AddWordame, get16(name) + sizeof (c_byte))


#define CapiMsg_AddWord(name, val)                                      \
    set16 ((name) + get16 (name), (val));                               \
    set16 ((name), get16 (name) + sizeof (c_word))


#define CapiMsg_AddDWord(name, val)                                     \
    set32 ((name) + get16 (name), (val));                               \
    set16 ((name), get16 (name) + sizeof (c_dword))

static inline void CapiMsg_AddStruct (c_struct name, c_struct val) {
    if (val) {
        c_word len = CalcCapiStructLen (val);
        memcpy ((name + get16 (name)), val, len);
        set16 (name, (tUShort)(get16 (name) + len));
    } else  {
        CapiMsg_AddByte (name, 0);
    }
}
#endif


/*===========================================================================*\
    CCAPI20_MsgBase Konstruktor
\*===========================================================================*/

CCAPI20_MsgBase::CCAPI20_MsgBase (void)
  : PutMsgbufferQueue (CAPI_MAX_MESSAGE_SIZE, CAPI_MAX_MESSAGE_COUNTER),
    m_hCapiBase (vIllegalHandle),
    m_Applid (IllegalApplID),
    m_CurPutMessage (0),
    m_EmptyPMQInUse (0) {

    dhead ("CCAPI20_MsgBase-Konstruktor", DEBUG_CCAPI20_MsgBase);

    #ifndef NDEBUG
        m_RegisterState = REGISTER_STATE_RELEASED;
    #endif
}


/*===========================================================================*\
    CCAPI20_MsgBase Destruktor
\*===========================================================================*/

CCAPI20_MsgBase::~CCAPI20_MsgBase (void) {
    dhead ("CCAPI20_MsgBase-Destruktor", DEBUG_CCAPI20_MsgBase);
    dassert (m_hCapiBase == vIllegalHandle);
    dassert (GetApplID() == IllegalApplID);
    #ifndef NDEBUG
        dassert (m_RegisterState == REGISTER_STATE_RELEASED);
    #endif

    FreeCurPutMsg();
    if (m_hCapiBase != vIllegalHandle) {
        CapiBase_Release (m_hCapiBase);
        /*----- TODO: while (ApplID != IllegalApplID); ???? -----*/
    }
}


/*===========================================================================*\
    CCAPI20_MsgBase::REGISTER
\*===========================================================================*/

void CCAPI20_MsgBase::REGISTER (tUInt MaxB3Connection, tUInt MaxB3Blks, tUInt MaxSizeB3) {
    dhead ("CCAPI20_MsgBase::REGISTER", DEBUG_CCAPI20_MsgBase);
    dassert (GetApplID() == IllegalApplID);
    dassert (m_hCapiBase == vIllegalHandle);
    dwarning (m_CurPutMessage == 0);
    dwarning (PutMsgbufferQueue.IsQueueEmpty() == vTrue);
    #ifndef NDEBUG
        dassert (m_RegisterState == REGISTER_STATE_RELEASED);
        m_RegisterState = REGISTER_STATE_REGISTERING;
    #endif

    // TODO: Adjust size of PutMsgbufferQueue or MaxB3Connection!

    PutMsgbufferQueue.FreeQueue();
    FreeCurPutMsg();
    CapiBase_Register (this, CapiCalcRegMsgBufSize (MaxB3Connection), MaxB3Connection, MaxB3Blks, MaxSizeB3);
}


/*===========================================================================*\
    CCAPI20_MsgBase::RELEASE
\*===========================================================================*/

void CCAPI20_MsgBase::RELEASE  (void) {
    dhead ("CCAPI20_MsgBase::RELEASE", DEBUG_CCAPI20_MsgBase);
    #ifndef NDEBUG
        dassert (m_RegisterState == REGISTER_STATE_REGISTERED);
        m_RegisterState = REGISTER_STATE_RELEASING;
    #endif
    dassert (GetApplID() != IllegalApplID);
    dassert (m_hCapiBase != vIllegalHandle);
    dwarning (PutMsgbufferQueue.IsQueueEmpty() == vTrue);
    PutMsgbufferQueue.FreeQueue();      // TODO: hier geht oft der DISCONNECT_RESP verloren!
    if (GetApplID() != IllegalApplID) {
        m_Applid = IllegalApplID;
        CapiBase_Release (m_hCapiBase);
    }
}


/*===========================================================================*\
    CCAPI20_MsgBase REQ / RESP
\*===========================================================================*/

void CCAPI20_MsgBase::ALERT_REQ (c_word Messagenumber, c_dword PLCI, cp_additionalinfo *pAdditionalInfo) {
    if (pAdditionalInfo) {
        tUInt ailen = CalcCapiStructLen (pAdditionalInfo->BChannelInformation) +
                      CalcCapiStructLen (pAdditionalInfo->KeypadFacility) +
                      CalcCapiStructLen (pAdditionalInfo->UserUserData) +
                      CalcCapiStructLen (pAdditionalInfo->FacilityDataArray);
        if (ailen > 4 * sizeof (c_byte)) {
            AllocAndPutMsg (CAPI_ALERT, CAPI_REQ, Messagenumber, PLCI, "54444", ailen,
                            pAdditionalInfo->BChannelInformation, pAdditionalInfo->KeypadFacility,
                            pAdditionalInfo->UserUserData, pAdditionalInfo->FacilityDataArray);
            return;
        }
    }
    AllocAndPutMsg (CAPI_ALERT, CAPI_REQ, Messagenumber, PLCI, "1", 0);
}


void CCAPI20_MsgBase::CONNECT_REQ (c_word Messagenumber, c_dword Controller, cp_connectreq *Params) {
    char *format = "244252224442154444";
    tUInt ailen  = CalcCapiStructLen (Params->ai.BChannelInformation) +
                   CalcCapiStructLen (Params->ai.KeypadFacility) +
                   CalcCapiStructLen (Params->ai.UserUserData) +
                   CalcCapiStructLen (Params->ai.FacilityDataArray);
    if (ailen <= 4 * sizeof (c_byte)) {
        ailen  = 0;
        format = "24425222444211";
    }
    AllocAndPutMsg (CAPI_CONNECT, CAPI_REQ, Messagenumber, Controller, format, Params->CIPValue,
                    Params->CalledPartyNumber, Params->CallingPartyNumber, 0,
                    (CalcCapiStructLen (Params->bp.B1Configuration) +
                     CalcCapiStructLen (Params->bp.B2Configuration) +
                     CalcCapiStructLen (Params->bp.B3Configuration) + 3 * sizeof (c_word)),
                    Params->bp.B1Protocol,
                    Params->bp.B2Protocol, Params->bp.B3Protocol, Params->bp.B1Configuration,
                    Params->bp.B2Configuration, Params->bp.B3Configuration, 0, 0,
                    ailen, Params->ai.BChannelInformation, Params->ai.KeypadFacility,
                    Params->ai.UserUserData, Params->ai.FacilityDataArray);
}


void CCAPI20_MsgBase::CONNECT_RESP (c_word Messagenumber, c_dword PLCI, cp_connectresp *Params) {
    char *format = "252224444254444";
    tUInt ailen  = CalcCapiStructLen (Params->ai.BChannelInformation) +
                   CalcCapiStructLen (Params->ai.KeypadFacility) +
                   CalcCapiStructLen (Params->ai.UserUserData) +
                   CalcCapiStructLen (Params->ai.FacilityDataArray);
    if (ailen <= 4 * sizeof (c_byte)) {
        ailen  = 0;
        format = "25222444421";
    }
    AllocAndPutMsg (CAPI_CONNECT, CAPI_RESP, Messagenumber, PLCI, format, Params->Reject,
                    (CalcCapiStructLen (Params->bp.B1Configuration) +
                     CalcCapiStructLen (Params->bp.B2Configuration) +
                     CalcCapiStructLen (Params->bp.B3Configuration) + 3 * sizeof (c_word)),
                    Params->bp.B1Protocol, Params->bp.B2Protocol, Params->bp.B3Protocol, Params->bp.B1Configuration,
                    Params->bp.B2Configuration, Params->bp.B3Configuration, Params->ConnectedNumber, 0,
                    ailen, Params->ai.BChannelInformation, Params->ai.KeypadFacility,
                    Params->ai.UserUserData, Params->ai.FacilityDataArray);
}


/*===========================================================================*\
    CCAPI20_MsgBase::HandleMessageError
\*===========================================================================*/

void CCAPI20_MsgBase::HandleMessageError (tUInt type, c_info ErrorCode, c_message cmsg) {
    dhead ("CCAPI20_MsgBase::HandleMessageError", DEBUG_CCAPI20_MsgBase);
    dparams ("%x,%x", ErrorCode);
    if (ErrorCode == ci_Msg_IllAppNr) {
        if (GetApplID() != IllegalApplID) {
            RELEASE();
        }
    } else {
        switch (type) {
        case 0:
            // PutMessage Error
            dassert (cmsg != 0);
            switch (CapiMsg_GetSubCmd (cmsg)) {
            case CAPI_REQ:
                CapiMsg_SetSubCmd (cmsg, CAPI_CONF);
                switch (CapiMsg_GetCmd (cmsg)) {
                case CAPI_ALERT:
                case CAPI_CONNECT:
                case CAPI_DISCONNECT_B3:
                case CAPI_DISCONNECT:
                case CAPI_CONNECT_B3:
                case CAPI_INFO:
                case CAPI_LISTEN:
                case CAPI_RESET_B3:
                case CAPI_SELECT_B_PROTOCOL:
                    set16 (cmsg + CAPI_MSG_HEADER_ADDR_SIZE, (tUShort)ErrorCode);
                    CapiMsg_SetMsgLen (cmsg, CAPI_MSG_HEADER_ADDR_SIZE + sizeof (c_word));
                    break;

                case CAPI_DATA_B3:
                    set16 (cmsg + CAPI_MSG_HEADER_ADDR_SIZE, CapiMsg_GetWord (cmsg, sizeof (c_dword) + sizeof (c_word)));
                    set16 (cmsg + CAPI_MSG_HEADER_ADDR_SIZE + sizeof (c_word), (tUShort)ErrorCode);
                    CapiMsg_SetMsgLen (cmsg, CAPI_MSG_HEADER_ADDR_SIZE + 2 * sizeof (c_word));
                    break;

                case CAPI_FACILITY: {
                    for (tUInt i = CAPI_MSG_HEADER_ADDR_SIZE; i < CapiMsg_GetMsgLen (cmsg); i++) {
                        set8 (cmsg + i + sizeof (c_word), cmsg[i]);
                    }
                    set16 (cmsg + CAPI_MSG_HEADER_ADDR_SIZE, (tUShort)ErrorCode);
                    CapiMsg_SetMsgLen (cmsg, (c_word)(CapiMsg_GetMsgLen (cmsg) + sizeof (c_word)));
                    break;
                }

                case CAPI_MANUFACTURER:
                    /*----- Keine Ahnung was man hier tun sollte -----*/
                    dprintErr ("Error with MANUFACTURER_REQ not handled!");
                    break;

                default:
                    dprintErr ("Illegal Command for a CAPI-Request!");
                    break;
                }
                HandleGetMessage (cmsg);
                break;
            case CAPI_RESP:
                dprintErr ("RESP-Error: MsgNr=%x, Address=%x, Command=%x, Error=%x!",
                           CapiMsg_GetMsgNr (cmsg), CapiMsg_GetAddr (cmsg), CapiMsg_GetCmd(cmsg), ErrorCode);
                break;
            default:
                dprintErr ("Wrong Subcommand!");
                break;
            }
            break;
        case 1:
            // GetMessage Error
            RELEASE();
            // or make a GetMessage();
            break;
        case 2:
            // GetProfile Error
            break;
        default:
            dwarning (0);
            break;
        }
    }
}


/*===========================================================================*\
    CCAPI20_MsgBase::AllocAndPutMsg
\*===========================================================================*/

void CCAPI20_MsgBase::AllocAndPutMsg (c_byte Command, c_byte SubCommand, c_word Messagenumber,
                                      c_dword Address, char *format, ...) {
    dhead ("CCAPI20_MsgBase::AllocAndPutMsg", DEBUG_CCAPI20_MsgBase);
    dparams ("%x,%x,%x,%x,%s", Command, SubCommand, Messagenumber, Address, format);
    dassert (Command != 0);
    dassert (SubCommand != 0);
    //dassert (Address != 0);
    dassert (format != 0);
    va_list param;
    va_start (param, format);

    c_message cmsg = PutMsgbufferQueue.Alloc();
    dwarning (cmsg != 0);
    if (cmsg) {
        dprint ("Message(%x,%x) ", Command, SubCommand);
        FillMessage (cmsg, Command, SubCommand, Messagenumber, Address, format, param);
        //dprint ("Buffer=%x, Len=%x", cmsg, CapiMsg_GetMsgLen (cmsg));
        //dprintBuf (cmsg, CapiMsg_GetMsgLen(cmsg));
        PutMsgbufferQueue.InQueue (cmsg);
        EmptyPutMsgbufferQueue();
    } else {
        dwarning (0);
        c_byte lMessage[CAPI_MAX_MESSAGE_SIZE];
        FillMessage (lMessage, Command, SubCommand, Messagenumber, Address, format, param);
        HandleMessageError (0, ci_Int_MemoryFullErr, lMessage);
    }
    va_end (param);
}


/*===========================================================================*\
    CCAPI20_MsgBase::FillMessage
\*===========================================================================*/

void CCAPI20_MsgBase::FillMessage (c_message cmsg, c_byte Command, c_byte SubCommand, c_word Messagenumber,
                                   c_dword Address, char *format, va_list param) {
    dhead ("CCAPI20_MsgBase::FillMessage", DEBUG_CCAPI20_MsgBase);
    dparams ("%x,%x,%x,%x,%s", Command, SubCommand, Messagenumber, Address, format);
    dassert (cmsg != 0);
    dassert (Command != 0);
    dassert (SubCommand != 0);
    dassert (format != 0);
    dassert (param != 0);

    CapiMsg_SetFullHeader (cmsg, GetApplID(), Command, SubCommand, Messagenumber, Address);

    c_message lauf = cmsg + CAPI_MSG_HEADER_ADDR_SIZE;
    c_struct  s;
    c_word    l;
    while (*format != '\0') {
        switch (*format++) {
        case '1':
            dprint ("Byte-Add\n ");
            set8 (lauf++, (c_byte) va_arg (param, int));
            break;
        case '2':
            dprint ("Word-Add\n ");
            set16 (lauf, (c_word) va_arg (param, int));
            lauf += sizeof (c_word);
            break;
        case '3':
            dprint ("DWord-Add\n ");
            set32 (lauf, va_arg (param, c_dword));
            lauf += sizeof (c_dword);
            break;
        case '4':
            dprint ("Struct-Add\n ");
            s = va_arg (param, c_struct);
            if (s) {
                l = CalcCapiStructLen (s);
                memcpy (lauf, s, l);
                lauf += l;
            } else {
                set8 (lauf++, 0);
            }
            break;
        case '5':
            dprint ("StructLen-Add\n ");
            l = (c_word) va_arg (param, int);
            if (l > 254) {
                set8  (lauf++, 255);
                set16 (lauf, l);
                lauf += sizeof (c_word);
            } else {
                set8 (lauf++, (c_byte) l);
            }
            break;
        case '6':
            dprint ("QWord-Add\n ");
            set64 (lauf, va_arg (param, c_qword));
            lauf += sizeof (c_qword);
            break;
        default:
            dassert (0);
            break;
        }
    }
    CapiMsg_SetMsgLen (cmsg, (c_word)(lauf - cmsg));
}


/*===========================================================================*\
    CCAPI20_MsgBase::EmptyPutMsgbufferQueue
\*===========================================================================*/

void CCAPI20_MsgBase::EmptyPutMsgbufferQueue (void) {
    dhead ("CCAPI20_MsgBase::EmptyPutMsgbufferQueue", DEBUG_CCAPI20_MsgBase);
    if (atomicInc (&m_EmptyPMQInUse) == vTrue) {
        c_message DoIt = 0;
        do {
            if (m_CurPutMessage == 0) {
                DoIt = m_CurPutMessage = PutMsgbufferQueue.OutQueue();
                dassert ((m_CurPutMessage == 0) || (CapiMsg_GetMsgLen (m_CurPutMessage) >= CAPI_MSG_HEADER_ADDR_SIZE));
            }
            if (DoIt) {
                if (GetApplID() != IllegalApplID) {
                    dassert (CapiMsg_GetApplID (m_CurPutMessage) != IllegalApplID);
                    dassert (m_hCapiBase != vIllegalHandle);
                    CapiBase_PutMessage (m_hCapiBase, m_CurPutMessage);
                } else {
                    dprintWrn ("GetApplID() == IllegalApplID! ");
                    FreeCurPutMsg();
                }
            }
        } while (atomicDec (&m_EmptyPMQInUse) == vFalse);
    }
}


/*===========================================================================*\
    CCAPI20_MsgBase::RegisterComplete
\*===========================================================================*/

void CCAPI20_MsgBase::RegisterComplete (thCapiBase Base, c_word ApplId, c_info ErrorCode) {
    dhead ("CCAPI20_MsgBase::RegisterComplete", DEBUG_CCAPI20_MsgBase);
    dparams ("hBase=%p,ApplId=%x,ErrorCode=%x", Base, ApplId, ErrorCode);

    FreeCurPutMsg();
    if ((ErrorCode == ci_NoError) && (ApplId != IllegalApplID) && (Base != vIllegalHandle)) {
        m_hCapiBase = Base;
        m_Applid    = ApplId;
        #ifndef NDEBUG
            dassert (m_RegisterState == REGISTER_STATE_REGISTERING);
            m_RegisterState = REGISTER_STATE_REGISTERED;
        #endif
        REGISTER_COMPLETE (ci_NoError);
    } else {
        dassert (ErrorCode != ci_Reg_Busy);
        #ifndef NDEBUG
            dassert (m_RegisterState == REGISTER_STATE_REGISTERING);
            m_RegisterState = REGISTER_STATE_RELEASED;
        #endif
        REGISTER_COMPLETE ((ErrorCode) ? ErrorCode : ci_Reg_OSResourceErr);
    }
}


/*===========================================================================*\
    CCAPI20_MsgBase::ReleaseComplete
\*===========================================================================*/

void CCAPI20_MsgBase::ReleaseComplete (void) {
    dhead ("CCAPI20_MsgBase::ReleaseComplete", DEBUG_CCAPI20_MsgBase);
    FreeCurPutMsg();
    dassert (PutMsgbufferQueue.IsQueueEmpty() == vTrue);
    #ifndef NDEBUG
        dassert (m_RegisterState == REGISTER_STATE_RELEASING);
        m_RegisterState = REGISTER_STATE_RELEASED;
    #endif
    m_hCapiBase = vIllegalHandle;
    m_Applid    = IllegalApplID;
    PutMsgbufferQueue.FreeQueue();
    RELEASE_COMPLETE();
}


/*===========================================================================*\
    CCAPI20_MsgBase::PutMessageComplete
\*===========================================================================*/

void CCAPI20_MsgBase::PutMessageComplete (c_info ErrorCode) {
    dhead ("CCAPI20_MsgBase::PutMessageComplete", DEBUG_CCAPI20_MsgBase);
    dassert (m_CurPutMessage != 0);
    dassert ((m_CurPutMessage == 0) || (CapiMsg_GetMsgLen (m_CurPutMessage) >= CAPI_MSG_HEADER_ADDR_SIZE));
    #if !defined NDEBUG && defined USE_DEBUG_INFO_CLASS && (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_NONE)
        DbgCapiMsgTrace (m_CurPutMessage, ErrorCode, CMTRACE_PutMessage);
    #endif
    if (ErrorCode != ci_NoError) {
        dassert (ErrorCode != ci_Msg_Busy);
        dassert (ErrorCode != ci_Msg_SendQueueFull);
        HandleMessageError (0, ErrorCode, m_CurPutMessage);
    }
    FreeCurPutMsg();
    EmptyPutMsgbufferQueue();
}


/*===========================================================================*\
    CCAPI20_MsgBase::GetMessageReceived
\*===========================================================================*/

tBool CCAPI20_MsgBase::GetMessageReceived (c_message Message, c_info ErrorCode) {
    dhead ("CCAPI20_MsgBase::GetMessageReceived", DEBUG_CCAPI20_MsgBase);
    dparams ("buf=%p, len=%x, err=%x", Message, (Message) ? CapiMsg_GetMsgLen(Message) : 0,
             ErrorCode);
    dassert (ErrorCode != ci_Msg_Busy);
    dassert (ErrorCode != ci_Msg_SendQueueFull);
    #if !defined NDEBUG && defined USE_DEBUG_INFO_CLASS && (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_NONE)
        DbgCapiMsgTrace (Message, ErrorCode, CMTRACE_GetMessage);
    #endif
    #ifdef NDEBUG
        dwarning (m_RegisterState == REGISTER_STATE_REGISTERED);
    #endif
    if (GetApplID() == IllegalApplID) {
        RETURN ('x', vFalse);
    }
    if (ErrorCode == ci_NoError) {
        dassert (m_hCapiBase != vIllegalHandle);
        dassert (Message != 0);
        HandleGetMessage (Message);
    } else {
        dassert (ErrorCode != ci_Msg_Busy);
        dassert (ErrorCode != ci_Msg_SendQueueFull);
        HandleMessageError (1, ErrorCode, 0);
    }
    RETURN ('x', vTrue);
}


/*===========================================================================*\
    CCAPI20_MsgBase::HandleGetMessage
\*===========================================================================*/

void CCAPI20_MsgBase::HandleGetMessage (c_message Message) {
    dhead ("CCAPI20_MsgBase::HandleGetMessage", DEBUG_CCAPI20_MsgBase);
    dassert (Message != 0);
    dassert (CapiMsg_GetMsgLen (Message) >= CAPI_MSG_HEADER_ADDR_SIZE);
    //dprintBuf (Message, CapiMsg_GetMsgLen(Message));
    switch (CapiMsg_GetSubCmd (Message)) {
    /*-------------------------------------------------------------------*\
        CAPI - Confirmations
    \*-------------------------------------------------------------------*/
    case CAPI_CONF:
        switch (CapiMsg_GetCmd (Message)) {
        case CAPI_ALERT:
            ALERT_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                        CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_CONNECT:
            CONNECT_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                          CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_CONNECT_B3:
            CONNECT_B3_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                             CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_DATA_B3: {
            cp_datab3conf pdc = {
                CapiMsg_GetWord (Message, 0),
                CapiMsg_GetInfo (Message, sizeof (c_word))
            };
            DATA_B3_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pdc);
            break;
        }
        case CAPI_DISCONNECT_B3:
            DISCONNECT_B3_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_DISCONNECT:
            DISCONNECT_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                             CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_FACILITY: {
            cp_facilityconf pf;
            pf.Info             = CapiMsg_GetInfo (Message, 0);
            pf.FacilitySelector = (c_facilselect)CapiMsg_GetWord (Message, sizeof (c_word));
            pf.FacilityParam    = CapiMsg_GetStruct (Message, 2 * sizeof (c_word));
            FACILITY_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pf);
            break;
        }
        case CAPI_INFO:
            INFO_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                       CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_LISTEN:
            LISTEN_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                         CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_MANUFACTURER:
            if (CapiMsg_GetMsgLen (Message) > 3 * sizeof (c_dword)) {
                cp_manufacturer pm;
                memset (&pm, '\0', sizeof (cp_manufacturer));
                pm.ManuID   = CapiMsg_GetDWord (Message, 0);
                pm.Class    = CapiMsg_GetDWord (Message, sizeof (c_dword));
                pm.Function = CapiMsg_GetDWord (Message, 2 * sizeof (c_dword));
                pm.ManuData = CapiMsg_GetStruct (Message, 3 * sizeof (c_dword));
                MANUFACTURER_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pm);
            } else {
                dwarning (0);
            }
            break;
        case CAPI_RESET_B3:
            RESET_B3_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                           CapiMsg_GetInfo (Message, 0));
            break;

        case CAPI_SELECT_B_PROTOCOL:
            SELECT_B_PROTOCOL_CONF (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                    CapiMsg_GetInfo (Message, 0));
            break;
        }
        break;

    /*-------------------------------------------------------------------*\
        CAPI - Indications
    \*-------------------------------------------------------------------*/
    case CAPI_IND:
        switch (CapiMsg_GetCmd (Message)) {
        case CAPI_CONNECT: {
            cp_connectind pci;
            pci.CIPValue           = (c_cipvalue)CapiMsg_GetWord (Message, 0);
            pci.CalledPartyNumber  = CapiMsg_GetStruct (Message, sizeof (c_word));
            pci.CallingPartyNumber = CapiMsg_GetStruct (Message, CalcCapiStructLen (pci.CalledPartyNumber) +
                                                                 sizeof(c_word));
            if (CONNECT_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pci) == vFalse) {
                CCAPI20_MsgBase::CONNECT_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pci);
            }
            break;
        }
        case CAPI_CONNECT_ACTIVE:
            if (CONNECT_ACTIVE_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                CapiMsg_GetStruct (Message, 0)) == vFalse) {
                CCAPI20_MsgBase::CONNECT_ACTIVE_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                                     CapiMsg_GetStruct (Message, 0));
            }
            break;

        case CAPI_CONNECT_B3:
            if (CONNECT_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                            CapiMsg_GetStruct (Message, 0)) == vFalse) {
                CCAPI20_MsgBase::CONNECT_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                                 CapiMsg_GetStruct (Message, 0));
            }
            break;

        case CAPI_CONNECT_B3_ACTIVE:
            if (CONNECT_B3_ACTIVE_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                       CapiMsg_GetStruct (Message, 0)) == vFalse) {
                CCAPI20_MsgBase::CONNECT_B3_ACTIVE_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                                        CapiMsg_GetStruct (Message, 0));
            }
            break;

        case CAPI_CONNECT_B3_T90_ACTIVE:
            if (CONNECT_B3_T90_ACTIVE_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message)) == vFalse) {
                CCAPI20_MsgBase::CONNECT_B3_T90_ACTIVE_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message));
            }
            break;

        case CAPI_DATA_B3: {
            dassert (CapiMsg_GetMsgLen (Message) >= CAPI_MSG_DATA_POS + sizeof (c_dataptr));
            cp_datab3 pdb3;
            pdb3.Data       = CapiMsg_GetData (Message);
            pdb3.DataLength = CapiMsg_GetWord (Message, sizeof (c_dword));
            pdb3.DataHandle = CapiMsg_GetWord (Message, sizeof (c_dword) + sizeof (c_word));
            pdb3.Flags      = CapiMsg_GetWord (Message, sizeof (c_dword) + 2 * sizeof (c_word));
            if (DATA_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pdb3) == vFalse) {
                CCAPI20_MsgBase::DATA_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pdb3);
            }
            break;
        }
        case CAPI_DISCONNECT_B3: {
            cp_disconnectb3ind pdbi;
            pdbi.ReasonB3 = CapiMsg_GetInfo (Message, 0);
            pdbi.NCPI     = CapiMsg_GetStruct (Message, sizeof (c_word));
            if (DISCONNECT_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pdbi) == vFalse) {
                CCAPI20_MsgBase::DISCONNECT_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pdbi);
            }
            break;
        }
        case CAPI_DISCONNECT:
            if (DISCONNECT_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                CapiMsg_GetInfo (Message, 0)) == vFalse) {
                CCAPI20_MsgBase::DISCONNECT_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                CapiMsg_GetInfo (Message, 0));
            }
            break;

        case CAPI_FACILITY: {
            cp_facility pf = {
                (c_facilselect)CapiMsg_GetWord (Message, 0),
                CapiMsg_GetStruct (Message, sizeof (c_word))
            };
            if (FACILITY_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pf) == vFalse) {
                CCAPI20_MsgBase::FACILITY_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pf);
            }
            break;
        }
        case CAPI_INFO: {
            cp_infoind pii;
            pii.InfoNumber  = CapiMsg_GetWord (Message, 0);
            pii.InfoElement = CapiMsg_GetStruct (Message, sizeof (c_word));
            if (INFO_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pii) == vFalse) {
                CCAPI20_MsgBase::INFO_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pii);
            }
            break;
        }
        case CAPI_MANUFACTURER:
            if (CapiMsg_GetMsgLen (Message) > 3 * sizeof (c_dword)) {
                cp_manufacturer pm;
                memset (&pm, '\0', sizeof (cp_manufacturer));
                pm.ManuID   = CapiMsg_GetDWord (Message, 0);
                pm.Class    = CapiMsg_GetDWord (Message, sizeof (c_dword));
                pm.Function = CapiMsg_GetDWord (Message, 2 * sizeof (c_dword));
                pm.ManuData = CapiMsg_GetStruct (Message, 3 * sizeof (c_dword));
                if (MANUFACTURER_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pm) == vFalse) {
                    CCAPI20_MsgBase::MANUFACTURER_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message), &pm);
                }
            } else {
                dwarning (0);
            }
            break;

        case CAPI_RESET_B3:
            if (RESET_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                              CapiMsg_GetStruct (Message, 0)) == vFalse) {
                CCAPI20_MsgBase::RESET_B3_IND (CapiMsg_GetMsgNr (Message), CapiMsg_GetAddr (Message),
                                               CapiMsg_GetStruct (Message, 0));
            }
            break;
        }
    }
}


/*===========================================================================*\
\*===========================================================================*/
