/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#ifndef _CAPIMSG_H_
#define _CAPIMSG_H_

#include "CapiBase.h"
#include "c2msgdef.h"
#include "DynStore.h"
#include <stdarg.h>

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

extern cp_connectresp   cpV_ConnectResp_Ignore;
extern cp_connectresp   cpV_ConnectResp_Reject;
extern cp_connectresp   cpV_ConnectResp_Busy;
extern cp_connectb3resp cpV_ConnectB3Resp_Reject;

/*===========================================================================*\
\*===========================================================================*/

class CCAPI20_MsgBase : private CCallBackCapiBase {
public:
    CCAPI20_MsgBase (void);
    virtual ~CCAPI20_MsgBase (void);

    c_word GetApplID (void);

    /*-----------------------------------------------------------------------*\
        CAPI - REGISTER, RELEASE, GET_PROFILE
    \*-----------------------------------------------------------------------*/
    void REGISTER (tUInt MaxB3Connection, tUInt MaxB3Blks, tUInt MaxSizeB3);
    void RELEASE  (void);

    void GET_PROFILE (tUInt Controller);

    virtual void REGISTER_COMPLETE (c_info ErrorCode) = 0;
    virtual void RELEASE_COMPLETE (void) = 0;

    virtual void GET_PROFILE_COMPLETE (cp_getprofile * /*pProfile*/, c_info /*ErrorCode*/) {}

    /*-----------------------------------------------------------------------*\
        CAPI - Requests
    \*-----------------------------------------------------------------------*/
    void ALERT_REQ                         (c_word             Messagenumber,
                                            c_dword            PLCI,
                                            cp_additionalinfo *pAdditionalInfo);

    void CONNECT_REQ                       (c_word         Messagenumber,
                                            c_dword        Controller,
                                            cp_connectreq *Params);

    void CONNECT_B3_REQ                    (c_word   Messagenumber,
                                            c_dword  PLCI,
                                            c_struct NCPI);

    void DATA_B3_REQ                       (c_word     Messagenumber,
                                            c_dword    NCCI,
                                            cp_datab3 *Params);

    void DISCONNECT_B3_REQ                 (c_word   Messagenumber,
                                            c_dword  NCCI,
                                            c_struct NCPI);

    void DISCONNECT_REQ                    (c_word   Messagenumber,
                                            c_dword  PLCI);

    void FACILITY_REQ                      (c_word       Messagenumber,
                                            c_dword      Address,
                                            cp_facility *Params);

    void INFO_REQ                          (c_word      Messagenumber,
                                            c_dword     Address,
                                            cp_inforeq *Params);

    void LISTEN_REQ                        (c_word        Messagenumber,
                                            c_dword       Controller,
                                            cp_listenreq *Params);

    void MANUFACTURER_REQ                  (c_word           Messagenumber,
                                            c_dword          Controller,
                                            cp_manufacturer *Params);

    void RESET_B3_REQ                      (c_word   Messagenumber,
                                            c_dword  NCCI,
                                            c_struct NCPI);

    void SELECT_B_PROTOCOL_REQ             (c_word        Messagenumber,
                                            c_dword       PLCI,
                                            cp_bprotocol *Params);

    /*-----------------------------------------------------------------------*\
        virtual CAPI - Confirmation
    \*-----------------------------------------------------------------------*/

    virtual void ALERT_CONF                (c_word  /*Messagenumber*/,
                                            c_dword /*PLCI*/,
                                            c_info  /*Info*/) {}

    virtual void CONNECT_CONF              (c_word  /*Messagenumber*/,
                                            c_dword /*PLCI*/,
                                            c_info  /*Info*/) {}

    virtual void CONNECT_B3_CONF           (c_word  /*Messagenumber*/,
                                            c_dword /*NCCI*/,
                                            c_info  /*Info*/) {}

    virtual void DATA_B3_CONF              (c_word          /*Messagenumber*/,
                                            c_dword         /*NCCI*/,
                                            cp_datab3conf * /*Params*/) {}

    virtual void DISCONNECT_B3_CONF        (c_word  /*Messagenumber*/,
                                            c_dword /*NCCI*/,
                                            c_info  /*Info*/) {}

    virtual void DISCONNECT_CONF           (c_word  /*Messagenumber*/,
                                            c_dword /*PLCI*/,
                                            c_info  /*Info*/) {}

    virtual void FACILITY_CONF             (c_word            /*Messagenumber*/,
                                            c_dword           /*Address*/,
                                            cp_facilityconf * /*Params*/) {}

    virtual void INFO_CONF                 (c_word  /*Messagenumber*/,
                                            c_dword /*Address*/,
                                            c_info  /*Info*/) {}

    virtual void LISTEN_CONF               (c_word  /*Messagenumber*/,
                                            c_dword /*Controller*/,
                                            c_info  /*Info*/) {}

    virtual void MANUFACTURER_CONF         (c_word            /*Messagenumber*/,
                                            c_dword           /*Controller*/,
                                            cp_manufacturer * /*Params*/) {}

    virtual void RESET_B3_CONF             (c_word  /*Messagenumber*/,
                                            c_dword /*NCCI*/,
                                            c_info  /*Info*/) {}

    virtual void SELECT_B_PROTOCOL_CONF    (c_word  /*Messagenumber*/,
                                            c_dword /*PLCI*/,
                                            c_info  /*Info*/) {}

    /*-----------------------------------------------------------------------*\
        virtual CAPI - Indications
    \*-----------------------------------------------------------------------*/

    virtual tBool CONNECT_IND               (c_word         Messagenumber,
                                             c_dword        PLCI,
                                             cp_connectind *Params);

    virtual tBool CONNECT_ACTIVE_IND        (c_word   Messagenumber,
                                             c_dword  PLCI,
                                             c_struct ConnectedNumber);

    virtual tBool CONNECT_B3_IND            (c_word   Messagenumber,
                                             c_dword  NCCI,
                                             c_struct NCPI);

    virtual tBool CONNECT_B3_ACTIVE_IND     (c_word   Messagenumber,
                                             c_dword  NCCI,
                                             c_struct NCPI);

    virtual tBool CONNECT_B3_T90_ACTIVE_IND (c_word   Messagenumber,
                                             c_dword  NCCI);

    virtual tBool DATA_B3_IND               (c_word     Messagenumber,
                                             c_dword    NCCI,
                                             cp_datab3 *Params);

    virtual tBool DISCONNECT_B3_IND         (c_word              Messagenumber,
                                             c_dword             NCCI,
                                             cp_disconnectb3ind *Params);

    virtual tBool DISCONNECT_IND            (c_word  Messagenumber,
                                             c_dword PLCI,
                                             c_info  Reason);

    virtual tBool FACILITY_IND              (c_word       Messagenumber,
                                             c_dword      Address,
                                             cp_facility *Params);

    virtual tBool INFO_IND                  (c_word      Messagenumber,
                                             c_dword     Address,
                                             cp_infoind *Params);

    virtual tBool MANUFACTURER_IND          (c_word           Messagenumber,
                                             c_dword          Controller,
                                             cp_manufacturer *Params);

    virtual tBool RESET_B3_IND              (c_word   Messagenumber,
                                             c_dword  NCCI,
                                             c_struct NCPI);

    /*-----------------------------------------------------------------------*\
        CAPI - Responses
    \*-----------------------------------------------------------------------*/
    void CONNECT_RESP                      (c_word          Messagenumber,
                                            c_dword         PLCI,
                                            cp_connectresp *Params);

    void CONNECT_ACTIVE_RESP               (c_word  Messagenumber,
                                            c_dword PLCI);

    void CONNECT_B3_RESP                   (c_word            Messagenumber,
                                            c_dword           NCCI,
                                            cp_connectb3resp *Params);

    void CONNECT_B3_ACTIVE_RESP            (c_word  Messagenumber,
                                            c_dword NCCI);

    void CONNECT_B3_T90_ACTIVE_RESP        (c_word  Messagenumber,
                                            c_dword NCCI);

    void DATA_B3_RESP                      (c_word  Messagenumber,
                                            c_dword NCCI,
                                            c_word  DataHandle);

    void DISCONNECT_B3_RESP                (c_word  Messagenumber,
                                            c_dword NCCI);

    void DISCONNECT_RESP                   (c_word  Messagenumber,
                                            c_dword PLCI);

    void FACILITY_RESP                     (c_word       Messagenumber,
                                            c_dword      Address,
                                            cp_facility *Params);

    void INFO_RESP                         (c_word  Messagenumber,
                                            c_dword Address);

    void MANUFACTURER_RESP                 (c_word           Messagenumber,
                                            c_dword          Controller,
                                            cp_manufacturer *Params);

    void RESET_B3_RESP                     (c_word  Messagenumber,
                                            c_dword NCCI);

    /*-----------------------------------------------------------------------*\
        Error Handling
    \*-----------------------------------------------------------------------*/
    virtual void HandleMessageError (tUInt type, c_info ErrorCode, c_message cmsg);
    // type: 0 = PutMessage, 1 = GetMessage, 2 = GetProfile
    // cmsg: Nur bei PutMessage != 0 !!

private:
    /*-----------------------------------------------------------------------*\
        Virtuelle Funktionen von CCAPI20_Base
    \*-----------------------------------------------------------------------*/
    void  RegisterComplete (thCapiBase hBase, c_word ApplId, c_info ErrorCode);
    void  ReleaseComplete (void);
    void  PutMessageComplete (c_info ErrorCode);
    tBool GetMessageReceived (c_message ReturnMessage, c_info ErrorCode);
    void  GetProfileComplete (cp_getprofile *pProfile, c_info ErrorCode);

    /*-----------------------------------------------------------------------*\
        Interne Funtkionen
    \*-----------------------------------------------------------------------*/
    void AllocAndPutMsg (c_byte Command, c_byte SubCommand, c_word Messagenumber,
                         c_dword Address, char *format, ...);
    void FillMessage (c_message cmsg, c_byte Command, c_byte SubCommand, c_word Messagenumber,
                      c_dword Address, char *format, va_list param);
    void FreeCurPutMsg (void);
    void EmptyPutMsgbufferQueue (void);
    void HandleGetMessage (c_message ReturnMessage);

    /*-----------------------------------------------------------------------*\
        Interne Klassen und Variablen
    \*-----------------------------------------------------------------------*/
    CDynStoreQueue PutMsgbufferQueue;

    thCapiBase m_hCapiBase;
    c_word     m_Applid;
    c_message  m_CurPutMessage;
    tUInt      m_EmptyPMQInUse;
#ifndef NDEBUG
    tUInt      m_RegisterState;
#endif
};


/*===========================================================================*\
    inline Funktionen der Klasse CCAPI20_MsgBase
\*===========================================================================*/

inline c_word CCAPI20_MsgBase::GetApplID (void) {
    return m_Applid;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void CCAPI20_MsgBase::GET_PROFILE (tUInt Controller) {
    //dassert (m_hCapiBase != vIllegalHandle);
    CapiBase_GetProfile (m_hCapiBase, Controller);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void CCAPI20_MsgBase::GetProfileComplete (cp_getprofile *pProfile, c_info ErrorCode) {
    if (ErrorCode != ci_NoError) {
        //dassert (ErrorCode != ci_Msg_Busy);
        //dassert (ErrorCode != ci_Msg_SendQueueFull);
        HandleMessageError (2, ErrorCode, 0);
    }
    GET_PROFILE_COMPLETE (pProfile, ErrorCode);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void CCAPI20_MsgBase::FreeCurPutMsg (void) {
    c_byte *toDel   = m_CurPutMessage;
    m_CurPutMessage = 0;
    PutMsgbufferQueue.Free (toDel);
}


/*===========================================================================*\
    CCAPI20_MsgBase: Fast alle CAPI - Requests
\*===========================================================================*/

inline void CCAPI20_MsgBase::CONNECT_B3_REQ (c_word Messagenumber, c_dword PLCI, c_struct NCPI) {
    AllocAndPutMsg (CAPI_CONNECT_B3, CAPI_REQ, Messagenumber, PLCI, "4", NCPI);
}

inline void CCAPI20_MsgBase::DATA_B3_REQ (c_word Messagenumber, c_dword NCCI, cp_datab3 *Params) {
    #ifdef C_PLATFORM_64
        AllocAndPutMsg (CAPI_DATA_B3, CAPI_REQ, Messagenumber, NCCI, "32226", 0, Params->DataLength,
                        Params->DataHandle, Params->Flags, Params->Data);
    #else
        AllocAndPutMsg (CAPI_DATA_B3, CAPI_REQ, Messagenumber, NCCI, "3222", Params->Data, Params->DataLength,
                        Params->DataHandle, Params->Flags);
    #endif
}

inline void CCAPI20_MsgBase::DISCONNECT_B3_REQ (c_word Messagenumber, c_dword NCCI, c_struct NCPI) {
    AllocAndPutMsg (CAPI_DISCONNECT_B3, CAPI_REQ, Messagenumber, NCCI, "4", NCPI);
}

inline void CCAPI20_MsgBase::DISCONNECT_REQ (c_word Messagenumber, c_dword PLCI) {
    AllocAndPutMsg (CAPI_DISCONNECT, CAPI_REQ, Messagenumber, PLCI, "1", 0);
}

inline void CCAPI20_MsgBase::FACILITY_REQ (c_word Messagenumber, c_dword Address, cp_facility *Params) {
    AllocAndPutMsg (CAPI_FACILITY, CAPI_REQ, Messagenumber, Address, "24", Params->FacilitySelector,
                    Params->FacilityParam);
}

inline void CCAPI20_MsgBase::INFO_REQ (c_word Messagenumber, c_dword Address, cp_inforeq *Params) {
    c_word AddInfoLen = (c_word)(CalcCapiStructLen (Params->BChannelInformation) +
                                 CalcCapiStructLen (Params->KeypadFacility) +
                                 CalcCapiStructLen (Params->UserUserData) +
                                 CalcCapiStructLen (Params->FacilityDataArray));
    AllocAndPutMsg (CAPI_INFO, CAPI_REQ, Messagenumber, Address, "454444", Params->CalledPartyNumber, AddInfoLen,
                    Params->BChannelInformation, Params->KeypadFacility, Params->UserUserData, Params->FacilityDataArray);
}


inline void CCAPI20_MsgBase::LISTEN_REQ (c_word Messagenumber, c_dword Controller, cp_listenreq *Params) {
    AllocAndPutMsg (CAPI_LISTEN, CAPI_REQ, Messagenumber, Controller, "33341", Params->InfoMask, Params->CIPMask, 0,
                    Params->CallingPartyNumber, 0);
}


inline void CCAPI20_MsgBase::MANUFACTURER_REQ (c_word Messagenumber, c_dword Controller, cp_manufacturer *Params) {
    AllocAndPutMsg (CAPI_MANUFACTURER, CAPI_REQ, Messagenumber, Controller, "3334", Params->ManuID,
                    Params->Class, Params->Function, Params->ManuData);
}


inline void CCAPI20_MsgBase::RESET_B3_REQ (c_word Messagenumber, c_dword NCCI, c_struct NCPI) {
    AllocAndPutMsg (CAPI_RESET_B3, CAPI_REQ, Messagenumber, NCCI, "4", NCPI);
}

inline void CCAPI20_MsgBase::SELECT_B_PROTOCOL_REQ (c_word Messagenumber, c_dword PLCI, cp_bprotocol *Params) {
    c_word BProtLen = (c_word)(6 + CalcCapiStructLen (Params->B1Configuration) +
                                   CalcCapiStructLen (Params->B2Configuration) +
                                   CalcCapiStructLen (Params->B3Configuration));
    AllocAndPutMsg (CAPI_SELECT_B_PROTOCOL, CAPI_REQ, Messagenumber, PLCI, "5222444", BProtLen, Params->B1Protocol,
                    Params->B2Protocol, Params->B3Protocol, Params->B1Configuration, Params->B2Configuration,
                    Params->B3Configuration);
}


/*===========================================================================*\
    CCAPI20_MsgBase: Alle CAPI - Indications
\*===========================================================================*/

inline tBool CCAPI20_MsgBase::CONNECT_IND (c_word Messagenumber, c_dword PLCI, cp_connectind *) {
    CONNECT_RESP (Messagenumber, PLCI, &cpV_ConnectResp_Ignore);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::CONNECT_ACTIVE_IND (c_word Messagenumber, c_dword PLCI, c_struct) {
    CONNECT_ACTIVE_RESP (Messagenumber, PLCI);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::CONNECT_B3_IND (c_word Messagenumber, c_dword NCCI, c_struct) {
    CONNECT_B3_RESP (Messagenumber, NCCI, &cpV_ConnectB3Resp_Reject);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::CONNECT_B3_ACTIVE_IND (c_word Messagenumber, c_dword NCCI, c_struct) {
    CONNECT_B3_ACTIVE_RESP (Messagenumber, NCCI);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::CONNECT_B3_T90_ACTIVE_IND (c_word Messagenumber, c_dword NCCI) {
    CONNECT_B3_T90_ACTIVE_RESP (Messagenumber, NCCI);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::DATA_B3_IND (c_word Messagenumber, c_dword NCCI, cp_datab3 *Params) {
    DATA_B3_RESP (Messagenumber, NCCI, Params->DataHandle);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::DISCONNECT_B3_IND (c_word Messagenumber, c_dword NCCI, cp_disconnectb3ind *) {
    DISCONNECT_B3_RESP (Messagenumber, NCCI);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::DISCONNECT_IND (c_word Messagenumber, c_dword PLCI, c_info) {
    DISCONNECT_RESP (Messagenumber, PLCI);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::FACILITY_IND (c_word Messagenumber, c_dword Address, cp_facility *Params) {
    //cp_facility RespParams = { Params->FacilitySelector, 0 };
    //FACILITY_RESP (Messagenumber, Address, &RespParams);
    FACILITY_RESP (Messagenumber, Address, Params);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::INFO_IND (c_word Messagenumber, c_dword Address, cp_infoind *) {
    INFO_RESP (Messagenumber, Address);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::MANUFACTURER_IND (c_word Messagenumber, c_dword Controller, cp_manufacturer *Params) {
    MANUFACTURER_RESP (Messagenumber, Controller, Params);
    return vTrue;
}

inline tBool CCAPI20_MsgBase::RESET_B3_IND (c_word Messagenumber, c_dword NCCI, c_struct) {
    RESET_B3_RESP (Messagenumber, NCCI);
    return vTrue;
}

/*===========================================================================*\
    CCAPI20_MsgBase: Nearly All CAPI - Responses
\*===========================================================================*/

inline void CCAPI20_MsgBase::CONNECT_ACTIVE_RESP (c_word Messagenumber, c_dword PLCI) {
    AllocAndPutMsg (CAPI_CONNECT_ACTIVE, CAPI_RESP, Messagenumber, PLCI, "");
}

inline void CCAPI20_MsgBase::CONNECT_B3_RESP (c_word Messagenumber, c_dword NCCI, cp_connectb3resp *Params) {
    AllocAndPutMsg (CAPI_CONNECT_B3, CAPI_RESP, Messagenumber, NCCI, "24", Params->Reject, Params->NCPI);
}

inline void CCAPI20_MsgBase::CONNECT_B3_ACTIVE_RESP (c_word Messagenumber, c_dword NCCI) {
    AllocAndPutMsg (CAPI_CONNECT_B3_ACTIVE, CAPI_RESP, Messagenumber, NCCI, "");
}

inline void CCAPI20_MsgBase::CONNECT_B3_T90_ACTIVE_RESP (c_word Messagenumber, c_dword NCCI) {
    AllocAndPutMsg (CAPI_CONNECT_B3_T90_ACTIVE, CAPI_RESP, Messagenumber, NCCI, "");
}

inline void CCAPI20_MsgBase::DATA_B3_RESP (c_word Messagenumber, c_dword NCCI, c_word DataHandle) {
    AllocAndPutMsg (CAPI_DATA_B3, CAPI_RESP, Messagenumber, NCCI, "2", DataHandle);
}

inline void CCAPI20_MsgBase::DISCONNECT_B3_RESP (c_word Messagenumber, c_dword NCCI) {
    AllocAndPutMsg (CAPI_DISCONNECT_B3, CAPI_RESP, Messagenumber, NCCI, "");
}

inline void CCAPI20_MsgBase::DISCONNECT_RESP (c_word Messagenumber, c_dword PLCI) {
    AllocAndPutMsg (CAPI_DISCONNECT, CAPI_RESP, Messagenumber, PLCI, "");
}

inline void CCAPI20_MsgBase::FACILITY_RESP (c_word Messagenumber, c_dword Address, cp_facility *Params) {
    AllocAndPutMsg (CAPI_FACILITY, CAPI_RESP, Messagenumber, Address, "24", Params->FacilitySelector,
                    Params->FacilityParam);
}

inline void CCAPI20_MsgBase::INFO_RESP (c_word Messagenumber, c_dword Address) {
    AllocAndPutMsg (CAPI_INFO, CAPI_RESP, Messagenumber, Address, "");
}

inline void CCAPI20_MsgBase::MANUFACTURER_RESP (c_word Messagenumber, c_dword Controller, cp_manufacturer *Params) {
    AllocAndPutMsg (CAPI_MANUFACTURER, CAPI_RESP, Messagenumber, Controller, "3334", Params->ManuID,
                    Params->Class, Params->Function, Params->ManuData);
}

inline void CCAPI20_MsgBase::RESET_B3_RESP (c_word Messagenumber, c_dword NCCI) {
    AllocAndPutMsg (CAPI_RESET_B3, CAPI_RESP, Messagenumber, NCCI, "");
}


/*===========================================================================*\
\*===========================================================================*/

#endif
