/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "CapiStat.h"
#include "aString.h"
#include "aFlags.h"
#include "atomar.h"
#include "dbgCAPI.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define CAPI_MANU_ID_AVM            0x214D5641
#define CAPI_MANU_FUNCTION_ADSL     0x07

static c_byte      manudata_speed[]  = { 0x02, 0x03, 0x00 };

typedef struct _cp_LineParameters {
    c_word AttainableKBitrateUS;           /* in KBit/s */
    c_word AttainableKBitrateDS;           /* in KBit/s */
    c_word UsedOptionUS;                   /* [-1,3] */
    c_word UsedOptionDS;                   /* [-1,3] */
    c_byte US_FastUsed[4];                 /* bool */
    c_word US_DatabytesPerSymbol[4];       /* dezimal */
    c_byte US_CheckbytesPerCodeword[4];    /* dezimal */
    c_byte US_SymbolsPerCodeword[4];       /* dezimal */
    c_word US_InterleaveDepth[4];          /* dezimal */
    c_byte DS_FastUsed[4];                 /* bool */
    c_word DS_DatabytesPerSymbol[4];       /* dezimal */
    //c_byte DS_CheckbytesPerCodeword[4];    /* dezimal */
    //c_byte DS_SymbolsPerCodeword[4];       /* dezimal */
    //c_word DS_InterleaveDepth[4];          /* dezimal */
} cp_LineParameters;


/*===========================================================================*\
\*===========================================================================*/

CCAPI20_Statistic::CCAPI20_Statistic (ctStatDeep statdeep)
  : m_pCurrentData (0),
    m_pFreshingData (0),
    m_FreshingController (0),
    m_Deep (statdeep),
    m_CountReqCalls (0),
    m_CallBackFunc (0),
    m_CallBackHandle (0),
    m_FreshOnError (vFalse) {

    dhead ("CCAPI20_Statistic-Constructor", DEBUG_CCAPI20_Statistic);
    Protect_Create (&m_hProtect);
}

/*===========================================================================*\
\*===========================================================================*/

CCAPI20_Statistic::~CCAPI20_Statistic (void) {
    dhead ("CCAPI20_Statistic-Destructor", DEBUG_CCAPI20_Statistic);
    dassert (m_pFreshingData == 0);
    Protect_BeginWrite (m_hProtect);
    ctIntCapiData *pDelData = m_pCurrentData;
    m_pCurrentData          = 0;
    Protect_EndWrite (m_hProtect);
    if (pDelData) {
        delete [] pDelData->pControllerData;
        delete pDelData;
    }
    Protect_Destroy (&m_hProtect);
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::FreshUp (tBool FreshUpOnError) {
    dhead ("CCAPI20_Statistic::FreshUp", DEBUG_CCAPI20_Statistic);
    Protect_BeginWrite (m_hProtect);
    if (!m_pFreshingData) {
        m_pFreshingData  = new ctIntCapiData;
        Protect_EndWrite (m_hProtect);
        dassert (m_pFreshingData != 0);   // this case not handled at the moment
        dassert (m_CountReqCalls == 0);
        if (m_pFreshingData) {
            m_pFreshingData->CapiInstalled    = ci_Reg_NotInstalled;
            m_pFreshingData->CountControllers = 0;
            m_pFreshingData->pControllerData  = 0;
            m_FreshingController              = 0;
            m_CountReqCalls                   = 0;
            m_FreshOnError                    = FreshUpOnError;
            REGISTER (1, 1, CAPI_MAX_MESSAGE_SIZE);
        } else {
            FreshUpComplete (ci_Int_MemoryFullErr);
        }
    } else {
        Protect_EndWrite (m_hProtect);
    }
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::FreshUpComplete (c_info Info) {
    dhead ("CCAPI20_Statistic::FreshUpComplete", DEBUG_CCAPI20_Statistic);
    if (m_CallBackFunc) {
        m_CallBackFunc (m_CallBackHandle, Info);
    }
}


/*===========================================================================*\
\*===========================================================================*/

c_info CCAPI20_Statistic::CapiInstalled (void) {
    dhead ("CCAPI20_Statistic::CapiInstalled", DEBUG_CCAPI20_Statistic);
    c_info fret = ci_Reg_Busy;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData) {
        fret = m_pCurrentData->CapiInstalled;
    }
    Protect_EndRead (m_hProtect);
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt CCAPI20_Statistic::CountControllers (void) {
    dhead ("CCAPI20_Statistic::CountControllers", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CountCntrl);
    tUInt fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData) {
        fret = m_pCurrentData->CountControllers;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

tUInt CCAPI20_Statistic::CountBChannels (tUInt Controller) {
    dhead ("CCAPI20_Statistic::CountBChannels", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CntrlProps);
    tUInt fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->CountControllers) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        fret = m_pCurrentData->pControllerData[Controller - 1].NumBChannels;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

c_dword CCAPI20_Statistic::Profile_GlobalOptions (tUInt Controller) {
    dhead ("CCAPI20_Statistic::Profile_GlobalOptions", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CntrlProps);
    c_dword fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        fret = m_pCurrentData->pControllerData[Controller - 1].GlobalOptions;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

c_dword CCAPI20_Statistic::Profile_B1Protocol (tUInt Controller) {
    dhead ("CCAPI20_Statistic::Profile_B1Protocol", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CntrlProps);
    c_dword fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        fret = m_pCurrentData->pControllerData[Controller - 1].B1ProtocolSupport;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

c_dword CCAPI20_Statistic::Profile_B2Protocol (tUInt Controller) {
    dhead ("CCAPI20_Statistic::Profile_B2Protocol", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CntrlProps);
    c_dword fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        fret = m_pCurrentData->pControllerData[Controller - 1].B2ProtocolSupport;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

c_dword CCAPI20_Statistic::Profile_B3Protocol (tUInt Controller) {
    dhead ("CCAPI20_Statistic::Profile_B3Protocol", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CntrlProps);
    c_dword fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        fret = m_pCurrentData->pControllerData[Controller - 1].B3ProtocolSupport;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

c_dword CCAPI20_Statistic::SupplServices (tUInt Controller) {
    dhead ("CCAPI20_Statistic::SupplServices", DEBUG_CCAPI20_Statistic);
    dassert (m_Deep >= StatDeep_CntrlFacilities);
    c_dword fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        fret = m_pCurrentData->pControllerData[Controller - 1].SupplServices;
    }
    Protect_EndRead (m_hProtect);
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/

tUInt CCAPI20_Statistic::DSL_FindController (tUInt Controller) {
    dhead ("CCAPI20_Statistic::DSL_FindController", DEBUG_CCAPI20_Statistic);
    tUInt fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller <= m_pCurrentData->CountControllers);
        for (; Controller < m_pCurrentData->CountControllers; Controller++) {
            if (CheckOneFlag (m_pCurrentData->pControllerData[Controller].B1ProtocolSupport, b1prot_ATM) == vTrue) {
                fret = Controller + 1;
                break;
            }
        }
    }
    Protect_EndRead (m_hProtect);
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt CCAPI20_Statistic::DSL_GetModemState (tUInt Controller) {
    dhead ("CCAPI20_Statistic::DSL_GetModemState", DEBUG_CCAPI20_Statistic);
    tUInt fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        if (CheckOneFlag (m_pCurrentData->pControllerData[Controller - 1].B1ProtocolSupport, b1prot_ATM) == vTrue) {
            fret = m_pCurrentData->pControllerData[Controller - 1].DSL_ModemState;
        }
    }
    Protect_EndRead (m_hProtect);
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt CCAPI20_Statistic::DSL_GetSpeed (tUInt Controller) {
    dhead ("CCAPI20_Statistic::DSL_GetSpeed", DEBUG_CCAPI20_Statistic);
    tUInt fret = 0;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        if (CheckOneFlag (m_pCurrentData->pControllerData[Controller - 1].B1ProtocolSupport, b1prot_ATM) == vTrue) {
            fret = m_pCurrentData->pControllerData[Controller - 1].DSL_Speed;
            // Korregiere DSL_Speed (s. MANUFACTURER_IND)
            if (fret) {
                fret--;
            }
        }
    }
    Protect_EndRead (m_hProtect);
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCAPI20_Statistic::DSL_GetMACAddress (tUInt Controller, tUByte *pAddress) {
    dhead ("CCAPI20_Statistic::DSL_GetMACAddress", DEBUG_CCAPI20_Statistic);
    tBool fret = vFalse;
    Protect_BeginRead (m_hProtect);
    if (m_pCurrentData && m_pCurrentData->pControllerData) {
        dassert (Controller > 0);
        dassert (Controller <= m_pCurrentData->CountControllers);
        if (  (CheckOneFlag (m_pCurrentData->pControllerData[Controller - 1].B1ProtocolSupport, b1prot_ATM) == vTrue)
           && (*(tUInt32 *)m_pCurrentData->pControllerData[Controller - 1].DSL_MACAddress != 0)) {
            // seems to be a valid MAC Address
            s_memcpy (pAddress, m_pCurrentData->pControllerData[Controller - 1].DSL_MACAddress, 6);
            fret = vTrue;
        }
    }
    Protect_EndRead (m_hProtect);
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::REGISTER_COMPLETE (c_info ErrorCode) {
    dhead ("CCAPI20_Statistic::REGISTER_COMPLETE", DEBUG_CCAPI20_Statistic);
    dparams ("%x", ErrorCode);
    dassert (m_pFreshingData != 0);
    dassert (m_FreshingController == 0);
    m_pFreshingData->CapiInstalled = ErrorCode;
    if (ErrorCode == ci_NoError) {
        if (m_Deep > StatDeep_OnlyInstall) {
            GET_PROFILE (1);
        } else {
            RELEASE();
        }
    } else {
        UpdateCurrent (ErrorCode);
    }
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::RELEASE_COMPLETE (void) {
    dhead ("CCAPI20_Statistic::RELEASE_COMPLETE", DEBUG_CCAPI20_Statistic);
    dassert (m_pFreshingData != 0);
    UpdateCurrent (ci_NoError);
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::GET_PROFILE_COMPLETE (cp_getprofile *pProfile, c_info ErrorCode) {
    dhead ("CCAPI20_Statistic::GET_PROFILE_COMPLETE", DEBUG_CCAPI20_Statistic);
    dparams ("%p,%x", pProfile, ErrorCode);
    dassert (m_pFreshingData != 0);
    if (ErrorCode != ci_NoError) {
        //TODO?: Return error error code in Complete?
        RELEASE();
        return;
    }
    dassert (pProfile != 0);
    //dprintBuf (pProfile, sizeof (*pProfile));

    // Is this the first get profile call? Then check the numbers of controllers
    if (!m_pFreshingData->pControllerData) {
        dprint ("NumContrl=%x ", pProfile->NumControllers);
        m_pFreshingData->CountControllers = pProfile->NumControllers;
        if ((m_Deep <= StatDeep_CountCntrl) || (!pProfile->NumControllers)) {
            RELEASE();
            return;
        }
        m_pFreshingData->pControllerData = new ctIntCntrlData [pProfile->NumControllers];
        if (!m_pFreshingData->pControllerData) {
            delete m_pFreshingData;
            m_pFreshingData = 0;
            //TODO: FreshUpComplete (ci_Int_MemoryFullErr);
            dwarning (0);
            RELEASE();
            return;
        }
        s_memzero (m_pFreshingData->pControllerData, sizeof (ctIntCntrlData) * pProfile->NumControllers);
    }
    m_pFreshingData->pControllerData[m_FreshingController].NumBChannels = pProfile->NumBChannels;
    s_memcpy (&m_pFreshingData->pControllerData[m_FreshingController].GlobalOptions,
              &pProfile->GlobalOptions, 4 * sizeof (c_dword));

    if (++m_FreshingController < m_pFreshingData->CountControllers) {
        GET_PROFILE (m_FreshingController + 1);
    } else if (m_Deep >= StatDeep_DSLProps) {
        // GetProfile is finished, now start manu- and facility req's
        static c_byte      manudata_listen[] = { 0x02, 0x01, 0x00 };
        static c_byte      manudata_mac[]    = { 0x08, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
        static c_byte      FacilGetSuppServicesStruct[] = { 0x03, 0x00, 0x00, 0x00 };
        static cp_facility FacParams = { facilselect_SupplServices, &FacilGetSuppServicesStruct[0] };
        tUInt i;
        dassert (m_CountReqCalls == 0);
        atomicInc (&m_CountReqCalls);
        dprint ("m_CountReqCalls-Inc=%x ", m_CountReqCalls);
        for (i = 0; i < m_pFreshingData->CountControllers; ) {
            if (CheckOneFlag (m_pFreshingData->pControllerData[i++].B1ProtocolSupport, b1prot_ATM) == vTrue) {
                // Found DSL Controller
                atomicAdd (&m_CountReqCalls, 4); // 3 for the req's and one for the ind / nolisten-req
                DSLManuReq (i, manudata_listen);
                DSLManuReq (i, manudata_speed);
                DSLManuReq (i, manudata_mac);
            }
        }
        if (m_Deep >= StatDeep_CntrlFacilities) {
            for (i = 0; i < m_pFreshingData->CountControllers; ) {
                if (CheckOneFlag (m_pFreshingData->pControllerData[i++].GlobalOptions,
                                  gof_SupplementaryServices) == vTrue) {
                    // Supplementary Services are supported
                    atomicInc (&m_CountReqCalls);
                    FACILITY_REQ (GetNextMsgNr(), i, &FacParams);
                }
            }
        }
        CallFinished();
    } else {
        RELEASE();
    }
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::FACILITY_CONF (c_word, c_dword Address, cp_facilityconf *Params) {
    dhead ("CCAPI20_Statistic::FACILITY_CONF", DEBUG_CCAPI20_Statistic);
    dassert (Address > 0);
    dassert (Params != 0);
    dassert (m_pFreshingData != 0);
    dassert (Address <= m_pFreshingData->CountControllers);
    c_struct Data = 0;
    if (  (Params->Info == ci_NoError)
       && (Params->FacilitySelector == facilselect_SupplServices)
       && (Params->FacilityParam != 0)
       && (GetBeginAndLenOfStructData (Params->FacilityParam, &Data) >= 9)
       && (get16 (Data) == supplserv_GetSupportedServices)) {
        Data += sizeof (c_word);
        if (  (GetBeginAndLenOfStructData (Data, &Data) >= 6)
           && (get16 (Data) == ci_NoError)) {
            m_pFreshingData->pControllerData[Address - 1].SupplServices = get32 (Data + sizeof (c_word));
        }
    }
    CallFinished();
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::MANUFACTURER_CONF (c_word, c_dword Controller, cp_manufacturer *Params) {
    dhead ("CCAPI20_Statistic::MANUFACTURER_CONF", DEBUG_CCAPI20_Statistic);
    dassert (Params != 0);
    dassert (Params->ManuID == CAPI_MANU_ID_AVM);
    dassert (m_pFreshingData != 0);
    dwarning (Params->Class == 0);
    if (Params->Function == CAPI_MANU_FUNCTION_ADSL) {
        c_byte *pData;
        tUInt   len   = GetBeginAndLenOfStructData (Params->ManuData, &pData);
        dwarning (len > 0);
        if (len) {
            switch (*(c_word *)pData) {
            case 0x06:
                dwarning (len >= 8);
                if ((Params->Class == ci_NoError) && (len >= 8)) {
                    s_memcpy (m_pFreshingData->pControllerData[Controller - 1].DSL_MACAddress, pData + 2, 6);
                }
                break;
            case 0x01:
                if (Params->Class != ci_NoError) {
                    dwarning (0);
                    // NO LISTEN -> NO MANU_IND -> m_CountReqCalls is to great
                    dassert (atomicGet (&m_CountReqCalls) > 1);
                    atomicDec (&m_CountReqCalls);
                }
                break;
            case 0x00:
            case 0x03:
                // Nothing to do
                break;
            default:
                dprintWrn ("Unexpected Service=%x ", *(c_word *)pData);
                break;
            }
        }
    }
    CallFinished();
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCAPI20_Statistic::MANUFACTURER_IND (c_word, c_dword Controller, cp_manufacturer *Params) {
    dhead ("CCAPI20_Statistic::MANUFACTURER_IND", DEBUG_CCAPI20_Statistic);
    dassert (Params != 0);
    dassert (Params->ManuData != 0);
    dassert (Params->ManuID == CAPI_MANU_ID_AVM);
    dassert (m_pFreshingData != 0);
    MANUFACTURER_RESP (GetNextMsgNr(), Controller, Params);

    dprint ("Cntrl=%x, Func=%x, Speed=%x, Len=%x ", Controller, Params->Function,
            m_pFreshingData->pControllerData[Controller - 1].DSL_Speed, CalcCapiStructDataLen (Params->ManuData));
    if (  (!m_pFreshingData->pControllerData[Controller - 1].DSL_Speed)
       && (Params->Function == CAPI_MANU_FUNCTION_ADSL)) {
        c_byte *pData;
        if (GetBeginAndLenOfStructData (Params->ManuData, &pData) > 5) {
            m_pFreshingData->pControllerData[Controller - 1].DSL_ModemState = *(c_dword *)pData;
            pData += sizeof (c_dword);
            pData += CalcCapiStructLen (pData);
            cp_LineParameters *pLineParams;
            dinfo (CalcCapiStructDataLen (pData) >= sizeof (cp_LineParameters), "To Short=%x!! ",
                   CalcCapiStructDataLen (pData));
            if (  (GetBeginAndLenOfStructData (pData, (c_byte **)&pLineParams) >= sizeof (cp_LineParameters))
               && (pLineParams->UsedOptionDS < 4)) {
                dassert (pLineParams != 0);
                m_pFreshingData->pControllerData[Controller - 1].DSL_Speed =
                        (pLineParams->DS_DatabytesPerSymbol[pLineParams->UsedOptionDS] * 48 / 53 * 32) * 1000 + 1;
                // Damit CallFinished nicht zu oft aufgerufen wird, nutzte ich DSL_Speed als Testmerkmal.
                // Dafr mu obige Zuweisung immer != 0 sein. Deshalb wird vorsorglich eine 1 addiert,
                // die spter wieder abgezogen wird.
                dprint ("Controller=%x,DSL_Speed=%x ", Controller,
                        m_pFreshingData->pControllerData[Controller - 1].DSL_Speed);

                // stop dsl listen
                static c_byte manudata_nolisten[] = { 0x02, 0x00, 0x00 };
                DSLManuReq (Controller, manudata_nolisten);
                return vTrue;
            }
        }
        dwarning (0);
        atomicInc (&m_CountReqCalls);
        DSLManuReq (Controller, manudata_speed);
    }
    return vTrue;
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::DSLManuReq (tUInt Controller, c_struct ManuData) {
    dhead ("CCAPI20_Statistic::DSLManuReq", DEBUG_CCAPI20_Statistic);
    dassert (Controller > IllegalController);
    dassert (ManuData != 0);
    cp_manufacturer pm = { CAPI_MANU_ID_AVM, 0, CAPI_MANU_FUNCTION_ADSL, ManuData };
    MANUFACTURER_REQ (GetNextMsgNr(), Controller, &pm);
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::CallFinished (void) {
    dhead ("CCAPI20_Statistic::CallFinished", DEBUG_CCAPI20_Statistic);
    dparams ("%x", m_CountReqCalls);
    dassert (m_CountReqCalls > 0);
    dassert (m_pFreshingData != 0);
    if (atomicDec (&m_CountReqCalls) == vTrue) {
        RELEASE();
    }
    dprint ("new_count=%x ", m_CountReqCalls);
}


/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Statistic::UpdateCurrent (c_info Info) {
    dhead ("CCAPI20_Statistic::UpdateCurrent", DEBUG_CCAPI20_Statistic);
    ctIntCapiData *pDelData = 0;
    Protect_BeginWrite (m_hProtect);
    if ((Info != ci_NoError) && (m_FreshOnError == vFalse)) {
        // ignore this m_pFreshingData
        if (m_pFreshingData) {
            pDelData        = m_pFreshingData;
            m_pFreshingData = 0;
        }
    } else if (m_pFreshingData) {
        pDelData        = m_pCurrentData;
        m_pCurrentData  = m_pFreshingData;
        m_pFreshingData = 0;
    }
    Protect_EndWrite (m_hProtect);
    if (pDelData) {
        delete [] pDelData->pControllerData;
        delete pDelData;
    }
    FreshUpComplete (Info);
}


/*===========================================================================*\
\*===========================================================================*/
