/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "Channel.h"
#include "ConvCapi.h"
#include "atomar.h"
#include "dbgCAPI.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define CAPI_CALLEDPARTYNUM_MAXLEN          70

#define DTMF_STATE_N                         0    // N = Nothing
#define DTMF_STATE_B                         1    // B = Listen Begin
#define DTMF_STATE_E                         2    // E = Listen End
#define DTMF_STATE_S                         3    // S = Send
#define DTMF_STATE_BE                        4
#define DTMF_STATE_BS                        5
#define DTMF_STATE_EB                        6
#define DTMF_STATE_ES                        7
#define DTMF_STATE_SE                        8
#define DTMF_STATE_SB                        9
#define DTMF_STATE_BSE                      10
#define DTMF_STATE_ESB                      11

#define CHANNEL_STATUS_RESET                 1
#define CHANNEL_STATUS_DISCONNECT_WAIT_NO    2
#define CHANNEL_STATUS_DISCONNECT_WAIT_WITH  4
#define CHANNEL_STATUS_DISCONNECT_WAIT       6

/*===========================================================================*\
    CTransferChannel - Konstruktor
\*===========================================================================*/

CTransferChannel::CTransferChannel (void)
  : m_SendBlock (CAPI_MAX_OUTSTANDING_FRAMES),
    m_pMSNList (0),
    m_deleteMSNList (vFalse),
    m_WaitForDDINumbers (vFalse),
    m_szCallingNumber (0),
    m_OutController (IllegalController),
    m_ActiveController (IllegalController),
    m_CIPValue (cipvalue_Undefined),
    m_B1Prot (b1prot_HDLC),
    m_B2Prot (b2prot_X75),
    m_B3Prot (b3prot_Transparent),
    m_B1Config (0),
    m_B3Config (0),
    m_sMSN (0),
    m_CalledPartyNum (0),
    m_ConnectedPartyNum (0),
    m_MaxDataSize (CAPI_MAX_DATA_SIZE),
    m_DDIPrefixLen (0),
    m_ParkIdentity (0),
    m_SupplServType (0),
    m_DTMFState (DTMF_STATE_N),
    m_DTMFDigits (0),
    m_DTMFToneDuration (0),
    m_DTMFGapDuration (0),
    m_DTMFListen (vFalse),
    m_Reseting (0),
    m_CheckListenInUse (0) {

    dhead ("CTransferChannel-Constructor", DEBUG_CTransferChannel);
    s_memzero (&m_AdditionalInfo, sizeof (m_AdditionalInfo));
}


/*===========================================================================*\
    CTransferChannel - Destruktor
\*===========================================================================*/

CTransferChannel::~CTransferChannel (void) {
    dhead ("CTransferChannel-Destructor", DEBUG_CTransferChannel);
    dassert (CurState() == cs_Released);
    dassert (m_Reseting == 0);
    if (m_deleteMSNList == vTrue) {
        delete m_pMSNList;
    }
    m_pMSNList         = 0;
    m_deleteMSNList    = vFalse;
    m_OutController    = IllegalController;
    m_ActiveController = IllegalController;
    m_CIPValue         = cipvalue_Undefined;
    m_SupplServType    = 0;
    Conv_DuplicateStruct (&m_B1Config, 0);
    Conv_DuplicateStruct (&m_B3Config, 0);
    Conv_DuplicateStruct (&m_sMSN, 0);
    Conv_DuplicateStruct (&m_CalledPartyNum, 0);
    Conv_DuplicateStruct (&m_ConnectedPartyNum, 0);
    Conv_DuplicateString (&m_szCallingNumber, 0);
    Conv_DuplicateString (&m_ParkIdentity, 0);
    Conv_DuplicateString (&m_DTMFDigits, 0);
    Conv_DuplicateStruct (&m_AdditionalInfo.BChannelInformation, 0);
    Conv_DuplicateStruct (&m_AdditionalInfo.KeypadFacility, 0);
    Conv_DuplicateStruct (&m_AdditionalInfo.UserUserData, 0);
    Conv_DuplicateStruct (&m_AdditionalInfo.FacilityDataArray, 0);
}


/*===========================================================================*\
    CTransferChannel::ResetChannel
\*===========================================================================*/

void CTransferChannel::ResetChannel (void) {
    dhead ("CTransferChannel::ResetChannel", DEBUG_CTransferChannel);
    dparams ("%x,%x", CurState(), m_Reseting);
    dwarning (m_Reseting == 0);
    if ((atomicGet (&m_CheckListenInUse) == 0) && (CurState() == cs_Released)) {
        IsReseted();
    } else {
        m_Reseting = CHANNEL_STATUS_RESET;
        SetMSNList (0);
        Disconnect (DisconnectKind_Fast);
        CheckListen();
    }
}


/*===========================================================================*\
    CTransferChannel::CheckListen
\*===========================================================================*/

void CTransferChannel::CheckListen (void) {
    dhead ("CTransferChannel::CheckListen", DEBUG_CTransferChannel);
    dprint ("state=%x,pMSNList=%p ", GetState(), m_pMSNList);
    if (atomicInc (&m_CheckListenInUse) == vFalse) {
        dprint ("m_CheckListenInUse=%x ", m_CheckListenInUse);
        return;
    }
    do {
        CCntrlMSNList *pTmpMSNList = m_pMSNList;
        switch (GetState()) {
        case cs_Released:
            dprint ("cl-State=cs_Released\n");
            if (pTmpMSNList && (pTmpMSNList->IsEmpty() == vFalse)) {
                Register (m_MaxDataSize);
            }
            break;
        case cs_Registering:
            dprint ("cl-State=cs_Registering\n");
            // nothing to do!
            break;
        case cs_Registered:
            dprint ("cl-State=cs_Registered\n");
            if (!pTmpMSNList || (pTmpMSNList->IsEmpty() == vTrue)) {
                dprint ("Start Release: ");
                Release();
                dprint ("Finished Release! ");
                break;
            }
            // no break!
        default:
            dprint ("cl-State=default\n");
            if (pTmpMSNList) {
                tUInt              newCntrler;
                tUInt              countMSNs;
                cu_PMWakeupNumber  OneWakeupNumber;
                cu_PMWakeupNumber *WakeupList;
                cp_listenreq       newListenP;
                newListenP.CallingPartyNumber = 0;

                dprint ("cl-ResetGetNextMask\n");
                pTmpMSNList->ResetGetNextMask();
                dprint ("cl-MSNList-Loop\n");
                while (pTmpMSNList->GetNextMask (&newCntrler, &newListenP.InfoMask, &newListenP.CIPMask) == vTrue) {
                    dprint ("cl-newCntrler=%x\n", newCntrler);
                    dassert (newCntrler != IllegalController);
                    ListenReq (newCntrler, &newListenP);
                    dprint ("cl-AfterListen\n");
                    if (pTmpMSNList->GetWakeupSupport (newCntrler) == vTrue) {
                        dprint ("cl-WakeUpEnabled\n");
                        countMSNs = pTmpMSNList->GetCountMSN (newCntrler);
                        if (  !countMSNs
                           || CheckBitsAll (newListenP.InfoMask,
                                            (infomask_CalledPartyNum | infomask_SendingComplete))
                           || pTmpMSNList->ExistMSNRange (newCntrler)) {
                            // empty MSNList, MSN Ranges or DDI in use => disable Wakeup
                            fReq_PMWakeup (newCntrler, 0, 0);
                        } else {
                            if (countMSNs > 1) {
                                WakeupList = new cu_PMWakeupNumber [countMSNs];
                                if (WakeupList) {
                                    for (tUInt i = 0; i < countMSNs; i++) {
                                        WakeupList[i].CalledPartyNumber = pTmpMSNList->GetStructMSN (newCntrler, i);
                                        WakeupList[i].CIPMask           = newListenP.CIPMask;
                                    }
                                    dassert (countMSNs <= MAXVAL_tUShort);
                                    fReq_PMWakeup (newCntrler, (tUShort)countMSNs, WakeupList);
                                    delete [] WakeupList;
                                } else {
                                    fReq_PMWakeup (newCntrler, 0, 0);
                                }
                            } else  {
                                OneWakeupNumber.CalledPartyNumber = pTmpMSNList->GetStructMSN (newCntrler, 0);
                                OneWakeupNumber.CIPMask           = newListenP.CIPMask;
                                dassert (OneWakeupNumber.CalledPartyNumber != 0);
                                fReq_PMWakeup (newCntrler, 1, &OneWakeupNumber);
                            }
                        }
                    }
                    dprint ("cl-NextIteration\n");
                }
            }
            break;
        }
        dprint ("Cur-m_CheckListenInUse=%x ", m_CheckListenInUse);
    } while (atomicDec (&m_CheckListenInUse) == vFalse);
    dprint ("Loop terminated! ");
    if (GetState() == cs_Released) {
        dprint ("Calling IsReleased! ");
        Release_Complete();
    }
}


/*===========================================================================*\
    CTransferChannel::MakeCall
\*===========================================================================*/

tBool CTransferChannel::MakeCall (tUChar *pTelNumber, c_cipvalue CipValue) {
    dhead ("CTransferChannel::MakeCall", DEBUG_CTransferChannel);
    dwarning (pTelNumber != 0);
    dwarning (CipValue != cipvalue_Undefined);
    dassert (m_OutController != IllegalController);
    dassert (m_ActiveController == IllegalController);

    if (GetState() > cs_Registered) {
        RETURN ('x', vFalse);
    }
    if (pTelNumber && !s_strempty (pTelNumber)) {
        ConvUnion_CalledNumber (&m_CalledPartyNum, pTelNumber);
        if (!m_CalledPartyNum) {
            // converting failed
            dwarning (0);
            RETURN ('x', vFalse);
        }
    } else {
        ConvUnion_CalledNumber (&m_CalledPartyNum, 0);
        dassert (m_CalledPartyNum == 0);
    }
    m_ActiveController = m_OutController;
    m_CIPValue         = CipValue;
    m_SupplServType    = 0;
    dprint ("state=%x ", GetState());
    switch (GetState()) {
    case cs_Released:
        Register (m_MaxDataSize);
        break;
    case cs_Registering:
        /*----- Nothing To Do! -----*/
        break;
    default:
        dwarning (0);
        break;
    case cs_Registered:
        {
            DialingBegin();
            cp_connectreq ConParams = { m_CIPValue, m_CalledPartyNum, m_sMSN,
                                        { m_B1Prot, m_B2Prot, m_B3Prot, m_B1Config, 0, m_B3Config },
                                        { 0, 0, 0, 0 } };
            ConParams.ai = m_AdditionalInfo;
            if (ConnectReq (m_ActiveController, &ConParams) == vFalse) {
                m_ActiveController = IllegalController;
                m_CIPValue         = cipvalue_Undefined;
                ConvUnion_CalledNumber (&m_CalledPartyNum, 0);
                RETURN ('x', vFalse);
            }
        }
        break;
    }
    RETURN ('x', vTrue);
}


/*===========================================================================*\
    CTransferChannel::AnswerCall
\*===========================================================================*/

void CTransferChannel::AnswerCall (c_reject Reject) {
    dhead ("CTransferChannel::AnswerCall", DEBUG_CTransferChannel);
    dparams ("%x", Reject);
    cp_connectresp ConParams = { Reject, m_ConnectedPartyNum,
                                 { m_B1Prot, m_B2Prot, m_B3Prot, m_B1Config, 0, m_B3Config },
                                 { 0, 0, 0, 0 } };
    ConParams.ai = m_AdditionalInfo;
    ConnectResp (&ConParams);
}


/*===========================================================================*\
    CTransferChannel::PutData
\*===========================================================================*/

tBool CTransferChannel::PutData (tUByte *Data, tUShort DataLength, void *hDataID) {
    dhead ("CTransferChannel::PutData", DEBUG_CTransferChannel);
    dparams ("d=%p,l=%x,h=%x", Data, DataLength, hDataID);
    if ((m_Reseting & (CHANNEL_STATUS_DISCONNECT_WAIT_NO | CHANNEL_STATUS_RESET)) != 0) {
        dwarning (0);
        RETURN ('x',  vFalse);
    }
    tUShort CapiHandle;
    if (m_SendBlock.New (hDataID, &CapiHandle) == vFalse) {
        //PutDataComplete (hDataID, i_QueueFull);
        RETURN ('x', vFalse);
    }
    dassert (CapiHandle <= MAXVAL_tUShort);
    cp_datab3 DataParams = {Data, DataLength, CapiHandle, 0};
    DataB3Req (&DataParams);
    RETURN ('x', vTrue);
}


/*===========================================================================*\
    CTransferChannel::UpdateBParameters
\*===========================================================================*/

void CTransferChannel::UpdateBParameters (void) {
    dhead ("CTransferChannel::UpdateBParameters", DEBUG_CTransferChannel);
    cp_bprotocol bprot = {m_B1Prot, m_B2Prot, m_B3Prot, m_B1Config, 0, m_B3Config};
    SelectBProtocolReq (&bprot);
}


/*===========================================================================*\
    CTransferChannel::ManufacturerCall
\*===========================================================================*/

void CTransferChannel::ManufacturerCall (c_dword Function, c_struct Data) {
    dhead ("CTransferChannel::ManufacturerCall", DEBUG_CTransferChannel);
    dassert (m_OutController != IllegalController);
    // workaround to allow Register and Release for a Manufacturer call
    switch (Function) {
    case 0xFFFF0001:
        dwarning (GetState() == cs_Released);
        if (GetState() == cs_Released) {
            dassert (m_ActiveController == IllegalController);
            m_SupplServType = Function;
            Register (m_MaxDataSize);
        }
        break;
    default:
        dassert (GetState() >= cs_Registered);
        cp_manufacturer Params = { 0x214D5641, 0, Function, Data };
        ManufacturerReq (m_OutController, &Params);
        break;
    }
}


/*===========================================================================*\
    CTransferChannel::HoldCall
\*===========================================================================*/
#ifdef USE_SUPPLEMENTARY_SERVICES

void CTransferChannel::HoldCall (void) {
    dhead ("CTransferChannel::HoldCall", DEBUG_CTransferChannel);
    if (GetState() > cs_D_Connected) {
        m_SupplServType = supplserv_Hold;
        DisconnectReq();
    } else {
        fReq_Hold();
    }
}


/*===========================================================================*\
    CTransferChannel::UnholdCall
\*===========================================================================*/

void CTransferChannel::UnholdCall (void) {
    dhead ("CTransferChannel::UnholdCall", DEBUG_CTransferChannel);
    fReq_Retrieve();
}


/*===========================================================================*\
    CTransferChannel::ParkCall
\*===========================================================================*/

void CTransferChannel::ParkCall (tUChar *pIDStr, tUInt len) {
    dhead ("CTransferChannel::ParkCall", DEBUG_CTransferChannel);
    if (GetState() < cs_D_Connected) {
        CallParked (ci_Int_IllegalState);
    } else if (ConvUnion_String (&m_ParkIdentity, pIDStr, len) == 0) {
        dassert (m_ParkIdentity == 0);
        CallParked (ci_Int_MemoryFullErr);
    } else if (GetState() > cs_D_Connected) {
        m_SupplServType = supplserv_Suspend;
        DisconnectReq();
    } else {
        dassert (m_ParkIdentity != 0);
        fReq_Suspend (m_ParkIdentity);
        Conv_DuplicateStruct (&m_ParkIdentity, 0);
    }
}


/*===========================================================================*\
    CTransferChannel::UnparkCall
\*===========================================================================*/

void CTransferChannel::UnparkCall (tUChar *pIDStr, tUInt len) {
    dhead ("CTransferChannel::UnparkCall", DEBUG_CTransferChannel);
    dassert (m_OutController != IllegalController);
    dassert (m_ActiveController == IllegalController);
    if (GetState() > cs_Registered) {
        CallUnparked (ci_Int_IllegalState);
    } else if (ConvUnion_String (&m_ParkIdentity, pIDStr, len) == 0) {
        dassert (m_ParkIdentity == 0);
        CallUnparked (ci_Int_MemoryFullErr);
    } else {
        switch (GetState()) {
        case cs_Released:
        //case cs_Releasing:
            Register (m_MaxDataSize);
            // no break!
        case cs_Registering:
            m_ActiveController = m_OutController;
            m_CIPValue         = cipvalue_Undefined;
            m_SupplServType    = supplserv_Resume;
            break;
        default:
            dassert (0);
            // no break!
        case cs_Registered:
            m_ActiveController = m_OutController;
            m_CIPValue         = cipvalue_Undefined;
            fReq_Resume (m_ActiveController, m_ParkIdentity);
            Conv_DuplicateStruct (&m_ParkIdentity, 0);
            break;
        }
    }
}


/*===========================================================================*\
    CTransferChannel::RedirectCall
\*===========================================================================*/

void CTransferChannel::RedirectCall (tUChar *pTelNumber) {
    dhead ("CTransferChannel::RedirectCall", DEBUG_CTransferChannel);
    dassert (pTelNumber != 0);
    c_struct fanum = 0;
    if (ConvUnion_FacilityNumber (&fanum, pTelNumber) > 0) {
        dassert (fanum != 0);
        fReq_CD (vTrue, fanum);
        Conv_DuplicateStruct (&fanum, 0);
    } else {
        dassert (fanum == 0);
        CallRedirected (ci_Int_MemoryFullErr);
    }
}


/*===========================================================================*\
    CTransferChannel::TransferCall
\*===========================================================================*/

void CTransferChannel::TransferCall (CTransferChannel *pChannel) {
    dhead ("CTransferChannel::TransferCall", DEBUG_CTransferChannel);
    dassert (pChannel != 0);
    dassert (GetState() == cs_Holded);
    dassert (pChannel->GetState() == cs_Connected);
    fReq_ECT (pChannel);
}


/*===========================================================================*\
    CTransferChannel::Conference
\*===========================================================================*/

void CTransferChannel::Conference (CTransferChannel *pChannel, tBool doBegin) {
    dhead ("CTransferChannel::Conference", DEBUG_CTransferChannel);
    dparams ("%x", doBegin);
    dassert (pChannel != 0);
    if (doBegin == vTrue) {
        fReq_3PTY_Begin (pChannel);
    } else {
        fReq_3PTY_End (pChannel);
    }
}


/*===========================================================================*\
    CTransferChannel::Forward
\*===========================================================================*/

void CTransferChannel::Forward (c_dword /*Controller*/, c_forwardtype /*Type*/, tUChar * /*MSN*/, tUChar * /*ForwardNumber*/) {
    dhead ("CTransferChannel::Forward", DEBUG_CTransferChannel);
    //dassert (Controller != IllegalController);
}


/*===========================================================================*\
    CTransferChannel::DTMFListen
\*===========================================================================*/

void CTransferChannel::DTMFListen (tBool activate) {
    dhead ("CTransferChannel::DTMFListen", DEBUG_CTransferChannel);
    dparams ("%x", activate);
    dwarning (GetState() == cs_Connected);
    if (activate == m_DTMFListen) {
        return;
    }
    m_DTMFListen = activate;
    if (GetState() != cs_Connected) {
        return;
    }
    if (activate == vTrue) {
        switch (m_DTMFState) {
        case DTMF_STATE_N:
            m_DTMFState = DTMF_STATE_B;
            fReq_DTMF (dtmffunction_StartListen, 0, 0, 0, 0);
            break;
        case DTMF_STATE_B:
            // nothing to do!
            break;
        case DTMF_STATE_E:
            m_DTMFState = DTMF_STATE_BE;
            break;
        case DTMF_STATE_S:
            m_DTMFState = DTMF_STATE_SB;
            break;
        case DTMF_STATE_BE:
            m_DTMFState = DTMF_STATE_B;
            break;
        case DTMF_STATE_BS:
            m_DTMFState = DTMF_STATE_B;
            break;
        case DTMF_STATE_EB:
            // nothing to do!
            break;
        case DTMF_STATE_ES:
            m_DTMFState = DTMF_STATE_ESB;
            break;
        case DTMF_STATE_SE:
            m_DTMFState = DTMF_STATE_SB;
            break;
        case DTMF_STATE_SB:
            // nothing to do!
            break;
        case DTMF_STATE_BSE:
            m_DTMFState = DTMF_STATE_BS;
            break;
        default:
            dassert (0);
            break;
        }
    } else {
        switch (m_DTMFState) {
        case DTMF_STATE_N:
            m_DTMFState = DTMF_STATE_E;
            fReq_DTMF (dtmffunction_StopListen, 0, 0, 0, 0);
            break;
        case DTMF_STATE_B:
            m_DTMFState = DTMF_STATE_BE;
            break;
        case DTMF_STATE_E:
            // nothing to do!
            break;
        case DTMF_STATE_S:
            m_DTMFState = DTMF_STATE_SE;
            break;
        case DTMF_STATE_BE:
            // nothing to do!
            break;
        case DTMF_STATE_BS:
            m_DTMFState = DTMF_STATE_BSE;
            break;
        case DTMF_STATE_EB:
            m_DTMFState = DTMF_STATE_E;
            break;
        case DTMF_STATE_ES:
            // nothing to do!
            break;
        case DTMF_STATE_SE:
            // nothing to do!
            break;
        case DTMF_STATE_SB:
            m_DTMFState = DTMF_STATE_SE;
            break;
            break;
        case DTMF_STATE_BSE:
            // nothing to do!
            break;
        default:
            dassert (0);
            break;
        }
    }
}


/*===========================================================================*\
    CTransferChannel::DTMFSend
\*===========================================================================*/

void CTransferChannel::DTMFSend (tUInt count, tUChar *pDTMFDigits, tUShort ToneDuration, tUShort GapDuration) {
    dhead ("CTransferChannel::DTMFSend", DEBUG_CTransferChannel);
    dassert (count > 0);
    dassert (pDTMFDigits != 0);
    dassert (GetState() == cs_Connected);
    dwarning (m_DTMFDigits == 0);
    if (m_DTMFDigits) {
        DTMFSendComplete (ci_Int_FunctionInUse);
    } else if (ConvUnion_String (&m_DTMFDigits, pDTMFDigits, count) > 0) {
        dassert (m_DTMFDigits != 0);
        m_DTMFToneDuration = ToneDuration;
        m_DTMFGapDuration  = GapDuration;
        switch (m_DTMFState) {
        case DTMF_STATE_N:
            m_DTMFState = DTMF_STATE_S;
            fReq_DTMF (dtmffunction_SendDigits, m_DTMFToneDuration, m_DTMFGapDuration, m_DTMFDigits, 0);
            Conv_DuplicateStruct (&m_DTMFDigits, 0);
            break;
        case DTMF_STATE_B:
            m_DTMFState = DTMF_STATE_BS;
            break;
        case DTMF_STATE_E:
            m_DTMFState = DTMF_STATE_ES;
            break;
        case DTMF_STATE_BE:
            m_DTMFState = DTMF_STATE_BSE;
            break;
        case DTMF_STATE_EB:
            m_DTMFState = DTMF_STATE_ESB;
            break;
        default:
        case DTMF_STATE_BS:
        case DTMF_STATE_ES:
        case DTMF_STATE_BSE:
        case DTMF_STATE_ESB:
            dassert (0);
            // no break!
        case DTMF_STATE_S:
        case DTMF_STATE_SE:
        case DTMF_STATE_SB:
            dwarning (0);
            Conv_DuplicateStruct (&m_DTMFDigits, 0);
            DTMFSendComplete (ci_Int_FunctionInUse);
            break;
        }
    } else {
        DTMFSendComplete (ci_Int_MemoryFullErr);
    }
}

/*===========================================================================*\
    CTransferChannel::fConf_DTMF
\*===========================================================================*/

void CTransferChannel::fConf_DTMF (c_info Info) {
    dhead ("CTransferChannel::fConf_DTMF", DEBUG_CTransferChannel);
    dprint ("%x", Info);
    tUInt now  = DTMF_STATE_N;
    tUInt next = DTMF_STATE_N;
    switch (m_DTMFState) {
    case DTMF_STATE_N:
        dwarning (0);
        break;
    case DTMF_STATE_B:
        m_DTMFState = DTMF_STATE_N;
        now         = DTMF_STATE_B;
        break;
    case DTMF_STATE_E:
        m_DTMFState = DTMF_STATE_N;
        now         = DTMF_STATE_E;
        break;
    case DTMF_STATE_S:
        m_DTMFState = DTMF_STATE_N;
        now         = DTMF_STATE_S;
        break;
    case DTMF_STATE_BE:
        m_DTMFState = DTMF_STATE_E;
        next        = DTMF_STATE_E;
        now         = DTMF_STATE_B;
        break;
    case DTMF_STATE_BS:
        m_DTMFState = DTMF_STATE_S;
        next        = DTMF_STATE_S;
        now         = DTMF_STATE_B;
        break;
    case DTMF_STATE_ES:
        m_DTMFState = DTMF_STATE_S;
        next        = DTMF_STATE_S;
        now         = DTMF_STATE_E;
        break;
    case DTMF_STATE_SB:
        m_DTMFState = DTMF_STATE_B;
        next        = DTMF_STATE_B;
        now         = DTMF_STATE_S;
        break;
    case DTMF_STATE_SE:
        m_DTMFState = DTMF_STATE_E;
        next        = DTMF_STATE_E;
        now         = DTMF_STATE_S;
        break;
    case DTMF_STATE_EB:
        m_DTMFState = DTMF_STATE_B;
        next        = DTMF_STATE_B;
        now         = DTMF_STATE_E;
        break;
    case DTMF_STATE_BSE:
        m_DTMFState = DTMF_STATE_SE;
        next        = DTMF_STATE_S;
        now         = DTMF_STATE_B;
        break;
    case DTMF_STATE_ESB:
        m_DTMFState = DTMF_STATE_SB;
        next        = DTMF_STATE_S;
        now         = DTMF_STATE_E;
        break;
    default:
        dassert (0);
        break;
    }
    switch (now) {
    case DTMF_STATE_N:
        break;
    case DTMF_STATE_B:
        if (CapiInfoIsError (Info)) {
            m_DTMFListen = vFalse;
            DTMFListenError (Info);
        }
        break;
    case DTMF_STATE_E:
        break;
    case DTMF_STATE_S:
        DTMFSendComplete (Info);
        break;
    default:
        dassert (0);
        break;
    }
    switch (next) {
    case DTMF_STATE_N:
        break;
    case DTMF_STATE_B:
        fReq_DTMF (dtmffunction_StartListen, 0, 0, 0, 0);
        break;
    case DTMF_STATE_E:
        fReq_DTMF (dtmffunction_StopListen, 0, 0, 0, 0);
        break;
    case DTMF_STATE_S:
        dwarning (m_DTMFDigits != 0);
        fReq_DTMF (dtmffunction_SendDigits, m_DTMFToneDuration, m_DTMFGapDuration, m_DTMFDigits, 0);
        Conv_DuplicateStruct (&m_DTMFDigits, 0);
        break;
    default:
        dassert (0);
        break;
    }
}


#endif
/*===========================================================================*\
    CTransferChannel::Register_Complete
\*===========================================================================*/

void CTransferChannel::Register_Complete (c_info ErrorCode) {
    dhead ("CTransferChannel::Register_Complete", DEBUG_CTransferChannel);
    dwarning (ErrorCode == ci_NoError);
    if (CapiInfoIsSuccess (ErrorCode)) {
        if (m_ActiveController != IllegalController) {
            DialingBegin();
            #ifdef USE_SUPPLEMENTARY_SERVICES
                if (m_SupplServType == supplserv_Resume) {
                    dassert (m_ParkIdentity != 0);
                    fReq_Resume (m_ActiveController, m_ParkIdentity);
                } else
            #endif
            {
                cp_connectreq ConParams = { m_CIPValue, m_CalledPartyNum, m_sMSN,
                                            { m_B1Prot, m_B2Prot, m_B3Prot, m_B1Config, 0, m_B3Config },
                                            { 0, 0, 0, 0 } };
                ConParams.ai = m_AdditionalInfo;
                if (ConnectReq (m_ActiveController, &ConParams) == vFalse) {
                    dassert (0);
                }
            }
        } else if (m_SupplServType == 0xFFFF0001) {
            m_SupplServType = 0;
            ManufacturerCallComplete (ErrorCode, 0xFFFF0001, 0);
            return;
        }
        m_SupplServType = 0;
        Conv_DuplicateStruct (&m_ParkIdentity, 0);
        CheckListen();
    } else {
        if (m_ActiveController != IllegalController) {
            m_ActiveController = IllegalController;
            m_CIPValue         = cipvalue_Undefined;
            ConvUnion_CalledNumber (&m_CalledPartyNum, 0);
            IsDisconnected (ErrorCode);
        }
        if (m_SupplServType == 0xFFFF0001) {
            m_SupplServType = 0;
            ManufacturerCallComplete (ErrorCode, 0xFFFF0001, 0);
        }
        if (m_pMSNList && (m_pMSNList->IsEmpty() == vFalse)) {
            ListenError (0, ErrorCode);
        }
        if (atomicGet (&m_CheckListenInUse) == 0) {
            tUInt doIt = m_Reseting;
            m_Reseting = 0;
            if (doIt) {
                IsReseted();
            }
        }
    }
}


/*===========================================================================*\
    CTransferChannel::Release_Complete
\*===========================================================================*/

void CTransferChannel::Release_Complete (void) {
    dhead ("CTransferChannel::Release_Complete", DEBUG_CTransferChannel);
    dassert (m_ActiveController == IllegalController);
    if (  (m_ActiveController != IllegalController)
       || (m_pMSNList && (m_pMSNList->IsEmpty() == vFalse))) {
        Register (m_MaxDataSize);
    } else if (atomicGet (&m_CheckListenInUse) == 0) {
        tUInt doIt = m_Reseting;
        m_Reseting = 0;
        if (doIt) {
            IsReseted();
        }
    }
}


/*===========================================================================*\
    CTransferChannel::DataB3Conf
\*===========================================================================*/

void CTransferChannel::DataB3Conf (cp_datab3conf *Params) {
    dhead ("CTransferChannel::DataB3Conf", DEBUG_CTransferChannel);
    dassert (Params != 0);
    void *hDataID;
    if (m_SendBlock.Free (Params->DataHandle, &hDataID) == vTrue) {
        PutDataComplete (hDataID, Params->Info);
        if ((m_Reseting & CHANNEL_STATUS_DISCONNECT_WAIT) && (IsPutQueueEmpty() == vTrue)) {
            Disconnect();
        }
    } else {
        dprintErr ("Great CAPI Error (Now disconnecting)!!!\n");
        Disconnect();
        dassert (0);
    }
}


/*===========================================================================*\
    CTransferChannel::ListenConf
\*===========================================================================*/

void CTransferChannel::ListenConf (c_dword Controller, c_info Info) {
    dhead ("CTransferChannel::ListenConf", DEBUG_CTransferChannel);
    dparams ("c=%x,i=%x", Controller, Info);
    if (CapiInfoIsError (Info) == vTrue) {
        //if (Info == ci_Cod_IllegalAddress) {
        //    // illegal controller-number can be deleted
        //    m_pMSNList.Remove (Controller);
        //    CheckListen();
        //}
        ListenError (Controller, Info);
    }
}


/*===========================================================================*\
    CTransferChannel::ManufacturerConf
\*===========================================================================*/

void CTransferChannel::ManufacturerConf (c_dword, cp_manufacturer *Params) {
    dhead ("CTransferChannel::ManufacturerConf", DEBUG_CTransferChannel);
    ManufacturerCallComplete (Params->Class, Params->Function, Params->ManuData);
}


/*===========================================================================*\
    CTransferChannel::SelectBProtocolConf
\*===========================================================================*/

void CTransferChannel::SelectBProtocolConf (c_info Info) {
    dhead ("CTransferChannel::SelectBProtocolConf", DEBUG_CTransferChannel);
    UpdateBParametersComplete (Info);
}


/*===========================================================================*\
    CTransferChannel::ConnectInd
\*===========================================================================*/

void CTransferChannel::ConnectInd (c_dword Controller, cp_connectind *Params) {
    dhead ("CTransferChannel::ConnectInd", DEBUG_CTransferChannel);
    if (m_pMSNList != 0) {
        dassert (m_ActiveController == IllegalController);
        switch (m_pMSNList->FindCallInfo (Controller, Params->CIPValue, Params->CalledPartyNumber, &m_DDIPrefixLen)) {
        case iWrn_MSNList_DDICall:
            dassert (CAPI_CALLEDPARTYNUM_MAXLEN < 254);
            dassert (m_DDIPrefixLen < CAPI_CALLEDPARTYNUM_MAXLEN);
            dassert (m_DDIPrefixLen > 0);
            Conv_DuplicateStruct (&m_CalledPartyNum, 0);
            m_CalledPartyNum = new c_byte [CAPI_CALLEDPARTYNUM_MAXLEN + 3];  // 1 for len, 1 for 0x80, 1 for '\0'
            if (m_CalledPartyNum) {
                s_memset (m_CalledPartyNum, '\0', CAPI_CALLEDPARTYNUM_MAXLEN + 3);
                m_CalledPartyNum[0] = 1;
                m_CalledPartyNum[1] = 0x80;
                ConvStruct_CallingNumber (&m_szCallingNumber, Params->CallingPartyNumber);
                m_ActiveController  = Controller;
                m_CIPValue          = Params->CIPValue;
                return;
            }
            dwarning (0);
            m_DDIPrefixLen = 0;
            break;
        case i_Okay:
            if (AcceptIncomingCall() == vTrue) {
                m_ActiveController = Controller;
                m_CIPValue         = Params->CIPValue;
                tUChar *szN        = 0;
                ConvStruct_CalledNumber (&szN, Params->CalledPartyNumber);
                ConvStruct_CallingNumber (&m_szCallingNumber, Params->CallingPartyNumber);
                ConvUnion_CallingNumber (&m_ConnectedPartyNum, szN);
                dprint ("(%s)(%s)\n", m_szCallingNumber, szN);
                IncomingCall (m_CIPValue, m_szCallingNumber, szN);
                Conv_DuplicateString (&szN, 0);
            }
            return;
        case iWrn_MSNList_RejectCall:
            break;
        default:
            dassert (0);
            break;
        }
    }
    ConnectResp (&cpV_ConnectResp_Ignore);
}


/*===========================================================================*\
    CTransferChannel::ConnectActiveInd
\*===========================================================================*/

void CTransferChannel::ConnectActiveInd (c_struct ConnectedNumber) {
    dhead ("CTransferChannel::ConnectActiveInd", DEBUG_CTransferChannel);
    //tUInt        newController;
    //tUInt        newInfoMask;
    //tUInt        newCIPMask;
    //cp_listenreq newListenParams = {0};
    //MSNList.ResetGetNextMask();
    //while (MSNList.GetNextMask (&newController, &newInfoMask, &newCIPMask) == vTrue) {
    //    ListenReq (newController, &newListenParams);
    //}

    tUChar *ConNum = 0;
    ConvStruct_CallingNumber (&ConNum, ConnectedNumber);
    ConnectProceeding (ConNum);
    ConvStruct_CallingNumber (&ConNum, 0);
}


/*===========================================================================*\
    CTransferChannel::ConnectB3ActiveInd
\*===========================================================================*/

void CTransferChannel::ConnectB3ActiveInd (c_struct NCPI) {
    dhead ("CTransferChannel::ConnectB3ActiveInd", DEBUG_CTransferChannel);
    cu_NCPI *NCPIcapi = 0;
    ConvStruct_NCPI (&NCPIcapi, NCPI, m_B3Prot);

    //// Check ExtendedFax-Options, if CAPI tries to Build up with a set of not allowed features
    //if ((B3Prot == b3prot_G3Fax_extended) && (NCPIcapi->Options != 0)) {
    //    c_word *FAXoption;
    //    if (GetBeginAndLenOfStructData (B3Config, &(c_byte *)FAXoption) > 4 * sizeof (c_byte)) {
    //       FAXoption += 1;
    //       if (FAXoption & NCPIcapi->Options);
    //    }
    //}

    if (m_DTMFListen == vTrue) {
        fReq_DTMF (dtmffunction_StartListen, 0, 0, 0, 0);
    }
    IsConnected (NCPIcapi);
    ConvStruct_NCPI (&NCPIcapi, 0, b3prot_Transparent);
}


/*===========================================================================*\
    CTransferChannel::DataB3Ind
\*===========================================================================*/

void CTransferChannel::DataB3Ind (cp_datab3 *Params) {
    dhead ("CTransferChannel::DataB3Ind", DEBUG_CTransferChannel);
    dassert (Params != 0);
    GetData (Params->Data, Params->DataLength, Params->DataHandle);
}


/*===========================================================================*\
    CTransferChannel::DisconnectB3Ind
\*===========================================================================*/

void CTransferChannel::DisconnectB3Ind (cp_disconnectb3ind *Params) {
    dhead ("CTransferChannel::DisconnectB3Ind", DEBUG_CTransferChannel);
    dassert (Params != 0);
    m_Reseting &= ~CHANNEL_STATUS_DISCONNECT_WAIT;
    ResetSendBlock();
    cu_NCPI *pNCPI = 0;
    ConvStruct_NCPI (&pNCPI, Params->NCPI, m_B3Prot);
    DoDisconnecting (Params->ReasonB3, pNCPI);

#ifdef USE_SUPPLEMENTARY_SERVICES
    switch (m_SupplServType) {
    case supplserv_Hold:
        fReq_Hold();
        break;
    case supplserv_Suspend:
        dassert (m_ParkIdentity != 0);
        fReq_Suspend (m_ParkIdentity);
        break;
    }
    m_SupplServType = 0;
    Conv_DuplicateStruct (&m_ParkIdentity, 0);
#endif
}


/*===========================================================================*\
    CTransferChannel::DisconnectInd
\*===========================================================================*/

void CTransferChannel::DisconnectInd (c_info Reason) {
    dhead ("CTransferChannel::DisconnectInd", DEBUG_CTransferChannel);
    dwarning (m_ActiveController != IllegalController);
    if (m_ActiveController != IllegalController) {
        m_ActiveController  = IllegalController;
        m_CIPValue          = cipvalue_Undefined;
        m_DTMFListen        = vFalse;
        m_Reseting         &= ~CHANNEL_STATUS_DISCONNECT_WAIT;
        Conv_DuplicateStruct (&m_CalledPartyNum, 0);
        Conv_DuplicateStruct (&m_ConnectedPartyNum, 0);
        Conv_DuplicateString (&m_szCallingNumber, 0);
        Conv_DuplicateString (&m_ParkIdentity, 0);
        ResetSendBlock();
        if (m_DDIPrefixLen > 0) {
            m_DDIPrefixLen = 0;
        } else {
            IsDisconnected (Reason);
        }
        dprint ("DisconectInd-CheckListen ");
        CheckListen();
    }
}


/*===========================================================================*\
    CTransferChannel::InfoInd
\*===========================================================================*/

void CTransferChannel::InfoInd (cp_infoind *Params) {
    dhead ("CTransferChannel::InfoInd", DEBUG_CTransferChannel);
    dparams ("%x", Params->InfoNumber);
    dassert (Params != 0);
    // at the moment we only used this call for DDI
    switch (Params->InfoNumber) {
    case 0x70: // called party number info element
        if (GetState() != cs_D_ConnectIndication) {
            break;
        } else {
            c_byte *Data = 0;
            c_word  DataLen = GetBeginAndLenOfStructData (Params->InfoElement, &Data);
            dassert ((DataLen == 0) || (Data != 0));
            if ((DataLen <= 1) || (IsCapiStructEmpty (m_CalledPartyNum) == vTrue)) {
                break;
            }
            Data++;
            DataLen--;
            dassert (CalcCapiStructDataLen (m_CalledPartyNum) >= 1);
            tUInt NumLen = CalcCapiStructDataLen (m_CalledPartyNum) - 1;
            dprint ("CurDDILen=%x ", NumLen);
            dassert (NumLen <= CAPI_CALLEDPARTYNUM_MAXLEN);
            dassert (CAPI_CALLEDPARTYNUM_MAXLEN < 254);
            if (DataLen >= CAPI_CALLEDPARTYNUM_MAXLEN) {
                Data   += DataLen - CAPI_CALLEDPARTYNUM_MAXLEN;
                DataLen = CAPI_CALLEDPARTYNUM_MAXLEN;
                NumLen  = 0;
            } else if (DataLen + NumLen > CAPI_CALLEDPARTYNUM_MAXLEN) {
                tUByte *dst = m_CalledPartyNum + 2;
                tUByte *src = dst + CAPI_CALLEDPARTYNUM_MAXLEN - DataLen;
                for (tUInt i = NumLen + DataLen - CAPI_CALLEDPARTYNUM_MAXLEN; i > 0; i--) {
                    *dst++ = *src++;
                }
                NumLen = CAPI_CALLEDPARTYNUM_MAXLEN - DataLen;
            }
            s_memcpy (m_CalledPartyNum + NumLen + 2, Data, DataLen);
            m_CalledPartyNum[0] = (c_byte)(NumLen + DataLen + 1);
        }
        // no break!
    case 0xA1: // sending complete info element
        dprintBuf (m_CalledPartyNum, m_CalledPartyNum[0] + 1);
        dassert (m_ActiveController != IllegalController);
        if (  (GetState() == cs_D_ConnectIndication)
           && (m_CalledPartyNum[0] > m_DDIPrefixLen)
           && (m_pMSNList != 0)) {
            switch (m_pMSNList->TestDDINumber (m_ActiveController, m_CalledPartyNum + m_DDIPrefixLen + 1,
                                               m_CalledPartyNum[0] - m_DDIPrefixLen)) {
            case iWrn_MSNList_DDICall:
                // DDI Number not complete received -> wait for more numbers
                break;
            case i_Okay:
                if (AcceptIncomingCall() == vTrue) {
                    c_struct dummyDDINumber = m_CalledPartyNum + m_DDIPrefixLen + 1;
                    m_DDIPrefixLen          = 0;
                    tUChar *szN             = 0;
                    ConvStruct_CalledNumber (&szN, m_CalledPartyNum);
                    ConvUnion_CallingNumber (&m_ConnectedPartyNum, szN);
                    Conv_DuplicateString (&szN, 0);
                    IncomingCall (m_CIPValue, m_szCallingNumber, dummyDDINumber);
                }
                break;
            case iWrn_MSNList_RejectCall:
                Conv_DuplicateStruct (&m_CalledPartyNum, 0);
                ConnectResp (&cpV_ConnectResp_Ignore);
                break;
            default:
                dassert (0);
                break;
            }
        }
        break;
    }
}


/*===========================================================================*\
    CTransferChannel::ManufacturerInd
\*===========================================================================*/

void CTransferChannel::ManufacturerInd (c_dword, cp_manufacturer *Params) {
    dhead ("CTransferChannel::ManufacturerInd", DEBUG_CTransferChannel);
    dassert (Params != 0);
    ManufacturerNotify (Params->Function, Params->ManuData);
}


/*===========================================================================*\
    CTransferChannel::fConf_PMWakeup
\*===========================================================================*/

void CTransferChannel::fConf_PMWakeup (c_dword Controller, c_info Info, c_word NumberAccepted) {
    dhead ("CTransferChannel::fConf_PMWakeup", DEBUG_CTransferChannel);
    dassert (m_pMSNList != 0);
    if (CapiInfoIsSuccess (Info) == vTrue) {
        if (NumberAccepted < m_pMSNList->GetCountMSN (Controller)) {
            dwarning (0);
            fReq_PMWakeup (Controller, 0, 0);
        }
    } else {
        dprint ("fConf_PMWakeup returned with %x. ", Info);
        m_pMSNList->SetWakeupSupport (Controller, vFalse);
    }
}


/*===========================================================================*\
    CTransferChannel::ResetSendBlock
\*===========================================================================*/

void CTransferChannel::ResetSendBlock (void) {
    dhead ("CTransferChannel::ResetSendBlock", DEBUG_CTransferChannel);
    void *DelHandle;

    while (m_SendBlock.DelNextEntry (&DelHandle) == vTrue) {
        PutDataComplete (DelHandle, ci_Int_NotConnected);
    }
}


/*===========================================================================*\
\*===========================================================================*/
