/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#ifndef _CHANNEL_H_
#define _CHANNEL_H_

#include "Facility.h"
#include "DataHand.h"
#include "MSNList.h"
#include "ConvCapi.h"
#include "aFlags.h"

/*===========================================================================*\
\*===========================================================================*/

class CTransferChannel : private CCAPI20_Facility {
public:
    enum tDisconnectKind {
        DisconnectKind_Hard = 1,    // Makes Release
        DisconnectKind_Fast,        // Makes DisconnectReq (FastDisconnect = vTrue)
        DisconnectKind_Normal,      // Makes DisconnectReq
        DisconnectKind_WaitNoPut,   // Wait for all Outstanding DATA_B3_CONF's (without new PutData's)
        DisconnectKind_WaitWithPut  // Wait for all Outstanding DATA_B3_CONF's (with new PutData's)
    };

    CTransferChannel (void);
    virtual ~CTransferChannel (void);

    tCapiState CurState (void);
    tUShort    CountPutQueueItems (void);
    tBool      IsPutQueueEmpty (void);
    tBool      IsPutQueueFull (void);
    void       SetMaxDataSize (tUInt MaxDataSize);  // MaxDataSize == 0 => CAPI_MAX_DATA_SIZE
    void       SetRingingTimeout (tUInt secs);

    void SetOutController (tUInt Controller);       // OutController is used by MakeCall and UnparkCall
    tUInt GetOutController (void);                  // ActiveController is the Controller who is currently in use
    tUInt GetActiveController (void);               // both can differ if listen is active for another Controller
    void SetOwnNumber (tUChar *pMyNumber, tUByte Indicator = CALLNUM_INDICATOR_DEFAULT);
    tUChar *GetOwnNumber (void);
    void SetB1Parameters (c_b1prot B1Protocol, cu_B1Configuration *pB1Config = 0);
    void SetB2Parameters (c_b2prot B2Protocol);
    void SetB3Parameters (c_b3prot B3Protocol, cu_B3Configuration *pB3Config = 0);
    void SetAdditionalInfo (tUInt setflags, c_struct BChannelInfo, c_struct KeypadFacility, c_struct UserUserData,
                            c_struct FacilityDataArray);
    void SetNCPI (c_b3prot B3Protocol, cu_NCPI *pNCPI);

    tBool SetMSNList (CCntrlMSNList *pCntrlMSNList);
    CCntrlMSNList *GetMSNList (void);
    void CheckListen (void);
    virtual void ListenError (tUInt /*Controller*/, c_info /*Info*/) {}     // Controller == 0 => For all Controller

    tBool MakeCall (tUChar *pTelNumber, c_cipvalue CIPValue);
    void  AlertCall (void);
    void  AnswerCall (c_reject Reject = creject_Accept);
    tBool PutData (tUByte *Data, tUShort DataLength, void *hDataID);        // vFalse => QueueFull or Disconnect
    void  GetDataComplete (tUShort DataHandle);
    void  Disconnect (tDisconnectKind kind = DisconnectKind_Normal);
    void  ResetChannel (void);
    // Call ResetChannel always before deleting this class !!!
    void  UpdateBParameters (void);
    void  ManufacturerCall (c_dword Function, c_struct Data);

    virtual void IncomingCall (c_cipvalue CIPValue, tUChar *pOppositeNumber, tUChar *pMyCalledNumber) = 0;
    virtual void DialingBegin (void) {}                                     // (ConnectReq)
    virtual void DialingEnd (void) {}                                       // (ConnectConf)
    virtual void ConnectProceeding (tUChar * /*ConnectedNumber */) {}       // (ConnectActiveInd)
    virtual void IsConnected (cu_NCPI *pNCPI) = 0;
    virtual void PutDataComplete (void *hDataID, c_info Info) = 0;
    virtual void GetData (tUByte *Data, tUShort DataLength, tUShort DataHandle) = 0;
    virtual void DoDisconnecting (c_info /*ReasonB3*/, cu_NCPI * /*pNCPI*/) {}
    virtual void IsDisconnected (c_info Reason) = 0;
    virtual void IsReseted (void) = 0;
    virtual void UpdateBParametersComplete (c_info /*Info*/) {}
    virtual void ManufacturerCallComplete (c_dword /*Error*/, c_dword /*Function*/, c_struct /*Data*/) {}
    virtual void ManufacturerNotify (c_dword /*Function*/, c_struct /*Data*/) {}

#ifdef USE_SUPPLEMENTARY_SERVICES
    void HoldCall (void);
    void UnholdCall (void);
    void ParkCall (tUChar *pIDStr, tUInt len = 0);
    void UnparkCall (tUChar *pIDStr, tUInt len = 0);
    void RedirectCall (tUChar *pTelNumber);
    void TransferCall (CTransferChannel *pChannel);                 // this is in hold, pChannel is active
    void Conference (CTransferChannel *pChannel, tBool doBegin);    // this is in hold, pChannel is active
    void Forward (c_dword Controller, c_forwardtype Type, tUChar *MSN, tUChar *ForwardNumber);
                                                                    // ForwardNumber = 0 => Deactivate
    void DTMFListen (tBool activate);
    void DTMFSend (tUInt count, tUChar *pDTMFDigits, tUShort ToneDuration = 0, tUShort GapDuration = 0);

    virtual void CallHolded (c_info /*Info*/) {}
    virtual void CallUnholded (c_info /*Info*/) {}
    virtual void CallParked (c_info /*Info*/) {}
    virtual void CallUnparked (c_info /*Info*/) {}
    virtual void CallRedirected (c_info /*Info*/) {}
    virtual void CallTransfered (c_info /*Info*/) {}
    virtual void ConferenceComplete (tBool /*Begin*/, c_info /*Info*/) {}
    virtual void ForwardComplete (c_info /*Info*/) {}
    virtual void DTMFReceived (tUInt /*count*/, tUChar * /*pDTMFChars*/) {}
    virtual void DTMFSendComplete (c_info /*Info*/) {}
    virtual void DTMFListenError (c_info /*Info*/) {}

private:
    /*-----------------------------------------------------------------------*\
        virtual functions from CCAPI20_Facility
    \*-----------------------------------------------------------------------*/
    void fConf_DTMF (c_info Info);
    void fInd_DTMF (c_struct DTMFdigits);
    void fInd_Hold (c_info Reason);
    void fInd_Retrieve (c_info Reason);
    void fInd_Suspend (c_info Reason);
    void fInd_Resume (c_info Reason);
    void fInd_CD (c_info Reason);
    void fInd_ECT (c_info Reason);
    void fInd_3PTY_Begin (c_info Reason);
    void fInd_3PTY_End (c_info Reason);
#endif

private:
    /*-----------------------------------------------------------------------*\
        virtual functions from CCAPI20_Channel
    \*-----------------------------------------------------------------------*/
    void Register_Complete (c_info ErrorCode);
    void Release_Complete (void);
    void ConnectConf (void);
    void DataB3Conf (cp_datab3conf *Params);
    void ListenConf (c_dword Controller, c_info Info);
    void ManufacturerConf (c_dword Controller, cp_manufacturer *Params);
    void SelectBProtocolConf (c_info Info);
    void ConnectInd (c_dword Controller, cp_connectind *Params);
    void ConnectActiveInd (c_struct ConnectedNumber);
    void ConnectB3ActiveInd (c_struct NCPI);
    void DataB3Ind (cp_datab3 *Params);
    void DisconnectB3Ind (cp_disconnectb3ind *Params);
    void DisconnectInd (c_info Reason);
    void InfoInd (cp_infoind *Params);
    void ManufacturerInd (c_dword Controller, cp_manufacturer *Params);
    void fConf_PMWakeup (c_dword Controller, c_info Info, c_word NumberAccepted);

    /*-----------------------------------------------------------------------*\
        Interne Funktionen
    \*-----------------------------------------------------------------------*/
    void ResetSendBlock (void);

    /*-----------------------------------------------------------------------*\
        Interne Variablen
    \*-----------------------------------------------------------------------*/
    CDataHandleAdmin  m_SendBlock;
    CCntrlMSNList    *m_pMSNList;
    tBool             m_deleteMSNList;
    tBool             m_WaitForDDINumbers;
    tUChar           *m_szCallingNumber;

    c_dword           m_OutController;
    c_dword           m_ActiveController;
    c_cipvalue        m_CIPValue;
    c_b1prot          m_B1Prot;
    c_b2prot          m_B2Prot;
    c_b3prot          m_B3Prot;
    c_struct          m_B1Config;
    c_struct          m_B3Config;
    c_struct          m_sMSN;
    c_struct          m_CalledPartyNum;
    c_struct          m_ConnectedPartyNum;
    tUInt             m_MaxDataSize;
    tUInt             m_DDIPrefixLen;
    c_struct          m_ParkIdentity;
    tUInt             m_SupplServType;
    tUInt             m_DTMFState;
    c_struct          m_DTMFDigits;
    tUShort           m_DTMFToneDuration;
    tUShort           m_DTMFGapDuration;
    tBool             m_DTMFListen;
    tUInt             m_Reseting;
    tUInt             m_CheckListenInUse;
    cp_additionalinfo m_AdditionalInfo;
};


/*===========================================================================*\
    Inline Funktionen
\*===========================================================================*/

inline tCapiState CTransferChannel::CurState (void) {
    return GetState();
}

inline tUShort CTransferChannel::CountPutQueueItems (void) {
    return m_SendBlock.CurCount();
}

inline tBool CTransferChannel::IsPutQueueEmpty (void) {
    return m_SendBlock.IsEmpty();
}

inline tBool CTransferChannel::IsPutQueueFull (void) {
    return m_SendBlock.IsFull();
}

inline void CTransferChannel::SetMaxDataSize (tUInt MaxDataSize) {
    if (!MaxDataSize || (MaxDataSize > CAPI_MAX_DATA_SIZE)) {
        MaxDataSize = CAPI_MAX_DATA_SIZE;
    }
    m_MaxDataSize = MaxDataSize;
}

inline void CTransferChannel::SetRingingTimeout (tUInt secs) {
    CCAPI20_Channel::SetRingingTimeout (secs);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void CTransferChannel::SetOutController (tUInt Controller) {
    // assert (Controller != IllegalController);
    m_OutController = Controller;
}

inline tUInt CTransferChannel::GetOutController (void) {
    return m_OutController;
}

inline tUInt CTransferChannel::GetActiveController (void) {
    return m_ActiveController;
}

inline void CTransferChannel::SetOwnNumber (tUChar *pMyNumber, tUByte Indicator) {
    ConvUnion_CallingNumber (&m_sMSN, pMyNumber, Indicator);
}

inline tUChar *CTransferChannel::GetOwnNumber (void) {
    tUChar *ownNumber = 0;
    if (GetBeginAndLenOfStructData (m_sMSN, &ownNumber) > 2) {
        return ownNumber + 2;
    }
    return 0;
}

inline void CTransferChannel::SetB1Parameters (c_b1prot B1Protocol, cu_B1Configuration *pB1Config) {
    m_B1Prot = B1Protocol;
    ConvUnion_B1Configuration (&m_B1Config, pB1Config, B1Protocol);
}

inline void CTransferChannel::SetB2Parameters (c_b2prot B2Protocol) {
    m_B2Prot = B2Protocol;
}

inline void CTransferChannel::SetB3Parameters (c_b3prot B3Protocol, cu_B3Configuration *pB3Config) {
    m_B3Prot = B3Protocol;
    ConvUnion_B3Configuration (&m_B3Config, pB3Config, B3Protocol);
}

inline void CTransferChannel::SetAdditionalInfo (tUInt setflags, c_struct BChannelInfo, c_struct KeypadFacility,
                                                 c_struct UserUserData, c_struct FacilityDataArray) {
    if (CheckOneFlag (setflags, 0)) {
        Conv_DuplicateStruct (&m_AdditionalInfo.BChannelInformation, BChannelInfo);
    }
    if (CheckOneFlag (setflags, 1)) {
        Conv_DuplicateStruct (&m_AdditionalInfo.KeypadFacility, KeypadFacility);
    }
    if (CheckOneFlag (setflags, 2)) {
        Conv_DuplicateStruct (&m_AdditionalInfo.UserUserData, UserUserData);
    }
    if (CheckOneFlag (setflags, 3)) {
        Conv_DuplicateStruct (&m_AdditionalInfo.FacilityDataArray, FacilityDataArray);
    }
}

inline void CTransferChannel::SetNCPI (c_b3prot B3Protocol, cu_NCPI *pNCPI) {
    c_struct NCPIuser = 0;
    ConvUnion_NCPI (&NCPIuser, pNCPI, B3Protocol);
    CCAPI20_Channel::SetNCPI (NCPIuser);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tBool CTransferChannel::SetMSNList (CCntrlMSNList *pCntrlMSNList) {
    if (m_deleteMSNList == vTrue) {
        delete m_pMSNList;
    }
    m_pMSNList      = pCntrlMSNList;
    m_deleteMSNList = vFalse;
    CheckListen();
    return vTrue;
}

inline CCntrlMSNList *CTransferChannel::GetMSNList (void) {
    if (!m_pMSNList) {
        m_pMSNList      = new CCntrlMSNList;
        m_deleteMSNList = vTrue;
    }
    return m_pMSNList;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void CTransferChannel::AlertCall (void) {
    if (GetState() == cs_D_ConnectIndication) {
        AlertReq (0);
    }
}

inline void CTransferChannel::GetDataComplete (tUShort DataHandle) {
    DataB3Resp (DataHandle);
}

inline void CTransferChannel::Disconnect (tDisconnectKind kind) {
    switch (kind) {
    case DisconnectKind_WaitWithPut:
        if (IsPutQueueEmpty() == vFalse) {
            m_Reseting = 4;
        } else {
            DisconnectReq();
        }
        break;
    case DisconnectKind_WaitNoPut:
        if (IsPutQueueEmpty() == vFalse) {
            m_Reseting = 2;
        } else {
            DisconnectReq();
        }
        break;
    default:
    case DisconnectKind_Normal:
        DisconnectReq();
        break;
    case DisconnectKind_Fast:
        DisconnectReq (vTrue);
        break;
    case DisconnectKind_Hard:
        Release();
        break;
    }
}

inline void CTransferChannel::ConnectConf (void) {
    DialingEnd();
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#ifdef USE_SUPPLEMENTARY_SERVICES

inline void CTransferChannel::fInd_DTMF (c_struct DTMFdigits) {
    // dwarning (m_DTMFListen == vTrue);
    if (m_DTMFListen == vTrue) {
        tUChar *digits = 0;
        tUInt len = GetBeginAndLenOfStructData (DTMFdigits, &digits);
        if (len) {
            DTMFReceived (len, digits);
        }
    }
}

inline void CTransferChannel::fInd_Hold (c_info Reason) {
    CallHolded (Reason);
}

inline void CTransferChannel::fInd_Retrieve (c_info Reason) {
    CallUnholded (Reason);
}

inline void CTransferChannel::fInd_Suspend (c_info Reason) {
    CallParked (Reason);
}

inline void CTransferChannel::fInd_Resume (c_info Reason) {
    if (CapiInfoIsSuccess (Reason)) {
        cp_bprotocol bprot = {m_B1Prot, m_B2Prot, m_B3Prot, m_B1Config, 0, m_B3Config};
        SelectBProtocolReq (&bprot);
    }
    CallUnparked (Reason);
}

inline void CTransferChannel::fInd_CD (c_info Reason) {
    CallRedirected (Reason);
}

inline void CTransferChannel::fInd_ECT (c_info Reason) {
    CallTransfered (Reason);
}

inline void CTransferChannel::fInd_3PTY_Begin (c_info Reason) {
    ConferenceComplete (vTrue, Reason);
}

inline void CTransferChannel::fInd_3PTY_End (c_info Reason) {
    ConferenceComplete (vFalse, Reason);
}

#endif // USE_SUPPLEMENTARY_SERVICES
/*===========================================================================*\
\*===========================================================================*/

#endif
