/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "ConvCapi.h"
#include "c2msgdef.h"
#include "dbgCAPI.h"

/*===========================================================================*\
    From Capi-Structs to Capi-Unions
\*===========================================================================*/

tUInt ConvStruct_CalledNumber (tUChar **pTelNum, c_struct PartyNumber) {
    dhead ("ConvStruct_CalledNumber", DEBUG_CAPI_CONVERT);
    dassert (pTelNum != 0);

    Conv_DuplicateString (pTelNum, 0);
    if (PartyNumber) {
        tUInt len = PartyNumber[0];
        if (len == 0xFF) {
            len = get16 (PartyNumber + 1);
            PartyNumber += 2;
        }
        PartyNumber += 2;
        if (len > 1) {
            tUChar *tmpStr = new tUChar [len];
            dwarning (tmpStr != 0);
            if (tmpStr) {
                len--;
                s_memcpy (tmpStr, PartyNumber, len);
                tmpStr[len] = '\0';
                *pTelNum = tmpStr;
                return len;
            }
        }
    }
    return 0;
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvStruct_CallingNumber (tUChar **pTelNum, c_struct PartyNumber) {
    dhead ("ConvStruct_CallingNumber", DEBUG_CAPI_CONVERT);
    dassert (pTelNum != 0);

    Conv_DuplicateString (pTelNum, 0);
    if (PartyNumber) {
        tUInt len = PartyNumber[0];
        if (len == 0xFF) {
            len = get16 (PartyNumber + 1);
            PartyNumber += 2;
        }
        PartyNumber += 3;
        if (len > 2) {
            len--;
            tUChar *tmpStr = new tUChar [len];
            dwarning (tmpStr != 0);
            if (tmpStr) {
                len--;
                s_memcpy (tmpStr, PartyNumber, len);
                tmpStr[len] = '\0';
                *pTelNum = tmpStr;
                return len;
            }
        }
    }
    return 0;
}


/*===========================================================================*\
\*===========================================================================*/

void ConvStruct_NCPI (cu_NCPI **pCU_NCPI, c_struct CS_NCPI, c_b3prot B3Protocol) {
    dhead ("ConvStruct_NCPI", DEBUG_CAPI_CONVERT);
    dassert (pCU_NCPI != 0);
    cu_NCPI *tmpNCPI;
    tBool Okay = vFalse;

    if (*pCU_NCPI) {
        tmpNCPI = *pCU_NCPI;
        *pCU_NCPI = 0;
        delete [] tmpNCPI->ReceiveID;
        delete [] tmpNCPI;
    }

    if (CS_NCPI) {
        tmpNCPI = new cu_NCPI [1];
        dwarning (tmpNCPI != 0);
        if (tmpNCPI) {
            s_memzero (tmpNCPI, sizeof (cu_NCPI));
            tUInt len = CS_NCPI[0];
            CS_NCPI++;
            if (len == 0xFF) {
                len = get16 (CS_NCPI);
                CS_NCPI += sizeof (c_word);
            }
            switch (B3Protocol) {
            case b3prot_ISO8208:
            case b3prot_X25:
                if (len > 2) {
                    if (B3Protocol == 2) {
                        tmpNCPI->Flags   = CS_NCPI[0];
                    } else {
                        tmpNCPI->Options = CS_NCPI[0];
                    }
                    tmpNCPI->Group     = CS_NCPI[1];
                    tmpNCPI->Channel   = CS_NCPI[2];
                    Okay = vTrue;
                    break;
                }
                break;
            case b3prot_G3Fax:
            case b3prot_G3Fax_extended:
                if (len > 8) {
                    tmpNCPI->Rate      = get16 (CS_NCPI);
                    tmpNCPI->Format    = get16 (CS_NCPI + 2 * sizeof (c_word));
                    tmpNCPI->Pages     = get16 (CS_NCPI + 3 * sizeof (c_word));
                    if (B3Protocol == b3prot_G3Fax) {
                        tmpNCPI->Resolution = get16 (CS_NCPI + sizeof (c_word));
                        tmpNCPI->Options    = 0;
                    } else {
                        tmpNCPI->Options    = get16 (CS_NCPI + sizeof (c_word));
                        tmpNCPI->Resolution = 0;
                    }
                    len      = CS_NCPI[4 * sizeof (c_word)];
                    CS_NCPI += 4 * sizeof (c_word) + sizeof (c_byte);
                    if (len == 0xFF) {
                        len = get16 (CS_NCPI);
                        CS_NCPI += sizeof (c_word);
                    }
                    dprint ("RecvID-len=%x ", len);
                    if (len > 0) {
                        tmpNCPI->ReceiveID = new tUChar [len + 1];
                        // TODO: Zur Zeit kann nicht erkannt werden, ob new schief
                        //       geht oder keine RecieveID vorhanden ist
                        dwarning (tmpNCPI->ReceiveID != 0);
                        if (tmpNCPI->ReceiveID) {
                            s_memcpy (tmpNCPI->ReceiveID, CS_NCPI, len);
                            tmpNCPI->ReceiveID[len] = '\0';
                        }
                    }
                    Okay = vTrue;
                }
                break;
            case b3prot_Modem:
                if (len > 3) {
                    tmpNCPI->Rate      = *((tUShort *)(CS_NCPI));
                    tmpNCPI->Protocol  = *((tUShort *)(CS_NCPI + 2));
                    Okay = vTrue;
                }
                break;
            default:
                // nothing to do!
                break;
            }
            if (Okay == vFalse) {
                dwarning (0);
                delete [] tmpNCPI->ReceiveID;
                delete [] tmpNCPI;
                tmpNCPI = 0;
            }
        }
        *pCU_NCPI = tmpNCPI;
    }
}


/*===========================================================================*\
    From Capi-Unions to Capi-Structs
\*===========================================================================*/

tUInt ConvUnion_CalledNumber (c_struct *pCalledNumber, tUChar *TelNum) {
    dhead ("ConvUnion_CalledNumber", DEBUG_CAPI_CONVERT);
    dassert (pCalledNumber != 0);
    tUInt len = (tUChar)((TelNum) ? strlen ((char *)TelNum) : 0);
    dassert (len < 65000);
    Conv_DuplicateStruct (pCalledNumber, 0);
    if (len) {
        dassert (TelNum != 0);
        tUInt needLen = CapiStruct_AllocSize (len + 1);
        c_struct tmpNumber = new c_byte [needLen + 1];
        dwarning (tmpNumber != 0);
        if (tmpNumber) {
            dprint ("newP=%p ", tmpNumber);
            *pCalledNumber = tmpNumber;
            CapiStruct_AddLen    (&tmpNumber, len + 1);
            CapiStruct_AddByte   (&tmpNumber, 0x80);
            CapiStruct_AddBuffer (&tmpNumber, TelNum, len + 1);
            // a little trick: we save the final '\0' so we can use this struct as a string, too
            RETURN ('x', needLen);
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvUnion_CallingNumber (c_struct *pCallingNumber, tUChar *TelNum, tUByte Indicator) {
    dhead ("ConvUnion_CallingNumber", DEBUG_CAPI_CONVERT);
    dassert (pCallingNumber != 0);
    tUInt len = (tUChar)((TelNum) ? strlen ((char *)TelNum) : 0);
    dassert (len < 65000);

    Conv_DuplicateStruct (pCallingNumber, 0);
    if (len || (Indicator != CALLNUM_INDICATOR_DEFAULT)) {
        tUInt needLen = CapiStruct_AllocSize (len + 2);
        c_struct tmpNumber = new c_byte [needLen + 1];
        dwarning (tmpNumber != 0);
        if (tmpNumber) {
            dprint ("newP=%p ", tmpNumber);
            *pCallingNumber = tmpNumber;
            CapiStruct_AddLen    (&tmpNumber, len + 2);
            CapiStruct_AddByte   (&tmpNumber, 0x00);
            CapiStruct_AddByte   (&tmpNumber, Indicator);
            CapiStruct_AddBuffer (&tmpNumber, TelNum, len);
            CapiStruct_AddByte   (&tmpNumber, 0x00);
            // a little trick: we add a final '\0' so we can use this struct as a string, too
            RETURN ('x', needLen);
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvUnion_FacilityNumber (c_struct *pFacilityNumber, tUChar *TelNum, tUByte Indicator) {
    dhead ("ConvUnion_CallingNumber", DEBUG_CAPI_CONVERT);
    dassert (pFacilityNumber != 0);
    tUInt len = (tUChar)((TelNum) ? strlen ((char *)TelNum) : 0);
    dassert (len < 65000);

    Conv_DuplicateStruct (pFacilityNumber, 0);
    if (len || (Indicator != CALLNUM_INDICATOR_DEFAULT)) {
        tUInt needLen = len + 3 + (len > 251) ? 3 : 1;
        c_struct tmpNumber = new c_byte [needLen + 1];
        dwarning (tmpNumber != 0);
        if (tmpNumber) {
            dprint ("newP=%p ", tmpNumber);
            *pFacilityNumber = tmpNumber;
            CapiStruct_AddLen    (&tmpNumber, len + 3);
            CapiStruct_AddByte   (&tmpNumber, 0x01);        // Public Party Number
            CapiStruct_AddByte   (&tmpNumber, 0x00);
            CapiStruct_AddByte   (&tmpNumber, Indicator);
            CapiStruct_AddBuffer (&tmpNumber, TelNum, len);
            CapiStruct_AddByte   (&tmpNumber, 0x00);
            // a little trick: we add a final '\0' so we can use this struct as a string, too
            RETURN ('x', needLen);
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvUnion_B1Configuration (c_struct *pCS_B1Config, cu_B1Configuration *CU_B1Conf, c_b1prot B1Protocol) {
    dhead ("ConvUnion_B1Config", DEBUG_CAPI_CONVERT);
    dassert (pCS_B1Config != 0);
    tUInt len = 0;
    c_struct tmpB1Config = 0;

    Conv_DuplicateStruct (pCS_B1Config, 0);
    if (CU_B1Conf) {
        switch (B1Protocol) {
        case b1prot_V110_asynch:
        case b1prot_V110_synch:
        case b1prot_G3Fax:
            len = 4 * 2 + 1;                // Wichtig: 2 = sizeof (c_word)
            tmpB1Config = new c_byte [len];
            dwarning (tmpB1Config != 0);
            break;

        case b1prot_Modem_negotiate:
        case b1prot_Modem_asynch:
        case b1prot_Modem_synch:
            len = 6 * 2 + 1;                // Wichtig: 2 = sizeof (c_word)
            tmpB1Config = new c_byte [len];
            dwarning (tmpB1Config != 0);
            if (tmpB1Config) {
                set16 (tmpB1Config +  9, CU_B1Conf->Options);
                set16 (tmpB1Config + 11, CU_B1Conf->SpeedNegotiation);
            }
            break;
        default:
            // nothing to do!
            break;
        }
        if (tmpB1Config) {
            dassert (len > 0);
            dprint ("newP=%p ", tmpB1Config);
            s_memzero (tmpB1Config, len);
            /*----- Das gilt fr Protocol: 2, 3, 4, 7, 8, 9 -----*/
            *tmpB1Config = (tUChar)(len - 1);
            set16 (tmpB1Config + 1, CU_B1Conf->MaximumBitRate);

            switch (B1Protocol) {
            case b1prot_G3Fax:
                dassert (len > 3);
                set16 (tmpB1Config + 3, CU_B1Conf->TransmitLevelInDB);
                break;
            case b1prot_V110_asynch:
            case b1prot_Modem_negotiate:
            case b1prot_Modem_asynch:
                dassert (len > 7);
                set16 (tmpB1Config + 3, CU_B1Conf->BitsPerCharacter);
                set16 (tmpB1Config + 5, CU_B1Conf->Parity);
                set16 (tmpB1Config + 7, CU_B1Conf->StopBits);
                break;
            default:
                // nothing to do!
                break;
            }
        }
    }
    *pCS_B1Config = tmpB1Config;
    dassert ((len == 0) ? (tmpB1Config == 0) : (tmpB1Config != 0));
    RETURN ('x', (tmpB1Config) ? len : 0);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvUnion_B3Configuration (c_struct *pCS_B3Config, cu_B3Configuration *CU_B3Conf, c_b3prot B3Protocol) {
    dhead ("ConvUnion_B3Config", DEBUG_CAPI_CONVERT);
    dassert (pCS_B3Config != 0);
    tUInt len = 0;

    Conv_DuplicateStruct (pCS_B3Config, 0);
    if (CU_B3Conf) {
        switch (B3Protocol) {
        case b3prot_T90NL:
        case b3prot_ISO8208:
        case b3prot_X25: {
            len = 17;
            c_struct tmpB3Config = new c_byte [len];
            dwarning (tmpB3Config != 0);
            if (!tmpB3Config) {
                break;
            }
            dprint ("newP=%p ", tmpB3Config);
            set8  (tmpB3Config,       16);
            set16 (tmpB3Config + 0x1, CU_B3Conf->LIC);
            set16 (tmpB3Config + 0x3, CU_B3Conf->HIC);
            set16 (tmpB3Config + 0x5, CU_B3Conf->LTC);
            set16 (tmpB3Config + 0x7, CU_B3Conf->HTC);
            set16 (tmpB3Config + 0x9, CU_B3Conf->LOC);
            set16 (tmpB3Config + 0xB, CU_B3Conf->HOC);
            set16 (tmpB3Config + 0xD, CU_B3Conf->ModuloMode);
            set16 (tmpB3Config + 0xF, CU_B3Conf->WindowSize);
            *pCS_B3Config = tmpB3Config;
            break;
        }
        case b3prot_G3Fax:
        case b3prot_G3Fax_extended: {
            tUShort StationIdLen = (tUShort)((CU_B3Conf->StationID) ? strlen ((char *)CU_B3Conf->StationID) : 0);
            tUShort HeadLineLen  = (tUShort)((CU_B3Conf->HeadLine) ? strlen ((char *)CU_B3Conf->HeadLine) : 0);
            len = 4 + CapiStruct_AllocSize (StationIdLen) + CapiStruct_AllocSize (HeadLineLen);
            c_struct tmpB3Config = CapiStruct_AllocStruct (len);
            dwarning (tmpB3Config != 0);
            if (!tmpB3Config) {
                break;
            }
            dprint ("newP2=%p ", tmpB3Config);
            *pCS_B3Config = tmpB3Config;
            CapiStruct_AddLen    (&tmpB3Config, len);
            CapiStruct_AddWord   (&tmpB3Config, CU_B3Conf->Options);
            CapiStruct_AddWord   (&tmpB3Config, CU_B3Conf->Format);
            CapiStruct_AddLen    (&tmpB3Config, StationIdLen);
            CapiStruct_AddBuffer (&tmpB3Config, CU_B3Conf->StationID, StationIdLen);
            CapiStruct_AddLen    (&tmpB3Config, HeadLineLen);
            CapiStruct_AddBuffer (&tmpB3Config, CU_B3Conf->HeadLine, HeadLineLen);
            break;
        }
        default:
            // nothing to do!
            break;
        }
    }
    dassert ((len == 0) ? (*pCS_B3Config == 0) : (*pCS_B3Config != 0));
    RETURN ('x', len);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvUnion_NCPI (c_struct *pCS_NCPI, cu_NCPI *CU_NCPI, c_b3prot B3Protocol) {
    dhead ("ConvUnion_NCPI", DEBUG_CAPI_CONVERT);
    dassert (pCS_NCPI != 0);
    tUInt len = 0;
    c_struct tmpNCPI = 0;

    Conv_DuplicateStruct (pCS_NCPI, 0);
    if (CU_NCPI) {
        switch (B3Protocol) {
        case b3prot_ISO8208:
        case b3prot_X25:
            len = 4;
            tmpNCPI = new c_byte [len];
            dwarning (tmpNCPI != 0);
            if (!tmpNCPI) {
                break;
            }
            dprint ("newP=%p ", tmpNCPI);
            set8 (tmpNCPI, 3);
            set8 (tmpNCPI + 1, (tUChar)((B3Protocol == b3prot_ISO8208) ? CU_NCPI->Flags : (CU_NCPI->Options & 0xFF)));
            set8 (tmpNCPI + 2, CU_NCPI->Group);
            set8 (tmpNCPI + 3, CU_NCPI->Channel);
            /*----- TODO: Contents ist noch nicht drin!! -----*/
            break;
      //case b3prot_G3Fax fehlt, da diese Conv-Func nur fr Request's ist => empty struct!!
        case b3prot_G3Fax_extended:
            len = 10;
            tmpNCPI = new c_byte [len];
            dwarning (tmpNCPI != 0);
            if (!tmpNCPI) {
                break;
            }
            dprint ("newP2=%p ", tmpNCPI);
            set8  (tmpNCPI,     9);
            set16 (tmpNCPI + 1, 0);
            set16 (tmpNCPI + 3, CU_NCPI->Options);
            set16 (tmpNCPI + 5, CU_NCPI->Format);
            set16 (tmpNCPI + 7, 0);
            set8  (tmpNCPI + 9, 0);
            break;
        case b3prot_Modem:
            len = 5;
            tmpNCPI = new c_byte [len];
            dprint ("newP3=%p ", tmpNCPI);
            dwarning (tmpNCPI != 0);
            if (!tmpNCPI) {
                break;
            }
            set8  (tmpNCPI,     4);
            set16 (tmpNCPI + 1, CU_NCPI->Rate);
            set16 (tmpNCPI + 3, CU_NCPI->Protocol);
            break;
        default:
            // nothing to do!
            break;
        }
    }
    *pCS_NCPI = tmpNCPI;
    dassert ((len == 0) ? (tmpNCPI == 0) : (tmpNCPI != 0));
    RETURN ('x', len);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt ConvUnion_String (c_struct *pStruct, tUChar *String, tUInt len) {
    dhead ("ConvUnion_String", DEBUG_CAPI_CONVERT);
    dassert (pStruct != 0);
    Conv_DuplicateStruct (pStruct, 0);
    if (String) {
        if (!len) {
            len = s_strlen (String);
        }
        if (len) {
            tUInt needLen = len + ((len > 254) ? 3 : 1);
            c_struct tmpStruct = new c_byte [needLen];
            if (tmpStruct)  {
                *pStruct = tmpStruct;
                CapiStruct_AddLen    (&tmpStruct, len);
                CapiStruct_AddBuffer (&tmpStruct, String, len);
                RETURN ('x', needLen);
            }
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
    Duplicate Structs
\*===========================================================================*/

tUInt Conv_DuplicateStruct (c_struct *pDest, c_struct source) {
    dhead ("Conv_DuplicateStruct", DEBUG_CAPI_CONVERT);
    dassert (pDest != 0);
    dprint ("dest=%p,src=%p ", *pDest, source);
    c_struct tmpStruct;
    if (*pDest != 0) {
        tmpStruct = *pDest;
        *pDest = 0;
        delete [] tmpStruct;
    }
    tUInt length = CalcCapiStructLen (source);
    dprint ("len=%x ", length);
    if (length > 1) {
        dassert (source != 0);
        tmpStruct = new c_byte [length];
        dwarning (tmpStruct != 0);
        if (tmpStruct) {
            dprint ("newP=%p ", tmpStruct);
            s_memcpy (tmpStruct, source, length);
            *pDest = tmpStruct;
            RETURN ('x', length);
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
    Duplicate Strings
\*===========================================================================*/

tUInt Conv_DuplicateString (tUChar **ppDest, tUChar *pSource) {
    dhead ("DuplicateString", DEBUG_CAPI_CONVERT);
    dassert (ppDest != 0);
    tUChar *tmpString;
    dprint ("dest=%p,src=%p ", *ppDest, pSource);

    if (*ppDest) {
        dprint ("del='%s' ", *ppDest);
        tmpString = *ppDest;
        *ppDest = 0;
        delete [] tmpString;
    }
    if (pSource) {
        tUInt len = strlen ((char *)pSource);
        dprint ("set=(%x)'%s' ", len, pSource);
        if (len) {
            len++;
            tmpString = new tUChar [len];
            dwarning (tmpString != 0);
            if (tmpString) {
                dprint ("newP=%p ", tmpString);
                s_memcpy (tmpString, pSource, len);
                *ppDest = tmpString;
                RETURN ('x', len);
            }
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
\*===========================================================================*/
