/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "Facility.h"
#include "aString.h"
#include "dbgCAPI.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define MAX_CALL_IDENTITY_LEN           100     // < 250 !!!

static c_dword unique_Handle = 1;

/*===========================================================================*\
\*===========================================================================*/

/*===========================================================================*\
    CCAPI20_Facility Konstruktor
\*===========================================================================*/


CCAPI20_Facility::CCAPI20_Facility(void) {
   Handle = unique_Handle++;
}

void CCAPI20_Facility::fReq_DTMF (c_dtmffunction Function, c_word ToneDuration, c_word GapDuration,
                                  c_struct DTMFDigits, c_struct DTMFCharacteristics) {
    dhead ("CCAPI20_Facility::fReq_DTMF", DEBUG_CCAPI20_Facility);
    // generate cp_facility struct
    c_dword     localNCCI;
    tUInt       dlen     = CalcCapiStructLen (DTMFDigits);
    tUInt       clen     = CalcCapiStructLen (DTMFCharacteristics);
    tUInt       len      = 3 * sizeof (c_word) + dlen + clen;
    c_struct data = CapiStruct_AllocStruct (len);
    if (!data) {
        fConf_DTMF (ci_Int_MemoryFullErr);
        return;
    }
    cp_facility cpfac;
    cpfac.FacilitySelector = facilselect_DTMF;
    cpfac.FacilityParam    = data;

    CapiStruct_AddLen (&data, len);
    CapiStruct_AddWord (&data, (tUShort)Function);
    CapiStruct_AddWord (&data, ToneDuration);
    CapiStruct_AddWord (&data, GapDuration);
    CapiStruct_AddStruct (&data, DTMFDigits);
    CapiStruct_AddStruct (&data, DTMFCharacteristics);

    // send facility-req
    localNCCI = GetNCCI();
    if (GetState() != cs_Connected) {
        fConf_DTMF (ci_Int_NotConnected);
        return;
    }
    dassert (localNCCI != IllegalNCCI);
    FACILITY_REQ (GetNextMsgNr(), localNCCI, &cpfac);
    CapiStruct_FreeStruct (cpfac.FacilityParam);
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_PMWakeup (c_dword Controller, c_word NumberRequested, cu_PMWakeupNumber *NumberList) {
    dhead ("CCAPI20_Facility::fReq_PMWakeup", DEBUG_CCAPI20_Facility);
    dassert ((NumberRequested == 0) || (NumberList != 0));
    tUInt              i;
    tUInt              CurLen;
    tUInt              Len    = sizeof (c_word);
    cu_PMWakeupNumber *CurNum = NumberList;

    // calculate needed size and allocate it
    if (!NumberRequested) {
        // empty struct
        c_byte data[4] = { 3, 0, 0, 0 };
        cp_facility cpfac;
        cpfac.FacilitySelector = facilselect_PMWakeup;
        cpfac.FacilityParam    = data;
        FACILITY_REQ (GetNextMsgNr(), Controller, &cpfac);
        return;
    }
    for (i = NumberRequested; i > 0; i--, CurNum++) {
        dassert (CalcCapiStructLen (CurNum->CalledPartyNumber) > 1);
        CurLen = CalcCapiStructLen (CurNum->CalledPartyNumber) + sizeof (c_dword) + sizeof (c_byte);
        if (CurLen > 254 + 1) {
            CurLen += sizeof (c_word);
        }
        Len += CurLen;
    }
    if (Len > CAPI_MAX_MESSAGE_SIZE - CAPI_MSG_HEADER_ADDR_SIZE - sizeof (c_word)) {
        fConf_PMWakeup (Controller, ci_Int_MsgSizeOverflow, 0);
        return;
    }
    c_struct data = CapiStruct_AllocStruct (Len);
    if (!data) {
        fConf_PMWakeup (Controller, ci_Int_MemoryFullErr, 0);
        return;
    }
    cp_facility cpfac;
    cpfac.FacilitySelector = facilselect_PMWakeup;
    cpfac.FacilityParam    = data;

    // fill struct
    CapiStruct_AddLen (&data, Len);
    CapiStruct_AddWord (&data, NumberRequested);
    for (i = NumberRequested, CurNum = NumberList; i > 0; i--, CurNum++) {
        CurLen = CalcCapiStructLen (CurNum->CalledPartyNumber);
        dassert (CurLen > 1);
        CapiStruct_AddLen (&data, CurLen + sizeof (c_dword));
        CapiStruct_AddBuffer (&data, CurNum->CalledPartyNumber, CurLen);
        CapiStruct_AddDWord (&data, CurNum->CIPMask);
    }

    // send request
    FACILITY_REQ (GetNextMsgNr(), Controller, &cpfac);
    CapiStruct_FreeStruct (cpfac.FacilityParam);
}

/*===========================================================================*\
\*===========================================================================*/
#ifdef USE_SUPPLEMENTARY_SERVICES

void CCAPI20_Facility::fReq_Listen (c_dword Controller, c_dword NotificationMask) {
    dhead ("CCAPI20_Facility::fReq_Listen", DEBUG_CCAPI20_Facility);
    c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word) + sizeof (c_dword)];
    set8  (fparam,     7);
    set16 (fparam + 1, supplserv_Listen);
    set8  (fparam + 3, 4);
    set32 (fparam + 4, NotificationMask);
    CallSupplServFacilityReq (Controller, fparam);
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_Hold (void) {
    dhead ("CCAPI20_Facility::fReq_Hold", DEBUG_CCAPI20_Facility);
    if (SetState (cs_HoldPending, cs_D_Connected) == vTrue) {
        c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word)];
        set8  (fparam,     3);
        set16 (fparam + 1, supplserv_Hold);
        set8  (fparam + 3, 0);
        CallSupplServFacilityReq (GetPLCI(), fparam);
    } else {
        dwarning (0);
        fInd_Hold (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_Retrieve (void) {
    dhead ("CCAPI20_Facility::fReq_Retrieve", DEBUG_CCAPI20_Facility);
    if (SetState (cs_RetrievePending, cs_Holded) == vTrue) {
        c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word)];
        set8  (fparam,     3);
        set16 (fparam + 1, supplserv_Retrieve);
        set8  (fparam + 3, 0);
        CallSupplServFacilityReq (GetPLCI(), fparam);
    } else {
        dwarning (0);
        fInd_Retrieve (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_Suspend (c_struct CallIdentity) {
    dhead ("CCAPI20_Facility::fReq_Suspend", DEBUG_CCAPI20_Facility);
    tUInt len = CalcCapiStructLen (CallIdentity);
    dprint ("len=%x\n ", len);
    if (len > MAX_CALL_IDENTITY_LEN) {
        dwarning (0);
        fInd_Suspend (ci_Int_MsgSizeOverflow);
    } else if (SetState (cs_SuspendPending, cs_D_Connected) == vTrue) {
        dassert (MAX_CALL_IDENTITY_LEN + sizeof (c_word) + 2 * sizeof (c_byte) < 254);
        c_byte fparam [MAX_CALL_IDENTITY_LEN + sizeof (c_word) + 2 * sizeof (c_byte)];
        set8  (fparam,     (tUByte)(len + sizeof (c_word) + sizeof (c_byte)));
        set16 (fparam + 1, supplserv_Suspend);
        set8  (fparam + 3, (tUByte)len);
        if (len > 1) {
            s_memcpy (fparam + 4, CallIdentity, len);
        } else {
            set8 (fparam + 4, 0);
        }
        CallSupplServFacilityReq (GetPLCI(), fparam);
    } else {
        dwarning (0);
        fInd_Suspend (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_Resume (c_dword Controller, c_struct CallIdentity) {
    dhead ("CCAPI20_Facility::fReq_Resume", DEBUG_CCAPI20_Facility);
    tUInt len = CalcCapiStructLen (CallIdentity);
    dprint ("len=%x ", len);
    if (len > MAX_CALL_IDENTITY_LEN) {
        dwarning (0);
        fInd_Suspend (ci_Int_MsgSizeOverflow);
    } else if (SetState (cs_ResumePending, cs_Registered) == vTrue) {
        dassert (MAX_CALL_IDENTITY_LEN + sizeof (c_word) + 2 * sizeof (c_byte) < 254);
        c_byte fparam [MAX_CALL_IDENTITY_LEN + sizeof (c_word) + 2 * sizeof (c_byte)];
        set8  (fparam,     (tUByte)(len + sizeof (c_word) + sizeof (c_byte)));
        set16 (fparam + 1, supplserv_Resume);
        set8  (fparam + 3, (tUByte)len);
        if (len > 1) {
            s_memcpy (fparam + 4, CallIdentity, len);
        } else {
            set8 (fparam + 4, 0);
        }
        CallSupplServFacilityReq (Controller, fparam);
    } else {
        dwarning (0);
        fInd_Resume (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_ECT (CCAPI20_Channel * /*pChannel*/) {
    dhead ("CCAPI20_Facility::fReq_ECT", DEBUG_CCAPI20_Facility);
    //dassert (pChannel != 0);
    dwarning (GetState() == cs_Holded);
    c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word) + sizeof (c_dword)];
    set8  (fparam,     7);
    set16 (fparam + 1, supplserv_ECT);
    set8  (fparam + 3, sizeof (c_dword));
    set32 (fparam + 4, GetPLCI());
    //set32 (fparam + 4, pChannel->GetPLCI());
    CallSupplServFacilityReq (GetPLCI(), fparam);
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_3PTY_Begin (CCAPI20_Channel * /*pChannel*/) {
    dhead ("CCAPI20_Facility::fReq_3PTY_Begin", DEBUG_CCAPI20_Facility);
    //dassert (pChannel != 0);
    dwarning (GetState() == cs_Holded);
    c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word) + sizeof (c_dword)];
    set8  (fparam,     7);
    set16 (fparam + 1, supplserv_3PTY_Begin);
    set8  (fparam + 3, sizeof (c_dword));
    set32 (fparam + 4, GetPLCI());
    //set32 (fparam + 4, pChannel->GetPLCI());
    CallSupplServFacilityReq (GetPLCI(), fparam);
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_3PTY_End (CCAPI20_Channel * /*pChannel*/) {
    dhead ("CCAPI20_Facility::fReq_3PTY_End", DEBUG_CCAPI20_Facility);
    //dassert (pChannel != 0);
    dwarning (GetState() == cs_Holded);
    c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word) + sizeof (c_dword)];
    set8  (fparam,     7);
    set16 (fparam + 1, supplserv_3PTY_End);
    set8  (fparam + 3, sizeof (c_dword));
    set32 (fparam + 4, GetPLCI());
    //set32 (fparam + 4, pChannel->GetPLCI());
    CallSupplServFacilityReq (GetPLCI(), fparam);
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_CF_Activate (c_dword Controller, c_forwardtype Type, c_struct ServedUserNumber,
                                         c_struct ForwardToNumber) {
    dhead ("CCAPI20_Facility::fReq_CF_Activate", DEBUG_CCAPI20_Facility);
    if (GetState() >= cs_Registered) {
        tUInt    lenS = 9 + CalcCapiStructLen (ServedUserNumber) + CalcCapiStructLen (ForwardToNumber);
        tUInt    lenA = CapiStruct_AllocSize (lenS) + sizeof (c_word);
        c_struct data = CapiStruct_AllocStruct (lenA);
        if (!data) {
            fInd_CF_Activate (ci_Int_MemoryFullErr);
            return;
        }
        cp_facility cpfac;
        cpfac.FacilitySelector = facilselect_SupplServices;
        cpfac.FacilityParam    = data;

        CapiStruct_AddLen    (&data, lenA);
        CapiStruct_AddWord   (&data, supplserv_CF_Activate);
        CapiStruct_AddLen    (&data, lenS);
        CapiStruct_AddDWord  (&data, Handle);
        CapiStruct_AddWord   (&data, (c_word)Type);
        CapiStruct_AddWord   (&data, 0);
        CapiStruct_AddStruct (&data, ServedUserNumber);
        CapiStruct_AddStruct (&data, ForwardToNumber);
        CapiStruct_AddByte   (&data, 0);

        FACILITY_REQ (GetNextMsgNr(), Controller, &cpfac);
        CapiStruct_FreeStruct (cpfac.FacilityParam);
    } else {
        fInd_CF_Activate (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_CF_Deactivate (c_dword Controller, c_forwardtype Type, c_struct ServedUserNumber) {
    dhead ("CCAPI20_Facility::fReq_CF_Deactivate", DEBUG_CCAPI20_Facility);
    if (GetState() >= cs_Registered) {
        tUInt    lenS = 8 + CalcCapiStructLen (ServedUserNumber);
        tUInt    lenA = CapiStruct_AllocSize (lenS) + sizeof (c_word);
        c_struct data = CapiStruct_AllocStruct (lenA);
        if (!data) {
            fInd_CF_Deactivate (ci_Int_MemoryFullErr);
            return;
        }
        cp_facility cpfac;
        cpfac.FacilitySelector = facilselect_SupplServices;
        cpfac.FacilityParam    = data;

        CapiStruct_AddLen    (&data, lenA);
        CapiStruct_AddWord   (&data, supplserv_CF_Deactivate);
        CapiStruct_AddLen    (&data, lenS);
        CapiStruct_AddDWord  (&data, Handle);
        CapiStruct_AddWord   (&data, (c_word)Type);
        CapiStruct_AddWord   (&data, 0);
        CapiStruct_AddStruct (&data, ServedUserNumber);

        FACILITY_REQ (GetNextMsgNr(), Controller, &cpfac);
        CapiStruct_FreeStruct (cpfac.FacilityParam);
    } else {
        fInd_CF_Deactivate (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_CD (tBool PresentationAllowed, c_struct DeflectToNumber) {
    dhead ("CCAPI20_Facility::fReq_CD", DEBUG_CCAPI20_Facility);
    tUInt lenS = CalcCapiStructLen (DeflectToNumber) + sizeof (c_word) + sizeof (c_byte);
    tUInt lenA = CapiStruct_AllocSize (lenS) + sizeof (c_word);
    c_struct data = CapiStruct_AllocStruct (lenA);
    if (!data) {
        fInd_CD (ci_Int_MemoryFullErr);
    } else if (SetState (cs_CallDeflecting, cs_D_ConnectIndication) == vTrue) {
        cp_facility cpfac;
        cpfac.FacilitySelector = facilselect_SupplServices;
        cpfac.FacilityParam    = data;

        CapiStruct_AddLen    (&data, lenA);
        CapiStruct_AddWord   (&data, supplserv_CD);
        CapiStruct_AddLen    (&data, lenS);
        CapiStruct_AddWord   (&data, (PresentationAllowed == vTrue) ? (tUShort)1 : (tUShort)0);
        CapiStruct_AddStruct (&data, DeflectToNumber);
        CapiStruct_AddByte   (&data, 0);

        FACILITY_REQ (GetNextMsgNr(), GetPLCI(), &cpfac);
        CapiStruct_FreeStruct (cpfac.FacilityParam);
    } else {
        dwarning (0);
        fInd_CD (ci_Int_IllegalState);
    }
}

/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::fReq_MCID (void) {
    dhead ("CCAPI20_Facility::fReq_MCID", DEBUG_CCAPI20_Facility);
    c_byte fparam [2 * sizeof (c_byte) + sizeof (c_word)];
    set8  (fparam,     3);
    set16 (fparam + 1, supplserv_MCID);
    set8  (fparam + 3, 0);
    CallSupplServFacilityReq (GetPLCI(), fparam);
}

/*===========================================================================*\
\*===========================================================================*/

tBool CCAPI20_Facility::HandleSupplServicesInd (c_dword Address, c_supplserv Function, c_struct Params) {
    dhead ("CCAPI20_Facility::HandleSupplServicesInd", DEBUG_CCAPI20_Facility);
    dparams ("%x,%x", Address, Function);
    if (Address != GetPLCI()) {
        if (Address > CAPI_MAX_CONTROLLER_NUMBER) {
            // Address must be a PLCI or NCCI => This Address is definitly not for this class
            RETURN ('x', vFalse);
        }
    }
    {
        // TODO: fr forwarding, CCBS ist das hier zu allgemeingltig!
        c_byte frespStruct[6];
        set8  (frespStruct,     3);
        set16 (frespStruct + 1, (c_word)Function);
        set8  (frespStruct + 3, 0);
        cp_facility fresp = { facilselect_SupplServices, frespStruct };
        FACILITY_RESP (GetNextMsgNr(), Address, &fresp);
    }
    c_info Info = ci_NoError;
    if (CalcCapiStructDataLen (Params) >= sizeof (c_word)) {
        Info = (c_info)get16 (GetBeginOfStructData (Params));
        dprint ("Info=%x ", Info);
    }
    switch (Function) {
    case supplserv_Hold:
        if (CapiInfoIsSuccess (Info)) {
            if (SetState (cs_Holded, cs_HoldPending) == vTrue) {
                dassert (GetNCCI() == IllegalNCCI);
                fInd_Hold (Info);
            }
        } else if (SetState (cs_D_Connected, cs_HoldPending) == vTrue) {
            dassert (GetNCCI() == IllegalNCCI);
            ConnectB3Req();
            fInd_Hold (Info);
        }
        break;
    case supplserv_Retrieve:
        if (CapiInfoIsSuccess (Info)) {
            dassert (GetNCCI() == IllegalNCCI);
            if (SetState (cs_D_Connected, cs_RetrievePending) == vTrue) {
                ConnectB3Req();
                fInd_Retrieve (Info);
            }
        } else if (SetState (cs_Holded, cs_RetrievePending) == vTrue) {
            dassert (GetNCCI() == IllegalNCCI);
            fInd_Retrieve (Info);
        }
        break;
    case supplserv_Suspend:
        if (CapiInfoIsSuccess (Info)) {
            if (SetState (cs_D_DisconnectPending, cs_SuspendPending) == vTrue) {
                fInd_Suspend (Info);
            }
        } else if (SetState (cs_D_Connected, cs_SuspendPending) == vTrue) {
            ConnectB3Req();
            fInd_Suspend (Info);
        }
        break;
    case supplserv_Resume:
        if (CapiInfoIsSuccess (Info)) {
            if (SetState(cs_D_Connected, cs_ResumePending) == vTrue) {
                fInd_Resume (Info);
                if (GetState() == cs_D_Connected) {
                    ConnectB3Req();
                }
            }
        } else {
            DisconnectReq();
            fInd_Resume (Info);
        }
        break;
    case supplserv_ECT:
        if (GetState() == cs_Holded) {
            fInd_ECT (Info);
        }
        break;
    case supplserv_3PTY_Begin:
        if (GetState() == cs_Holded) {
            fInd_3PTY_Begin (Info);
        }
        break;
    case supplserv_3PTY_End:
        if (GetState() >= cs_D_Connected) {
            fInd_3PTY_End (Info);
        }
        break;
    case supplserv_CF_Activate: {
        c_struct data = 0;
        if (  (GetBeginAndLenOfStructData (Params, &data) >= sizeof (c_word) + sizeof (c_dword))
           && (get32 (data + sizeof (c_word)) == Handle)) {
            fInd_CF_Activate (Info);
        }
        break;
    }
    case supplserv_CF_Deactivate: {
        c_struct data = 0;
        if (  (GetBeginAndLenOfStructData (Params, &data) >= sizeof (c_word) + sizeof (c_dword))
           && (get32 (data + sizeof (c_word)) == Handle)) {
            fInd_CF_Deactivate (Info);
        }
        break;
    }
    case supplserv_CD:
        if (SetState ((CapiInfoIsError (Info) == vTrue) ? cs_D_ConnectIndication : cs_D_DisconnectPending,
                      cs_CallDeflecting) == vTrue) {
            fInd_CD (Info);
        }
        break;
    case supplserv_MCID:
        fInd_MCID (Info);
        break;
    //case supplserv_CCBS_Request:
    //case supplserv_CCBS_Deactivate:
    //case supplserv_CCBS_Interrogate:
    //case supplserv_CCBS_Call:
    //    dassert (CalcCapiStructDataLen (Params) >= sizeof (c_dword);
    //    fInd_CCBS_Call (*(c_word *)GetBeginOfStructData (Params));
    //    break;
    case supplserv_Hold_Notify:
        fInd_Hold_Notify();
        break;
    case supplserv_Retrieve_Notify:
        fInd_Retrieve_Notify();
        break;
    case supplserv_Suspend_Notify:
        fInd_Suspend_Notify();
        break;
    case supplserv_Resume_Notify:
        fInd_Resume_Notify();
        break;
    case supplserv_CallIsDiverting_Notify:
        fInd_CallIsDiverting_Notify();
        break;
    //case supplserv_DiversionActive_Notify:
    //    fInd_DiversionActive_Notify();
    //    break;
    case supplserv_ConferenceEstablished:
        fInd_ConferenceEstablished();
        break;
    case supplserv_ConferenceDisconnected:
        fInd_ConferenceDisconnected();
        break;
    //case supplserv_CCBS_EraseCallLinkID:
    //case supplserv_CCBS_Status:
    //case supplserv_CCBS_RemoteUserFree:
    //case supplserv_CCBS_B_Free:
    //case supplserv_CCBS_Erase:
    //case supplserv_CCBS_StopAlerting:
    //case supplserv_CCBS_InfoRetain:
    default:
        dprintWrn ("Unhandled SupplServInd(%x,a=%x) ", Function, Address);
        break;
    }
    RETURN ('x', vTrue);
}

#endif // #ifdef USE_SUPPLEMENTARY_SERVICES
/*===========================================================================*\
\*===========================================================================*/

void CCAPI20_Facility::FACILITY_CONF (c_word, c_dword Address, cp_facilityconf *Params) {
    dhead ("CCAPI20_Facility::FACILITY_CONF", DEBUG_CCAPI20_Facility);
    dassert (Params != 0);
    switch (Params->FacilitySelector) {
    case facilselect_DTMF: {
        c_info di = Params->Info;
        if (CapiInfoIsSuccess (di) == vTrue) {
            dassert (CalcCapiStructDataLen (Params->FacilityParam) >= sizeof (c_word));
            switch (get16 (GetBeginOfStructData (Params->FacilityParam))) {
            case 0:
                // nothing to do
                break;
            case 1:
                di = ci_Int_IncorrectDTMFDigit;
                break;
            default:
                dassert (0);
                // no Break!
            case 2:
                di = ci_Int_UnknownDTMFRequest;
                break;
            }
        }
        fConf_DTMF (di);
        break;
    }
#ifdef USE_SUPPLEMENTARY_SERVICES
    case facilselect_SupplServices: {
        c_struct data = 0;
        if (GetBeginAndLenOfStructData (Params->FacilityParam, &data) >= sizeof (c_word) + sizeof (c_byte)) {
            dassert (data != 0);
            c_supplserv Function = (c_supplserv)get16 (data);
            c_info      Info     = Params->Info;
            dprint ("Info=%x ", Info);
            dassert (CalcCapiStructDataLen (data + sizeof (c_word)) >= sizeof (c_word));
            if (  (CapiInfoIsSuccess (Info))
               && (GetBeginAndLenOfStructData (data + sizeof (c_word), &data) >= sizeof (c_word))) {
                dassert (data != 0);
                Info = (c_info)get16 (data);
            }
            dprint ("SupplServ=%x,Info=%x ", Function, Info);
            if (CapiInfoIsSuccess (Info) == vTrue) {
                if (Function == supplserv_Resume) {
                    SetPLCI (Address);
                }
            } else {
                // error handling
                switch (Function) {
                case supplserv_Listen:
                    fConf_Listen (Info);
                    break;
                case supplserv_Hold:
                    dassert (GetPLCI() == Address);
                    dassert (GetState() == cs_HoldPending);
                    if (SetState (cs_D_Connected, cs_HoldPending) == vTrue) {
                        ConnectB3Req();
                        fInd_Hold (Info);
                    }
                    break;
                case supplserv_Retrieve:
                    dassert (GetPLCI() == Address);
                    dassert (GetState() == cs_RetrievePending);
                    if (SetState (cs_Holded, cs_RetrievePending) == vTrue) {
                        fInd_Retrieve (Info);
                    }
                    break;
                case supplserv_Suspend:
                    dassert (GetPLCI() == Address);
                    dassert (GetState() == cs_SuspendPending);
                    if (SetState (cs_D_Connected, cs_SuspendPending) == vTrue) {
                        ConnectB3Req();
                        fInd_Suspend (Info);
                    }
                    break;
                case supplserv_Resume:
                    dassert (Address != IllegalPLCI);
                    if (SetState (cs_Registered, cs_ResumePending) == vTrue) {
                        fInd_Resume (Info);
                    }
                    break;
                case supplserv_ECT:
                    fInd_ECT (Info);
                    break;
                case supplserv_3PTY_Begin:
                    fInd_3PTY_Begin (Info);
                    break;
                case supplserv_3PTY_End:
                    fInd_3PTY_End (Info);
                    break;
                case supplserv_CF_Activate:
                    fInd_CF_Activate (Info);
                    break;
                case supplserv_CF_Deactivate:
                    fInd_CF_Deactivate (Info);
                    break;
                case supplserv_CD:
                    dassert (GetPLCI() == Address);
                    dassert (GetState() == cs_CallDeflecting);
                    if (SetState (cs_D_ConnectIndication, cs_CallDeflecting) == vTrue) {
                        fInd_CD (Info);
                    }
                    break;
                case supplserv_MCID:
                    fInd_MCID (Info);
                    break;
                case supplserv_CCBS_Request:
                    //fInd_CCBS_Request (Info);
                    break;
                case supplserv_CCBS_Deactivate:
                    //fInd_CCBS_Deactivate (Info);
                    break;
                case supplserv_CCBS_Interrogate:
                    //fInd_CCBS_Interrogate (Info);
                    break;
                case supplserv_CCBS_Call:
                    //fInd_CCBS_Call (Info);
                    break;
                default:
                    dprintErr ("Illegal SupplServ=%x ", Function);
                    dassert (0);
                    break;
                }
            }
        } else {
            dassert (0);
        }
        break;
    }
#endif // #ifdef USE_SUPPLEMENTARY_SERVICES
    case facilselect_PMWakeup:
        dprint ("FP=%x ", Params->FacilityParam);
        dassert (  (CapiInfoIsSuccess (Params->Info) == vFalse)
                || (CalcCapiStructDataLen (Params->FacilityParam) >= sizeof (c_word)));
        fConf_PMWakeup (Address, Params->Info, (CapiInfoIsSuccess (Params->Info) == vTrue) ?
                        get16 (GetBeginOfStructData (Params->FacilityParam)) : (tUShort)0);
        break;
    default:
        dwarning (0);
        break;
    }
}

/*===========================================================================*\
\*===========================================================================*/

tBool CCAPI20_Facility::FACILITY_IND (c_word, c_dword Address, cp_facility *Params) {
    dhead ("CCAPI20_Facility::FACILITY_IND", DEBUG_CCAPI20_Facility);
    dassert (Params != 0);
    tBool fret = vFalse;
    switch (Params->FacilitySelector) {
    case facilselect_DTMF:
        if ((Address == GetPLCI()) && (CalcCapiStructDataLen (Params->FacilityParam) > 0)) {
            cp_facility fresp = { facilselect_DTMF, 0 };
            FACILITY_RESP (GetNextMsgNr(), Address, &fresp);
            fInd_DTMF (Params->FacilityParam);
            fret = vTrue;
        }
        break;
#ifdef USE_SUPPLEMENTARY_SERVICES
    case facilselect_SupplServices: {
        c_struct data = 0;
        if (GetBeginAndLenOfStructData (Params->FacilityParam, &data) >= sizeof (c_word) + sizeof (c_byte)) {
            dassert (data != 0);
            fret = HandleSupplServicesInd (Address, (c_supplserv)get16 (data), data + sizeof (c_word));
        }
        break;
    }
#endif
    default:
        // illegal Selector
        dwarning (0);
        // no break!
    case facilselect_Handset:
    case facilselect_V42Compression:
        // not supported now
        break;
    }
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/
