/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "MSNList.h"
#include "ConvCapi.h"
#include "DynArray.h"
#include "aString.h"
#include "dbgCAPI.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

class CControllerInfo : public CSortPListElement {
public:
    CControllerInfo (void);
    virtual ~CControllerInfo (void);

    void DelAllMSNs (void);
    tCompare Compare (void *RefCompData);
    void *GetRefCompData (void);

    tUInt         Controller;
    tUInt         InfoMask;
    tUInt         CIPMask;
    tUChar       *pTelNumPrefix;
    tUInt         PrefixLen;
    tUInt         DDILen;
    CDynamicArray MSNList;
    tBool         PMSupportable;
};


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

struct MSNList_ArrayFindInfo {
    tUInt   len;
    tUChar *num;
};

struct CntrlInfo_MSNRange {
    c_struct aMSN;
    c_struct bMSN;
};


/*===========================================================================*\
\*===========================================================================*/

static tCompare ArrayFindFunc_NumComp (c_struct Num1, c_struct Num2) {
    dhead ("ArrayFindFunc_NumComp", DEBUG_CCntrlMSNList);
    dassert (CalcCapiStructDataLen (Num1) > 1);
    tCompare fret = vHigher;
    if (Num2) {
        dassert (CalcCapiStructDataLen (Num2) > 1);
        tUInt len = GetBeginAndLenOfStructData (Num1, &Num1) - 1;
        fret = a_compare (len, GetBeginAndLenOfStructData (Num2, &Num2) - 1);
        if (fret == vEqual) {
            dprint ("cmp=(%s/%s) ", Num1 + 1, Num2 + 1);
            fret = s_strncmp (Num1 + 1, Num2 + 1, len);
        }
    }
    RETURN ('x', fret);
}

static tCompare ArrayFindFunc_MSNComp (c_struct Num1, MSNList_ArrayFindInfo *Num2) {
    dhead ("ArrayFindFunc_MSNComp", DEBUG_CCntrlMSNList);
    dassert (CalcCapiStructDataLen (Num1) > 1);
    dassert (Num2 != 0);
    dassert (Num2->len != 0);
    dassert (Num2->num > 0);
    tUInt len = GetBeginAndLenOfStructData (Num1, &Num1) - 1;
    Num1     += len;
    if (len > Num2->len) {
        len = Num2->len;
    }
    dprint ("cmp=%s<->%s. ", Num1 - len + 1, Num2->num + Num2->len - len);
    RETURN ('x', s_strncmp (Num1 - len + 1, Num2->num + Num2->len - len, len));
}

static tCompare ArrayFindFunc_DDIComp (c_struct Num1, MSNList_ArrayFindInfo *Num2) {
    dhead ("ArrayFindFunc_NumComp", DEBUG_CCntrlMSNList);
    dassert (CalcCapiStructDataLen (Num1) > 1);
    dassert (Num2 != 0);
    dassert (Num2->len != 0);
    dassert (Num2->num > 0);
    tUInt    len  = GetBeginAndLenOfStructData (Num1, &Num1) - 1;
    tCompare fret = a_compare (len, Num2->len);
    if (fret != vHigher) {
        dprint ("cmp=(%s/%s) ", Num1 + 1, Num2->num);
        fret = s_strncmp (Num1 + 1, Num2->num, len);
    }
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

static void Free_CntrlInfo_MSNRange (CntrlInfo_MSNRange *pRange) {
    dhead ("Free_CntrlInfo_MSNRange", DEBUG_CCntrlMSNList);
    dassert (pRange != 0);
    Conv_DuplicateStruct (&pRange->aMSN, 0);
    Conv_DuplicateStruct (&pRange->bMSN, 0);
}

static tSInfo Alloc_CntrlInfo_MSNRange (CntrlInfo_MSNRange *pRange, tUChar *aMSN, tUChar *bMSN) {
    dhead ("Free_CntrlInfo_MSNRange", DEBUG_CCntrlMSNList);
    dassert (pRange != 0);
    dassert (s_strnull (aMSN) == vFalse);
    if (ConvUnion_CalledNumber (&pRange->aMSN, aMSN) == 0) {
        RETURN ('x', iErr_OutOfMemory);
    }
    dassert (pRange->aMSN != 0);
    if (bMSN) {
        dassert (s_strempty (bMSN) == vFalse);
        if (ConvUnion_CalledNumber (&pRange->bMSN, bMSN) == 0) {
            Conv_DuplicateStruct (&pRange->aMSN, 0);
            RETURN ('x', iErr_OutOfMemory);
        }
        dassert (pRange->bMSN != 0);
        if (  (CalcCapiStructLen (pRange->aMSN) != CalcCapiStructLen (pRange->bMSN))
           || (ArrayFindFunc_NumComp (pRange->aMSN, pRange->bMSN) != vLower)) {
            Free_CntrlInfo_MSNRange (pRange);
            RETURN ('x', iErr_MSNList_InvalidValue);
        }
    }
    RETURN ('x', i_Okay);
}


/*===========================================================================*\
\*===========================================================================*/

static tResult ArrayFindFunc_Range (const void *data, const void *ItemFrag) {
    dhead ("ArrayFindFunc_Range", DEBUG_CCntrlMSNList);
    dparams ("%x,%x", data, ItemFrag);
    dassert (data != 0);
    dassert (CalcCapiStructDataLen (((CntrlInfo_MSNRange *)ItemFrag)->aMSN) > 1);
    // sorting criterion: 1.) length ; 2.) digits
    switch (ArrayFindFunc_NumComp (((CntrlInfo_MSNRange *)data)->aMSN,
                                   ((CntrlInfo_MSNRange *)ItemFrag)->aMSN)) {
    case vEqual:
        // Equal values are not allowed!
        dwarning (0);
        RETURN ('x', vError);
    case vHigher:
        if (ArrayFindFunc_NumComp (((CntrlInfo_MSNRange *)data)->aMSN,
                                   ((CntrlInfo_MSNRange *)ItemFrag)->bMSN) != vHigher) {
            // Equal values or overlapping ranges are not allowed!
            dwarning (0);
            RETURN ('x', vError);
        }
        RETURN ('x', vCancel);
    case vLower:
        if (  ((CntrlInfo_MSNRange *)data)->bMSN
           && (ArrayFindFunc_NumComp (((CntrlInfo_MSNRange *)data)->bMSN,
                                      ((CntrlInfo_MSNRange *)ItemFrag)->aMSN) != vLower)) {
            // Equal values or overlapping ranges are not allowed!
            dwarning (0);
            RETURN ('x', vError);
        }
        break;
    }
    RETURN ('x', vNo);
}


/*===========================================================================*\
\*===========================================================================*/

static tResult ArrayFindFunc_MSN (const void *data, const void *ItemFrag) {
    dhead ("ArrayFindFunc_MSN", DEBUG_CCntrlMSNList);
    dparams ("%x,%x", data, ItemFrag);
    dassert (data != 0);
    dassert (ItemFrag != 0);
    switch (ArrayFindFunc_MSNComp (((CntrlInfo_MSNRange *)data)->aMSN, (MSNList_ArrayFindInfo *)ItemFrag)) {
    case vEqual:
        RETURN ('x', vYes);
    case vHigher:
        //RETURN ('x', vCancel);
        // because of the length-specific sorting we can not stop here
        break;
    case vLower:
        if (((CntrlInfo_MSNRange *)data)->bMSN) {
            switch (ArrayFindFunc_MSNComp (((CntrlInfo_MSNRange *)data)->bMSN, (MSNList_ArrayFindInfo *)ItemFrag)) {
            case vEqual:
            case vHigher:
                RETURN ('x', vYes);
            case vLower:
                break;
            }
        }
        break;
    }
    RETURN ('x', vNo);
}


/*===========================================================================*\
\*===========================================================================*/

static tResult ArrayFindFunc_DDI (const void *data, const void *ItemFrag) {
    dhead ("ArrayFindFunc_DDI", DEBUG_CCntrlMSNList);
    dparams ("%x,%x", data, ItemFrag);
    dassert (data != 0);
    dassert (ItemFrag != 0);
    // Compare length and digits and ignore subaddresstype of MSN.
    switch (ArrayFindFunc_DDIComp (((CntrlInfo_MSNRange *)data)->aMSN, (MSNList_ArrayFindInfo *)ItemFrag)) {
    case vHigher:
        RETURN ('x', vCancel);
    case vEqual:
        RETURN ('x', vYes);
    case vLower:
        if (((CntrlInfo_MSNRange *)data)->bMSN) {
            switch (ArrayFindFunc_DDIComp (((CntrlInfo_MSNRange *)data)->bMSN, (MSNList_ArrayFindInfo *)ItemFrag)) {
            case vHigher:
                RETURN ('x', vYes);
            case vEqual:
                RETURN ('x', vYes);
                break;
            case vLower:
                break;
            }
        }
        break;
    }
    RETURN ('x', vNo);
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::Set (tUInt Controller, tUInt InfoMask, tUInt CIPMask) {
    dhead ("CCntrlMSNList::Set", DEBUG_CCntrlMSNList);
    dassert (Controller > 0);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (!pE) {
        dprint ("New ControllerInfo! ");
        if (!CIPMask && !InfoMask) {
            RETURN ('x', vTrue);
        }
        pE = new CControllerInfo;
        if (!pE) {
            RETURN ('x', vFalse);
        }
        pE->Controller = Controller;
        CSortPointerList::AddFromLow (pE);
    }
    if ((pE->DDILen > 0) || (pE->PrefixLen > 0)) {
        InfoMask |= infomask_CalledPartyNum | infomask_SendingComplete;
    }
    pE->CIPMask  = CIPMask;
    pE->InfoMask = InfoMask;
    RETURN ('x', vTrue);
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::Add (tUInt Controller, tUInt InfoMask, tUInt CIPMask) {
    dhead ("CCntrlMSNList::Add", DEBUG_CCntrlMSNList);
    dassert (Controller > 0);
    if (!CIPMask && !InfoMask) {
        return vTrue;
    }
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (!pE) {
        dprint ("New ControllerInfo! ");
        pE = new CControllerInfo;
        if (!pE) {
            RETURN ('x', vFalse);
        }
        pE->Controller = Controller;
        CSortPointerList::AddFromLow (pE);
    }
    pE->CIPMask  |= CIPMask;
    pE->InfoMask |= InfoMask;
    RETURN ('x', vTrue);
}

/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::SetDDI (tUInt Controller, tUChar *pTelNumPrefix, tUInt MaxDDILen) {
    dhead ("CCntrlMSNList::SetDDI", DEBUG_CCntrlMSNList);
    dassert (Controller > 0);
    tUChar *tmpTelNumPrefix = 0;
    tUInt   tmpPrefixLen    = 0;
    if (pTelNumPrefix) {
        tmpPrefixLen = s_strlen (pTelNumPrefix);
        if (tmpPrefixLen) {
            s_strdup (&tmpTelNumPrefix, pTelNumPrefix, tmpPrefixLen);
            if (!tmpTelNumPrefix) {
                RETURN ('x', vFalse);
            }
        }
    }
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (!pE) {
        dprint ("New ControllerInfo! ");
        if (!MaxDDILen && !tmpPrefixLen) {
            RETURN ('x', vTrue);
        }
        pE = new CControllerInfo;
        if (!pE) {
            s_strdel (&tmpTelNumPrefix);
            RETURN ('x', vFalse);
        }
        pE->Controller = Controller;
        CSortPointerList::AddFromLow (pE);
    }
    if (MaxDDILen || tmpPrefixLen) {
        // SET DDI
        pE->InfoMask |= infomask_CalledPartyNum | infomask_SendingComplete;
    } else {
        // CLEAR DDI
        // TODO?: perhaps somebody has set the infomask-flags externally
        pE->InfoMask &= ~(infomask_CalledPartyNum | infomask_SendingComplete);
    }
    dprint ("Prefix(%x)=%s,DDILen=%x ", tmpPrefixLen, tmpTelNumPrefix, MaxDDILen);
    pE->pTelNumPrefix = tmpTelNumPrefix;
    pE->PrefixLen     = tmpPrefixLen;
    pE->DDILen        = MaxDDILen;
    RETURN ('x', vTrue);
}


/*===========================================================================*\
\*===========================================================================*/

tSInfo CCntrlMSNList::SetMSN (tUInt Controller, tUChar *aMSN, tUChar *bMSN) {
    dhead ("CCntrlMSNList::SetMsn", DEBUG_CCntrlMSNList);
    dparams ("%x,%s,%s", Controller, aMSN, bMSN);
    dassert (Controller > 0);
    if (s_strnull (aMSN) == vTrue) {
        aMSN = 0;
    }
    if (s_strnull (bMSN) == vTrue) {
        bMSN = 0;
    }
    dassert ((aMSN != 0) || (bMSN == 0)); // (aMSN == 0 && bMSN != 0) is not allowed
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (!pE) {
        if (!aMSN) {
            RETURN ('x', i_Okay);
        }
        pE = new CControllerInfo;
        if (!pE) {
            RETURN ('x', iErr_OutOfMemory);
        }
        pE->Controller = Controller;
        CSortPointerList::AddFromLow (pE);
    } else {
        pE->DelAllMSNs();
    }

    if (aMSN) {
        CntrlInfo_MSNRange newRange = { 0, 0 };
        tSInfo fret = Alloc_CntrlInfo_MSNRange (&newRange, aMSN, bMSN);
        if (infoError (fret)) {
            dwarning (0);
            RETURN ('x', fret);
        }
        if (pE->MSNList.Append (&newRange) == vFalse) {
            dwarning (0);
            Free_CntrlInfo_MSNRange (&newRange);
            RETURN ('x', iErr_OutOfMemory);
        }
        dassert (CalcCapiStructDataLen (newRange.aMSN) > 1); // no empty MSN's allowed
    }
    RETURN ('x', i_Okay);
}

/*===========================================================================*\
\*===========================================================================*/

tSInfo CCntrlMSNList::AddMSN (tUInt Controller, tUChar *aMSN, tUChar *bMSN) {
    dhead ("CCntrlMSNList::AddMsn", DEBUG_CCntrlMSNList);
    dparams ("%x,%s,%s", Controller, aMSN, bMSN);
    dassert (Controller > 0);
    dassert (s_strnull (aMSN) == vFalse);
    if (s_strnull (bMSN) == vTrue) {
        bMSN = 0;
    }
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (!pE) {
        pE = new CControllerInfo;
        if (!pE) {
            RETURN ('x', iErr_OutOfMemory);
        }
        pE->Controller = Controller;
        CSortPointerList::AddFromLow (pE);
    }

    // MSNs are added to a sortedArray
    tUInt fPos;
    CntrlInfo_MSNRange newRange = { 0, 0 };
    tSInfo fret = Alloc_CntrlInfo_MSNRange (&newRange, aMSN, bMSN);
    if (infoError (fret)) {
        dwarning (0);
        RETURN ('x', fret);
    }
    dassert (CalcCapiStructDataLen (newRange.aMSN) > 1); // no empty MSN's allowed
    switch (pE->MSNList.Find (ArrayFindFunc_Range, &newRange, 0, &fPos)) {
    case vNo:
        // Add at the end
        fPos = pE->MSNList.CountAllItems();
        /*----- no Break! -----*/
    case vCancel:
        dprint ("InsPos=%x ", fPos);
        if (pE->MSNList.Ins (fPos, &newRange) == vFalse) {
            dwarning (0);
            Free_CntrlInfo_MSNRange (&newRange);
            RETURN ('x', iErr_OutOfMemory);
        }
        break;
    default:
        dassert (0);
        // no break
    case vError:
        // overlapping ranges found
        dwarning (0);
        Free_CntrlInfo_MSNRange (&newRange);
        RETURN ('x', iErr_MSNList_MSNOverlap);
    case vYes:
        // dupplicate Entry found
        dwarning (0);
        Free_CntrlInfo_MSNRange (&newRange);
        RETURN ('x', iErr_MSNList_MSNEqual);
    }
    RETURN ('x', vTrue);
}


/*===========================================================================*\
\*===========================================================================*/

void CCntrlMSNList::Remove (tUInt Controller) {
    dhead ("CCntrlMSNList::Remove", DEBUG_CCntrlMSNList);
    dparams ("%x", Controller);
    dassert (Controller > 0);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (pE) {
        delete pE;
    }
}


/*===========================================================================*\
\*===========================================================================*/

tSInfo CCntrlMSNList::FindCallInfo (tUInt Controller, c_cipvalue CIPValue, c_struct MSN, tUInt *pDDIPrefixLen) {
    dhead ("CCntrlMSNList::FindCallInfo", DEBUG_CCntrlMSNList);
    dparams ("%x,%x", Controller, CIPValue);
    dassert (Controller > 0);
    dassert (pDDIPrefixLen != 0);
    *pDDIPrefixLen = 0;
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    dassert (pE != 0);
    if (pE && ((1 << CIPValue) & pE->CIPMask)) {
        dprint ("Found Cntrl with valid CIPValue! ");
        if (pE->InfoMask & infomask_CalledPartyNum) {
            dwarning (pE->PrefixLen > 0);
            if ((pE->DDILen == 0) && (pE->MSNList.ExistUsedItems() == vFalse)) {
                dassert (0);
                RETURN ('x', iWrn_MSNList_RejectCall);
            }
            *pDDIPrefixLen = 1; // Minimum for ignoring the 0xA0 Byte
            c_byte *pData  = 0;
            tUInt   len    = GetBeginAndLenOfStructData (MSN, &pData);
            dprint ("-Handle DDI(%x)- ", len);
            dwarning (pE->PrefixLen < len);
            if ((len > 1) && (pE->PrefixLen < len)) {
                c_byte *pFind = s_findstring (pData + 1, len - 1, pE->pTelNumPrefix, pE->PrefixLen);
                // calculate len for DDI Number Prefix
                dwarning (pFind != 0);
                if (pFind) {
                    dassert (pFind > pData);
                    *pDDIPrefixLen = pFind - pData + pE->PrefixLen;
                    dprint ("DDIPrefixLen=%x ", *pDDIPrefixLen);
                }
            }
            //return (len > *pDDIPrefixLen) ? TestDDINumber (Controller, pData + *pDDIPrefixLen, len - *pDDIPrefixLen) :
            //                                iWrn_MSNList_DDICall;
            RETURN ('x', iWrn_MSNList_DDICall);
        }
        MSNList_ArrayFindInfo fi;
        fi.len = GetBeginAndLenOfStructData (MSN, &fi.num);
        dprint ("MSN_len %d ", fi.len);
        // Global Call
        if (fi.len <= 1) {
            dprint ("GlobalCall found %s ", AllowGlobalCall ? "OK" : "REJ");  
            if (AllowGlobalCall) {
                RETURN ('x', i_Okay);
            } else {
                RETURN ('x', iWrn_MSNList_RejectCall);
            }
        }
        // Empty List
        if (pE->MSNList.ExistUsedItems() == vFalse) {
            RETURN ('x', i_Okay);
        }
        // ignore callled party number subaddress
        fi.len--;
        fi.num++;
        if (pE->MSNList.Find (ArrayFindFunc_MSN, &fi) == vYes) {
            RETURN ('x', i_Okay);
        }
    }
    RETURN ('x', iWrn_MSNList_RejectCall);
}


/*===========================================================================*\
\*===========================================================================*/

tSInfo CCntrlMSNList::TestDDINumber (tUInt Controller, tUChar *DDINumber, tUInt DDILen) {
    dhead ("CCntrlMSNList::TestDDINumber", DEBUG_CCntrlMSNList);
    dparams ("%x,%s,%x", Controller, DDINumber, DDILen);
    dassert (Controller > 0);
    dassert (DDINumber != 0);
    dassert (s_strempty (DDINumber) == vFalse);
    tSInfo           fret = iWrn_MSNList_RejectCall;
    CControllerInfo *pE   = (CControllerInfo *)FindHigher (&Controller);
    dwarning (pE != 0);
    if (pE) {
        dassert (pE->InfoMask & infomask_CalledPartyNum);
        dwarning (pE->DDILen > 0);
        dprint ("%x,%x,%x ", pE->DDILen, DDILen, pE->MSNList.ExistUsedItems());
        if (pE->MSNList.ExistUsedItems() == vFalse) {
            fret = (DDILen >= pE->DDILen) ? i_Okay : iWrn_MSNList_DDICall;
        } else {
            MSNList_ArrayFindInfo fi = { DDILen, DDINumber };
            fret = (pE->MSNList.Find (ArrayFindFunc_DDI, &fi) == vYes) ? i_Okay : iWrn_MSNList_DDICall;
        }
    }
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::GetMask (tUInt Controller, tUInt *pInfoMask, tUInt *pCIPMask) {
    dhead ("CCntrlMSNList::BuildMask", DEBUG_CCntrlMSNList);
    dassert (Controller != IllegalController);
    dassert (pCIPMask != 0);
    dassert (pInfoMask != 0);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (!pE) {
        RETURN ('x', vFalse);
    }
    if ((pE->InfoMask & infomask_CalledPartyNum) && (pE->DDILen == 0) && (pE->MSNList.ExistUsedItems() == vFalse)) {
        dwarning (0);
        RETURN ('x', vFalse);
    }
    *pCIPMask  = pE->CIPMask;
    *pInfoMask = pE->InfoMask;
    //if (!*pCIPMask && !*pInfoMask) {
    //    pCurPos = (CControllerInfo *)pE->GetPrev();
    //    delete pE;
    //}
    RETURN ('x', vTrue);
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::GetNextMask (tUInt *pController, tUInt *pInfoMask, tUInt *pCIPMask) {
    dhead ("CCntrlMSNList::BuildNextMask", DEBUG_CCntrlMSNList);
    dassert (pController != 0);
    dassert (pCIPMask != 0);
    dassert (pInfoMask != 0);
    tBool fret          = vFalse;
    CControllerInfo *pE = pCurPos;
    pE = (CControllerInfo *)((pE != 0) ? pE->GetHigher() : GetLowest());
    if (pE) {
        while (  (pE->InfoMask & infomask_CalledPartyNum)
              && (pE->DDILen == 0)
              && (pE->MSNList.ExistUsedItems() == vFalse)) {
            dwarning (0);
            pE = (CControllerInfo *)pE->GetHigher();
            if (pE == 0) {
                pCurPos = 0;
                RETURN ('x', fret);
            }
        }
        dassert (pE != 0);
        *pController = pE->Controller;
        *pCIPMask    = pE->CIPMask;
        *pInfoMask   = pE->InfoMask;
        fret         = vTrue;
        dprint ("Cntrl=%x,Info=%x,CIP=%x ", *pController, *pInfoMask, *pCIPMask);
    }
    pCurPos = pE;
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/

tUChar *CCntrlMSNList::GetMSN (tUInt Controller, tUInt pos) {
    dhead ("CCntrlMSNList::GetMSN", DEBUG_CCntrlMSNList);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    dwarning (pE != 0);
    if (pE) {
        dassert (pE->MSNList.CountUsedItems() > pos);
        CntrlInfo_MSNRange MSNRange;
        if (pE->MSNList.Get (pos, &MSNRange) == vTrue) {
            dassert (CalcCapiStructDataLen (MSNRange.aMSN) > 1);
            dwarning (MSNRange.bMSN == 0);
            RETURN ('x', GetBeginOfStructData (MSNRange.aMSN) + 1);
        } else {
            dwarning (0);
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
\*===========================================================================*/

c_struct CCntrlMSNList::GetStructMSN (tUInt Controller, tUInt pos) {
    dhead ("CCntrlMSNList::GetStructMSN", DEBUG_CCntrlMSNList);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    dwarning (pE != 0);
    if (pE) {
        CntrlInfo_MSNRange MSNRange;
        dassert (pE->MSNList.CountUsedItems() > pos);
        if (pE->MSNList.Get (pos, &MSNRange) == vTrue) {
            dassert (CalcCapiStructDataLen (MSNRange.aMSN) > 1);
            dwarning (MSNRange.bMSN == 0);
            RETURN ('x', MSNRange.aMSN);
        } else {
            dwarning (0);
        }
    }
    RETURN ('x', 0);
}


/*===========================================================================*\
\*===========================================================================*/

tUInt CCntrlMSNList::GetCountMSN (tUInt Controller) {
    dhead ("CCntrlMSNList::GetCountMSN", DEBUG_CCntrlMSNList);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    RETURN ('x', (pE) ? pE->MSNList.CountUsedItems() : 0);
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::ExistMSNRange (tUInt Controller) {
    dhead ("CCntrlMSNList::ExistMSNRange", DEBUG_CCntrlMSNList);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    dwarning (pE != 0);
    if (pE) {
        CntrlInfo_MSNRange MSNRange;
        tUInt pos = pE->MSNList.CountUsedItems();
        while (pos > 0) {
            pos--;
            if (  (pE->MSNList.Get (pos, &MSNRange) == vTrue)
               && (MSNRange.bMSN != 0)) {
                RETURN ('x', vTrue);
            }
        }
    }
    RETURN ('x', vFalse);
}


/*===========================================================================*\
\*===========================================================================*/

void CCntrlMSNList::SetWakeupSupport (tUInt Controller, tBool IsSupported) {
    dhead ("CCntrlMSNList::SetWakeupSupport", DEBUG_CCntrlMSNList);
    dassert (Controller > 0);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    if (pE) {
        pE->PMSupportable = IsSupported;
    }
}


/*===========================================================================*\
\*===========================================================================*/

tBool CCntrlMSNList::GetWakeupSupport (tUInt Controller) {
    dhead ("CCntrlMSNList::GetWakeupSupport", DEBUG_CCntrlMSNList);
    dassert (Controller > 0);
    CControllerInfo *pE = (CControllerInfo *)FindHigher (&Controller);
    RETURN ('x', (pE) ? pE->PMSupportable : vFalse);
}


/*===========================================================================*\
    CControllerInfo - Funktionen
\*===========================================================================*/

CControllerInfo::CControllerInfo (void)
  : Controller (0),
    InfoMask (0),
    CIPMask (0),
    pTelNumPrefix (0),
    PrefixLen (0),
    DDILen (0),
    MSNList (1, sizeof (CntrlInfo_MSNRange)),
    PMSupportable (vTrue) {

    dhead ("CControllerInfo-Constructor", DEBUG_CCntrlMSNList);
}

CControllerInfo::~CControllerInfo (void) {
    dhead ("CControllerInfo-Destructor", DEBUG_CCntrlMSNList);
    DelAllMSNs();
    s_strdel (&pTelNumPrefix);
}

void CControllerInfo::DelAllMSNs (void) {
    dhead ("CControllerInfo::DelAllMSNs", DEBUG_CCntrlMSNList);
    CntrlInfo_MSNRange MSNRange = { 0, 0 };
    tUInt              fret;
    while ((fret = MSNList.DelLast (&MSNRange)) > 0) {
        if (fret == 2) {
            Free_CntrlInfo_MSNRange (&MSNRange);
        }
    }
}

tCompare CControllerInfo::Compare (void *RefCompData) {
    dhead ("CControllerInfo::Compare", DEBUG_CCntrlMSNList);
    dparams ("%x,%x", Controller, RefCompData);
    return (*((tUInt *)RefCompData) < Controller) ? vLower : (*((tUInt *)RefCompData) == Controller) ? vEqual : vHigher;
}

void *CControllerInfo::GetRefCompData (void) {
    dhead ("CControllerInfo::GetRefCompData", DEBUG_CCntrlMSNList);
    return &Controller;
}


/*===========================================================================*\
\*===========================================================================*/

