/*---------------------------------------------------------------------------*\
 * Modul : ProtClas.h vom Freitag, 8. Mrz 2002 13:44 von MRolf
 *
 * $Id: ProtClas.h,v 1.1 2002/10/11 08:35:29 mrolf Exp $
 *
 * $Log: ProtClas.h,v $
 * Revision 1.1  2002/10/11 08:35:29  mrolf
 * Some new files and ProtVal.h is not needed anymore
 *
 * Revision 1.2  2002/08/08 14:18:12Z  mrolf
 * - MRolf: UniqueStart/Stop-Methoden implementiert
 *          Templates rausgeflogen
 * Revision 1.1  2002/07/05 08:26:01Z  MRolf
 * Initial revision
 * Revision 1.3  2002/06/07 10:55:40Z  MRolf
 * - MRolf: TestAndSet, Inc & Dec liegen auch als unprotected Version vor
\*---------------------------------------------------------------------------*/
#ifndef _PROTCLAS_H_
#define _PROTCLAS_H_

#include "Protect.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define UNIQUE_INIT_VALUE       0

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tBool basicTestAndSet (tUInt *pVar, tUInt testVal, tUInt setVal) {
    if (*pVar != testVal) {
        return vFalse;
    }
    *pVar = setVal;
    return vTrue;
}

inline tUInt basicInc (tUInt *pVar) {
    tUInt fret = *pVar;
    (*pVar)++;
    return fret;
}

inline tUInt basicDec (tUInt *pVar) {
    tUInt fret = *pVar;
    (*pVar)--;
    return fret;
}

inline tBool basicUniqueStart (tUInt *pCounter) {
    return ((*pCounter)++) == 0;
}

inline tBool basicUniqueStop (tUInt *pCounter) {
    return ((*pCounter)--) == 1;
}


/*===========================================================================*\
\*===========================================================================*/

class CProtectClass {
public:
    CProtectClass (void);
    virtual ~CProtectClass (void);
    tBool Init (void);  // Init can be called multiple times!!

    void BeginRead (void);
    void EndRead (void);
    void BeginWrite (void);
    void EndWrite (void);

    // Set, Add, Sub, Inc & Dec return always the old/previous value of pVar!!
    tUInt Get (tUInt *pVar);
    tUInt Set (tUInt *pVar, tUInt value);
    tBool TestAndSet (tUInt *pVar, tUInt testVal, tUInt setVal);

    tUInt Add (tUInt *pVar, tUInt value);
    tUInt Sub (tUInt *pVar, tUInt value);

    tUInt Inc (tUInt *pVar);
    tUInt Dec (tUInt *pVar);

    tBool UniqueStart (tUInt *pCounter);
    tBool UniqueStop (tUInt *pCounter);

private:
    Protect_Define (m_hProtect);
};


/*===========================================================================*\
    CProtectClass - inline functions
\*===========================================================================*/

inline CProtectClass::CProtectClass (void) {
    Protect_InitHandle (&m_hProtect);
    Protect_Create (&m_hProtect);
}

inline CProtectClass::~CProtectClass (void) {
    Protect_Destroy (&m_hProtect);
}

inline tBool CProtectClass::Init (void) {
    return Protect_IsCreated (m_hProtect) ? vTrue : Protect_Create (&m_hProtect);
}

inline void CProtectClass::BeginRead (void) {
    Protect_BeginRead (m_hProtect);
}

inline void CProtectClass::EndRead (void) {
    Protect_EndRead (m_hProtect);
}

inline void CProtectClass::BeginWrite (void) {
    Protect_BeginWrite (m_hProtect);
}

inline void CProtectClass::EndWrite (void) {
    Protect_EndWrite (m_hProtect);
}


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tUInt CProtectClass::Get (tUInt *pVar) {
    BeginRead();
    tUInt fret = *pVar;
    EndRead();
    return fret;
}

inline tUInt CProtectClass::Set (tUInt *pVar, tUInt value) {
    BeginWrite();
    tUInt fret = *pVar;
    *pVar  = value;
    EndWrite();
    return fret;
}

inline tBool CProtectClass::TestAndSet (tUInt *pVar, tUInt testVal, tUInt setVal) {
    tBool fret = vFalse;
    BeginWrite();
    if (*pVar == testVal) {
        *pVar = setVal;
        fret  = vTrue;
    }
    EndWrite();
    return fret;
}

inline tUInt CProtectClass::Add (tUInt *pVar, tUInt value) {
    BeginWrite();
    tUInt fret = *pVar;
    *pVar += value;
    EndWrite();
    return fret;
}

inline tUInt CProtectClass::Sub (tUInt *pVar, tUInt value) {
    BeginWrite();
    tUInt fret = *pVar;
    *pVar -= value;
    EndWrite();
    return fret;
}

inline tUInt CProtectClass::Inc (tUInt *pVar) {
    BeginWrite();
    tUInt fret = *pVar;
    (*pVar)++;
    EndWrite();
    return fret;
}

inline tUInt CProtectClass::Dec (tUInt *pVar) {
    BeginWrite();
    tUInt fret = *pVar;
    (*pVar)--;
    EndWrite();
    return fret;
}

inline tBool CProtectClass::UniqueStart (tUInt *pCounter) {
    return Inc (pCounter) == 0;
}

inline tBool CProtectClass::UniqueStop (tUInt *pCounter) {
    return Dec (pCounter) == 1;
}


/*===========================================================================*\
\*===========================================================================*/

#endif



