/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "PtrList.h"
#include "dbgSTD.h"

/*===========================================================================*\
\*===========================================================================*/

CPointerListElement::~CPointerListElement (void) {
    dhead ("CPointerListElement-Destructor", DEBUG_CPointerList);
    if (pNext && pPrev) {
        // It crashes if RemoveFromList is called for "Head" or "Tail"
        RemoveFromList();
    }
}

void CPointerListElement::AddAfter (CPointerListElement *pElement) {
    dhead ("CPointerListElement::AddAfter", DEBUG_CPointerList);
    dparams ("%x", pElement);
    dassert (pElement != 0);
    dassert (pElement->pNext == 0);
    dassert (pElement->pPrev == 0);
    pElement->pPrev = this;
    #ifndef DONT_USE_PROTECT_FEATURE
        pElement->SetProtectionLock (hProtect);
    #endif
    Protect_BeginWrite (hProtect);
    if (pNext) {
        pNext->pPrev    = pElement;
        pElement->pNext = pNext;
    }
    pNext = pElement;
    Protect_EndWrite (hProtect);
}

void CPointerListElement::AddBefore (CPointerListElement *pElement) {
    dhead ("CPointerListElement::AddBefore", DEBUG_CPointerList);
    dparams ("%x", pElement);
    dassert (pElement != 0);
    dassert (pElement->pNext == 0);
    dassert (pElement->pPrev == 0);
    pElement->pNext = this;
    #ifndef DONT_USE_PROTECT_FEATURE
        pElement->SetProtectionLock (hProtect);
    #endif
    Protect_BeginWrite (hProtect);
    dassert (pPrev != 0);
    pPrev->pNext    = pElement;
    pElement->pPrev = pPrev;
    pPrev           = pElement;
    Protect_EndWrite (hProtect);
}

CPointerListElement *CPointerListElement::FindFwd (CPointerListElement::ctFindFunc func, void *RefData) {
    dhead ("CPointerListElement::FindFwd", DEBUG_CPointerList);
    for (CPointerListElement *pLauf = GetNext(); (pLauf != 0); pLauf = pLauf->GetNext()) {
        switch (func (pLauf, RefData)) {
        case vYes:
            return pLauf;
        case vCancel:
        case vError:
            return 0;
        default:
            dassert (0);
            // no break!
        case vNo:
            // Nothing to do
            break;
        }
    }
    return 0;
}

CPointerListElement *CPointerListElement::FindRev (CPointerListElement::ctFindFunc func, void *RefData) {
    dhead ("CPointerListElement::FindRev", DEBUG_CPointerList);
    for (CPointerListElement *pLauf = GetPrev(); (pLauf != 0); pLauf = pLauf->GetNext()) {
        switch (func (pLauf, RefData)) {
        case vYes:
            return pLauf;
        case vCancel:
        case vError:
            return 0;
        default:
            dassert (0);
            // no break!
        case vNo:
            // Nothing to do
            break;
        }
    }
    return 0;
}

void CPointerListElement::InternalRemove (void) {
    dhead ("CPointerListElement::InternalRemove", DEBUG_CPointerList);
    if (pPrev != 0) {
        dassert (pNext != 0);
        dassert (pPrev->pNext == this);
        dassert (pNext->pPrev == this);
        pPrev->pNext = pNext;
        pNext->pPrev = pPrev;
        pNext = 0;
        pPrev = 0;
    } else {
        dassert (pNext == 0);
    }
}


/*===========================================================================*\
\*===========================================================================*/

CPointerList::~CPointerList (void) {
    dhead ("CPointerList-Destructor", DEBUG_CPointerList);
    RemoveAndDeleteAll();
    dassert (CountElements() == 0);
    Protect_Destroy (&hProtect);
}

tUInt CPointerList::CountElements (void) {
    dhead ("CPointerList::CountElements", DEBUG_CPointerList);
    tUInt count = 0;
    for (CPointerListElement *lauf = GetFirst(); lauf != 0; lauf = lauf->GetNext()) {
        count++;
    }
    RETURN ('x', count);
}

void CPointerList::RemoveAndDeleteAll (void) {
    dhead ("CPointerList::RemoveAndDeleteAll", DEBUG_CPointerList);
    CPointerListElement *pDelItem;
    while ((pDelItem = GetLast()) != 0) {
        delete pDelItem;
    }
}

/*===========================================================================*\
\*===========================================================================*/
