/*---------------------------------------------------------------------------*\
 * Copyright (C) 2000 AVM GmbH. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
 * ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
 * with the use or performance of this software. See the
 * GNU General Public License for further details.
\*---------------------------------------------------------------------------*/

#ifndef _ATOMAR_H_
#define _ATOMAR_H_

#include "aTypes.h"

#define MAP_ATOMIC_FUNCTION_CALLS

/*===========================================================================*\
    Definition
\*===========================================================================*/

tUInt atomicGet (tUInt *pValue);
tSInt atomicGet (tSInt *pValue);
void *atomicGet (void **ppValue);
tUInt atomicSet (tUInt *pUInt, tUInt value);
tSInt atomicSet (tSInt *pSInt, tSInt value);
void *atomicSet (void **pVoid, void *value);
tUInt atomicAdd (tUInt *pUInt, tUInt value);
tUInt atomicAdd (tUInt *pUInt, tSInt value);
tSInt atomicAdd (tSInt *pSInt, tSInt value);
tUInt atomicSub (tUInt *pUInt, tUInt value);
tBool atomicInc (tUInt *pUInt);
tBool atomicDec (tUInt *pUInt);
tUInt atomicExchange (tUInt *pUInt,  tUInt value);
tSInt atomicExchange (tSInt *pSInt,  tSInt value);
void *atomicExchange (void **ppVoid, void *value);

/*===========================================================================*\
    inline implementation:
\*===========================================================================*/
#if defined DONT_USE_PROTECT_FEATURE

inline tUInt atomicGet (tUInt *pValue) {
    return *pValue;
}

inline tSInt atomicGet (tSInt *pValue) {
    return *pValue;
}

inline void *atomicGet (void **ppValue) {
    return *ppValue;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tUInt atomicSet (tUInt *pUInt, tUInt value) {
    return *pUInt = value;
}

inline tSInt atomicSet (tSInt *pSInt, tSInt value) {
    return *pSInt = value;
}

inline void *atomicSet (void **ppVoid, void *value) {
    return *ppVoid = value;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tUInt atomicAdd (tUInt *pUInt, tUInt value) {
    tUInt old = *pUInt;
    *pUInt += value;
    return old;
}

inline tUInt atomicAdd (tUInt *pUInt, tSInt value) {
    tUInt old = *pUInt;
    *pUInt += value;
    return old;
}

inline tSInt atomicAdd (tSInt *pSInt, tSInt value) {
    tSInt old = *pSInt;
    *pSInt += value;
    return old;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tUInt atomicSub (tUInt *pUInt, tUInt value) {
    tUInt old = *pUInt;
    *pUInt -= value;
    return old;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tBool atomicInc (tUInt *pUInt) {
    return ((*pUInt)++ == 0);
}

inline tBool atomicDec (tUInt *pUInt) {
    return (--(*pUInt) == 0);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline tUInt atomicExchange (tUInt *pUInt,  tUInt value) {
    tUInt old = *pUInt;
    *pUInt = value;
    return old;
}

inline tSInt atomicExchange (tSInt *pSInt,  tSInt value) {
    tSInt old = *pSInt;
    *pSInt = value;
    return old;
}

inline void *atomicExchange (void **ppVoid, void *value) {
    void *old = *ppVoid;
    *ppVoid = value;
    return old;
}


// end of DONT_USE_PROTECT_FEATURE
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#else //if defined MAP_ATOMIC_FUNCTION_CALLS

#ifdef C_PLATFORM_64
#error "atomar doesn't support 64bit at the moment"
#endif

inline tUInt atomicGet (tUInt *pValue) {
    return (tUInt)atomicGet ((tSInt *)pValue);
}

inline void *atomicGet (void **ppValue) {
    return (void *)atomicGet ((tSInt *)ppValue);
}

inline tUInt atomicSet (tUInt *pUInt, tUInt value) {
    return (tUInt)atomicSet ((tSInt *)pUInt, (tSInt)value);
}

inline void *atomicSet (void **ppVoid, void *value) {
    return (void *)atomicSet ((tSInt *)ppVoid, (tSInt)value);
}

inline tUInt atomicAdd (tUInt *pUInt, tUInt value) {
    return (tUInt)atomicAdd ((tSInt *)pUInt, (tSInt)value);
}

inline tUInt atomicAdd (tUInt *pUInt, tSInt value) {
    return (tUInt)atomicAdd ((tSInt *)pUInt, value);
}

inline tUInt atomicSub (tUInt *pUInt, tUInt value) {
    return (tUInt)atomicAdd ((tSInt *)pUInt, -(tSInt)value);
}

inline tUInt atomicExchange (tUInt *pSInt,  tUInt value) {
    return (tUInt)atomicExchange ((tSInt *)pSInt, (tSInt)value);
}

inline void *atomicExchange (void **ppVoid, void *value) {
    return (void *)atomicExchange ((tSInt *)ppVoid, (tSInt)value);
}


#endif
/*===========================================================================*\
\*===========================================================================*/
#endif

