/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#ifndef _osmem_h
#define _osmem_h

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#include "aString.h"

#define ISALIGNED(p)   (0 == (PtrToUlong (p) & 3))

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void*   getPtr(tUByte *m);
tUInt64 get64 (tUByte *m);
tUInt32 get32 (tUByte *m);
tUInt16 get16 (tUByte *m);
tUByte  get8  (tUByte *m);

/*---------------------------------------------------------------------------*\
    Network Byte Order
\*---------------------------------------------------------------------------*/

tUInt64 getNBO64 (tUByte *m);
tUInt32 getNBO32 (tUByte *m);
tUInt16 getNBO16 (tUByte *m);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

tUInt64 getLE64 (tUByte *m);        // Little Endian
tUInt32 getLE32 (tUByte *m);
tUInt16 getLE16 (tUByte *m);

tUInt64 getBE64 (tUByte *m);        // Big Endian
tUInt32 getBE32 (tUByte *m);
tUInt16 getBE16 (tUByte *m);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void setPtr (tUByte *m, void   *d);
void set64  (tUByte *m, tUInt64 d);
void set32  (tUByte *m, tUInt32 d);
void set16  (tUByte *m, tUInt16 d);
void set8   (tUByte *m, tUByte  d);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void *getPtr (tUByte *m) {
    #ifdef C_PLATFORM_64
        return (void *)get64 (m);
    #else
        return (void *)get32 (m);
    #endif
}

inline tUInt64 get64 (tUByte *m) {
    tUInt64 value;
    s_memcpy (&value, m, sizeof (value));
    return value;
}

inline tUInt32 get32 (tUByte *m) {
    tUInt32 value;
    s_memcpy (&value, m, sizeof (value));
    return value;
}

inline tUInt16 get16 (tUByte *m) {
    tUInt16 value;
    s_memcpy (&value, m, sizeof (value));
    return value;
}

inline tUByte get8 (tUByte *m) {
    return *m;
}


/*---------------------------------------------------------------------------*\
    Network Byte Order
\*---------------------------------------------------------------------------*/

inline tUInt64 getNBO64 (tUByte *m) {
    return getBE64 (m);
}

inline tUInt32 getNBO32 (tUByte *m) {
    return getBE32 (m);
}

inline tUInt16 getNBO16 (tUByte *m) {
    return getBE16 (m);
}


/*---------------------------------------------------------------------------*\
    Implementation for little endian computers
\*---------------------------------------------------------------------------*/

inline tUInt64 getLE64 (tUByte *m) {
    return (((tUInt64)getLE32 (m + sizeof (tUInt32))) << 32) | (tUInt32)getLE32 (m);
}

inline tUInt32 getLE32 (tUByte *m) {
    return (((tUInt32)getLE16 (m + sizeof (tUInt16))) << 16) | (tUInt32)getLE16 (m);
}

inline tUInt16 getLE16 (tUByte *m) {
    return (tUInt16)((((tUInt16)m[1]) << 8) | (tUInt16)m[0]);
}

inline tUInt64 getBE64 (tUByte *m) {
    return (((tUInt64)getBE32 (m)) << 32) | (tUInt32)getBE32 (m + sizeof (tUInt32));
}

inline tUInt32 getBE32 (tUByte *m) {
    return (((tUInt32)getBE16 (m)) << 16) | (tUInt32)getBE16 (m + sizeof (tUInt16));
}

inline tUInt16 getBE16 (tUByte *m) {
    return (tUInt16)((((tUInt16)m[0]) << 8) | (tUInt16)m[1]);
}


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

inline void setPtr (tUByte *m, void *d)  {
    //assert (m != 0);
    #ifdef C_PLATFORM_64
        set64 (m, (tUInt64)d);
    #else
        set32 (m, (tUInt32)d);
    #endif
}

inline void set64 (tUByte *m, tUInt64 d)  {
    //assert (m != 0);
    s_memcpy (m, &d, sizeof (tUInt64));
}

inline void set32 (tUByte *m, tUInt32 d) {
    //assert (m != 0);
    s_memcpy (m, &d, sizeof (tUInt32));
}

inline void set16 (tUByte *m, tUInt16 d) {
    //assert (m != 0);
    s_memcpy (m, &d, sizeof (tUInt16));
}

inline void set8 (tUByte *m, tUByte d) {
    //assert (m != 0);
    *m = d;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#endif
